#!/usr/bin/env python

"""
ImageParser.py   $Id: ImageParser.py,v 1.19 2000/10/26 20:42:05 heiserd Exp $

This defines various classes to parse an image to a PluckerImageDocument.

It will try to identify the best availabel solution to do so and
define that as a default_parser function.


Some parts Copyright 1999 by Ondrej Palkovsky <ondrap@penguin.cz> and
others Copyright 1999 by Holger Duerer <holly@starship.python.net>

Distributable under the GNU General Public License Version 2 or newer.
"""

import os, sys, string, tempfile, re
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
from PyPlucker import PluckerDocs



binary_flag = ""
if sys.platform == 'os2':
    binary_flag = 'b'


#####################################################################
##
## This is the Image Magick parser from Chris.  It depends on os.popen
## and the availability of ImageMagick (convert) plus the Tbmp-tools
##
##

class ImageMagickImageParser:
    "Convert an image to the PalmBitmap. Uses 'convert' from ImageMagick"

    def __init__(self, url, type, data, config, maxwidth, maxheight, bpp, compress=1):
        self._url = url
        self._verbose = config.get_int ('verbosity', 1)
        self._config = config
        self._maxwidth = maxwidth
        self._maxheight = maxheight
        self._bpp = bpp
        self._doc = PluckerDocs.PluckerImageDocument (str(url), config)

        tmpfile = tempfile.mktemp()

        try:
            try:
                self.convert_to_pnm(data, tmpfile)
                self._doc.set_data(self.convert_to_Tbmp(tmpfile))
            finally:
                try: os.unlink(tmpfile)
                except: pass
        except RuntimeError:
            # This we pass through...
            raise
        except:
            raise RuntimeError, "Error while converting image (%s)." % self._url


    def get_plucker_doc(self):
        return self._doc


    def convert_to_Tbmp(self, tmpfile):
        
        ppmtoTbmp = self._config.get_string ('ppmtoTbmp_program', 'ppmtoTbmp')
        if self._bpp == 2:
            ppmtoTbmp = ppmtoTbmp + " -2bit"
        elif self._bpp == 4:
            ppmtoTbmp = ppmtoTbmp + " -4bit"

        if self._verbose <= 1:
            ppmtoTbmp = ppmtoTbmp + " -quiet"

	outfile = tempfile.mktemp()

        command = ppmtoTbmp + " " + tmpfile + " > " + outfile

        if self._verbose > 1:
            print "Running: ", command

        try:
            if os.system (command):
                raise RuntimeError, "Error while executing command '%s'" % command
            f = open (outfile, "rb")
            data = f.read ()
            f.close ()

            if len (data) == 0:
                # Oops, nothing fetched?!?
                raise RuntimeError, "No data from parsing image! (%s)" % self._url
        finally:
            try: os.unlink(outfile) 
            except: pass
        return data


    def convert_to_pnm(self, data, tmpfile):        

	infile = tempfile.mktemp()

        try:
            #print "Open infile " + infile
            f = open(infile, "wb")
            f.write(data)
            f.close()

            convert = self._config.get_string ('convert_program', 'convert')
            maxwidth = self._config.get_string ('maxwidth', '150')
            maxheight = self._config.get_string ('maxheight', '250')

            if self._maxheight != None:
                maxheight = self._maxheight
            if self._maxwidth != None:
                maxwidth = self._maxwidth

            size = "\"" + maxwidth + "x" + maxheight + ">\""

            if self._bpp == 1:
                ncolors = "2"
            elif self._bpp == 2:
                ncolors = "4"
            else:
                ncolors = "16"

            if self._bpp == 1:
                convert = convert + " -monochrome"

            command = convert + " -colors " + ncolors + " -dither -geometry " + size + " " + infile +" ppm:" + tmpfile

            if self._verbose > 1:
                print "Running: ", command
            if os.system (command):
                raise RuntimeError, "Error while executing command '%s'" % command
        finally:
            try: os.unlink(infile) 
            except: pass



## This is the standard parser from Ondrej.  It depends on os.popen
## and the availability of the pbmtools plus the Tbmp-tools
##
##


class NetPBMImageParser:
    "Convert an image to the PalmBitmap. Uses netpbm."    

    def __init__(self, url, type, data, config, maxwidth, maxheight, bpp, compress=1):
        self._type = type
        self._url = url
        self.scale = 0
        self.width = 1
        self.height = 1
        self._verbose = config.get_int ('verbosity', 1)
        self._config = config
        self._maxwidth = maxwidth
        self._maxheight = maxheight
        self._bpp = bpp
        self._doc = PluckerDocs.PluckerImageDocument(str(url), config)

        tmpfile = tempfile.mktemp()

        try:
            self.convert_to_pnm(data, tmpfile)
            self.fetch_info(tmpfile)
            self.define_scale()
            self._doc.set_data(self.convert_to_Tbmp(tmpfile))        
        except:
            try: os.unlink(tmpfile)
            except: pass
            raise RuntimeError, "Error while converting image."
        try: os.unlink(tmpfile)
        except: pass


    def get_plucker_doc(self):
        return self._doc


    def convert_to_Tbmp(self, tmpfile):
        ppmquant = self._config.get_string ('ppmquant_program', 'ppmquant')
        ppmtoTbmp = self._config.get_string ('ppmtoTbmp_program', 'ppmtoTbmp')
        if self._bpp == 1:
            ppmquant = ppmquant + " -fs 2 "
        elif self._bpp == 2:
            ppmquant = ppmquant + " -fs 4 "
            ppmtoTbmp = ppmtoTbmp + " -2bit"
        else:
            ppmquant = ppmquant + " -fs 16 "
            ppmtoTbmp = ppmtoTbmp + " -4bit"

        if self._verbose <= 1:
            ppmtoTbmp = ppmtoTbmp + " -quiet "
            ppmquant = ppmquant + " -quiet "

        if not self.scale:
            command = ppmquant  + tmpfile + "|" + ppmtoTbmp
        else:
            command = self.scale + tmpfile + "|" + ppmquant + "|" + ppmtoTbmp

        if self._verbose > 1:
            print "Running: ", command
        pipe = os.popen(command, "r"+binary_flag)
        data = pipe.read()
        pipe.close()
        if len (data) == 0:
            # Oops, nothing fetched?!?
            raise RuntimeError, "No data from parsing image! (%s)" % self._url
        return data


    def define_scale(self):
        maxwidth = self._config.get_int ('maxwidth', 150)
        maxheight = self._config.get_int ('maxheight', 250)
        pnmscale = self._config.get_string ('pnmscale_program', 'pnmscale')

        if self._maxwidth != None:
            maxwidth = string.atoi(self._maxwidth)
        if self._maxheight != None:
            maxheight = string.atoi(self._maxheight)

        if (self.width>maxwidth or self.height>maxheight):
            if (self.width/maxwidth>self.height/maxheight):
                self.scale = pnmscale + (" -xsize %d " % maxwidth)
            else:
                self.scale = pnmscale + (" -ysize %d " % maxheight)


    def fetch_info(self, tmpfile):
        pnmfile = self._config.get_string ('pnmfile_program', 'pnmfile')
        pipe = os.popen(pnmfile + " " + tmpfile, "r")
        info = pipe.read()
        pipe.close()
        info = string.split(info) #info is now an array
        self.width = string.atoi(info[3])
        self.height = string.atoi(info[5])        


    def convert_to_pnm(self, data, tmpfile):        
        giftopnm = self._config.get_string ('giftopnm_program', 'giftopnm')
        djpeg = self._config.get_string ('djpeg_program', 'djpeg')
        pngtopnm = self._config.get_string ('pngtopnm_program', 'pngtopnm')
        if (self._type=='image/gif'):
            command = giftopnm + ' >' + tmpfile
        elif (self._type=='image/jpeg'):
            command = djpeg + ' -pnm ' + ' >' + tmpfile
        elif (self._type=='image/png'):
            command = pngtopnm + ' >' + tmpfile
        if self._verbose > 1:
            print "Running: ", command
        pipe = os.popen(command, "w"+binary_flag)
        pipe.write(data)
        pipe.close()



#####################################################################
##
## This is a parser that relies on the Python Imaging Library (PIL)
## to do all the work (except converting to Tbmp format at the end).
## 
##
##
class PythonImagingLibraryParser:

    "Convert an image to the PalmBitmap. Uses 'convert' from ImageMagick"

    def __init__(self, url, type, data, config, maxwidth, maxheight, bpp, compress=1):
        import Image
        self._url = url
        self._verbose = config.get_int ('verbosity', 1)
        self._config = config
        self._maxwidth = maxwidth
        self._maxheight = maxheight
        self._bpp = bpp
        self._doc = PluckerDocs.PluckerImageDocument (str(url), config)

        tmpfile = tempfile.mktemp()

        if self._bpp == 1:
            colors = 2
        elif self._bpp == 2:
            colors = 4
        else:
            colors = 16

        try:
            if self._verbose > 1:
                print "Converting image %s with PIL" % url
            im = Image.open (StringIO (data))
            im.draft("L", im.size)
            im = im.convert ("L", colors=colors)
            try:
                apply(im.save, (tmpfile, "PPM"))
                self._doc.set_data (self.convert_to_Tbmp (tmpfile))
            finally:
                try: os.unlink(tmpfile)
                except: pass
        except:
            if self._verbose > 1:
                import traceback
                traceback.print_exc()
            raise RuntimeError, "Error while converting image."


    def get_plucker_doc(self):
        return self._doc


    def convert_to_Tbmp(self, tmpfile):
        ppmtoTbmp = self._config.get_string ('ppmtoTbmp_program', 'ppmtoTbmp')
        if self._bpp == 2:
            ppmtoTbmp = ppmtoTbmp + " -2bit"
        elif self._bpp == 4:
            ppmtoTbmp = ppmtoTbmp + " -4bit"

        if self._verbose <= 1:
            ppmtoTbmp = ppmtoTbmp + " -quiet"

        command = ppmtoTbmp + " " + tmpfile

        if self._verbose > 1:
            print "Running: ", command
        pipe = os.popen(command, "r"+binary_flag)
        data = pipe.read()
        pipe.close()
        if len (data) == 0:
            # Oops, nothing fetched?!?
            raise RuntimeError, "No data from parsing image! (%s)" % self._url
        return data



#####################################################################
##
## This is a parser for Windows systems.  It relies on a 
## and the availability of the pbmtools plus the Tbmp-tools
##
##
class WindowsImageParser:
    """Do it on Windows.  Ask Dirk Heiser <plucker@dirk-heiser.de> if
    these tools goof up.  I cannot test it."""

    def __init__ (self, url, type, data, config, maxwidth, maxheight, bpp, compress=1):

        def quotestr(path):
            out = string.strip(path)
            if (string.find(out,' ') != -1) and (string.find(out,'"') == -1):
                out = "\""+out+"\""
            return out


        verbose = config.get_int ('verbosity', 1)

        imagemagick = "%s %s" % (quotestr(config.get_string ('convert_program','convert.exe')),config.get_string ('convert_program_parameter','%type%:%input% bmp:%output%'))
        bmp2tbmp = "%s %s" % (quotestr(config.get_string ('bmp_to_tbmp', 'Bmp2Tbmp.exe')),config.get_string ('bmp_to_tbmp_parameter','-i=%input% -o=%output% -maxwidth=%maxwidth% -maxheight=%maxheight% -compress=%compress% -color=%colors%'))

        if compress:
            compress_str = config.get_string ('tbmp_compression_type','yes')
        else:
            compress_str = 'no'

        if maxwidth == None:
            maxwidth_str = config.get_string ('maxwidth', '150')
        else:
            maxwidth_str = '%s' % (maxwidth)

        if maxheight == None:
            maxheight_str = config.get_string ('maxheight', '250')
        else:
            maxheight_str = '%s' % (maxheight)

        # Check for Bitmap format
        if bpp == 1:
            bpp = 'M'
        elif bpp == 2:
            bpp = 'G4'
        elif bpp == 4:
            bpp = 'G16'
        elif bpp == 8:
            bpp = 'C256'
        else:
            raise RuntimeError, "Color format not Supported"
        
        
        # First check the type
        if type == 'image/jpeg':
            type = 'JPG'
        elif type == 'image/gif':
            type = 'GIF'
        elif type == 'image/png':
            type = 'PNG'
        else:
            raise RuntimeError, "Windows parser can only handle gif, png and jpeg images"

        # Create Temp Files
        in_name   = tempfile.mktemp ()
        mid_name  = tempfile.mktemp ()
        out_name  = tempfile.mktemp ()

        # write the data to the in temp file
        f = open (in_name, "wb")
        f.write (data)
        f.close ()

        # convert to bmp
        imagemagick = string.replace (imagemagick , '%compress%', compress_str)
        imagemagick = string.replace (imagemagick , '%colors%', bpp)
        imagemagick = string.replace (imagemagick , '%maxwidth%', maxwidth_str)
        imagemagick = string.replace (imagemagick , '%maxheight%', maxheight_str)
        imagemagick = string.replace (imagemagick , '%input%', in_name)
        imagemagick = string.replace (imagemagick , '%output%', mid_name)
        imagemagick = string.replace (imagemagick , '%type%', type)
        command = imagemagick
        if verbose == 2:
            print "Running %s" % command
        if verbose < 2:
            command = command + " > nul"
        res = os.system (command)
        if res:
            raise RuntimeError, "\nCommand %s failed with code %d\n" % (command, res)
        if not os.path.isfile(mid_name):
            raise RuntimeError, "\nCommand %s failed\n" % (command)

        # convert to Tbmp
        bmp2tbmp = string.replace (bmp2tbmp, '%compress%', compress_str)
        bmp2tbmp = string.replace (bmp2tbmp, '%colors%', bpp)
        bmp2tbmp = string.replace (bmp2tbmp, '%maxwidth%', maxwidth_str)
        bmp2tbmp = string.replace (bmp2tbmp, '%maxheight%', maxheight_str)
        bmp2tbmp = string.replace (bmp2tbmp, '%input%', mid_name)
        bmp2tbmp = string.replace (bmp2tbmp, '%output%', out_name)
        bmp2tbmp = string.replace (bmp2tbmp , '%type%', type)
        command = bmp2tbmp
        if verbose == 2:
            print "Running %s" % command
        if verbose < 2:
            command = command + " > nul"
        res = os.system (command)
        if res:
            raise RuntimeError, "\nCommand %s failed with code %d\n" % (command, res)
        if not os.path.isfile(out_name):
            raise RuntimeError, "\nCommand %s failed\n" % (command)

        f = open (out_name, "rb")
        data = f.read ()
        f.close ()

        # clean up the temp. files
        try:
            os.unlink (in_name)
            os.unlink (mid_name)
            os.unlink (out_name)
        except:
            pass
        
        self._doc = PluckerDocs.PluckerImageDocument (str (url), config)
        self._doc.set_data (data)

    def get_plucker_doc(self):
        return self._doc


if sys.platform == 'win32':
    DefaultParser = WindowsImageParser
else:
    try:
        # try to determine whether PIL is installed
        from Image import FLOYDSTEINBERG
        del FLOYDSTEINBERG
        DefaultParser = PythonImagingLibraryParser
    except:
        ## try to determine whether Image Magick's convert is installed
        if sys.platform == 'os2':
            res = os.system ("convert.exe > nul")
        else:
            res = os.system ("convert >/dev/null 2>&1")
        if res:
            # ImageMagick not available
            DefaultParser = NetPBMImageParser
        else:
            DefaultParser = ImageMagickImageParser



def get_default_parser (config):
    parser = config.get_string ('image_parser')
    if parser is None:
        return DefaultParser

    parser = string.lower (parser)
    if parser == "windows":
        return WindowsImageParser
    elif parser == "netpbm":
        return NetPBMImageParser
    elif parser == "imagemagick":
        return ImageMagickImageParser
    elif parser == "pil":
        return PythonImagingLibraryParser
    else:
        return DefaultParser

if __name__ == '__main__':
    # Called as a script
    print "This file currently does nothing when called as a script"
