/*
 * $Id: anchor.c,v 1.22 2000/10/22 20:17:32 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "anchor.h"
#include "debug.h"
#include "link.h"
#include "prefsdata.h"
#include "util.h"

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define MAXANCHORS 100
#define NOT_FOUND -1

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static Int16 nextAnchorId = 1;	/* To link VisibleAnchors together */
static VisibleAnchor *currentAnchor;
static VisibleAnchor visibleAnchors[MAXANCHORS];

/***********************************************************************
 *
 * FUNCTION:    AnchorInViewport
 *
 * DESCRIPTION: Check if given anchor is within viewport's boundaries.
 *
 * PARAMETERS:  anchor - pointer to visible anchor
 *
 * RETURN:      True if anchor is located inside viewport
 *
 ***********************************************************************/
static Boolean AnchorInViewport(VisibleAnchor *anchor)
{
	RectangleType *bounds = &(anchor->bounds);

	if ((bounds->topLeft.y > TopLeftY() + ExtentY()) || 
	    (bounds->topLeft.y + bounds->extent.y < TopLeftY()))
		return false;
	else
		return true;
}

/***********************************************************************
 *
 * FUNCTION:    GetFreeAnchor
 *
 * DESCRIPTION: Retrieve free visible anchor.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Pointer to a visible anchor or NULL if no more anchors 
 *              are available
 *
 ***********************************************************************/
static VisibleAnchor *GetFreeAnchor(void)
{
	Int16 i;

	for (i = 0; i < MAXANCHORS; i++)
		if (visibleAnchors[i].state == ANCHOR_UNUSED)
			return visibleAnchors + i;

	return NULL;
}

/***********************************************************************
 *
 * FUNCTION:    DeleteLikeAnchors
 *
 * DESCRIPTION: Remove all anchors which have the same bounds as the 
 *              given anchor.
 *
 * PARAMETERS:  anchor - pointer to visible anchor
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void DeleteLikeAnchors(VisibleAnchor *anchor)
{
	RectangleType *anchorBounds = &(anchor->bounds);
	VisibleAnchor *thisAnchor = visibleAnchors;
	Int16 i;

	for (i = 0; i < MAXANCHORS; i++, thisAnchor++) {
		if (MemCmp(anchorBounds, &(visibleAnchors[i].bounds), sizeof(RectangleType)))
			continue;

		thisAnchor->state = ANCHOR_UNUSED;
	}
}

/***********************************************************************
 *
 * FUNCTION:    StrikeThrough
 *
 * DESCRIPTION: Add a strike-through for anchor
 *
 * PARAMETERS:  anchor - pointer to visible anchor
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void StrikeThrough(const VisibleAnchor *anchor)
{
	Int16 anchorY = anchor->bounds.topLeft.y + anchor->bounds.extent.y -
			FntCharHeight() / 2;

	WinDrawLine(anchor->bounds.topLeft.x, 
		    anchorY, 
		    anchor->bounds.topLeft.x + anchor->bounds.extent.x, 
		    anchorY);
}

/***********************************************************************
 *
 * FUNCTION:    SetAnchorData
 *
 * DESCRIPTION: Allocate new anchor and store relevant data in it.
 *
 * PARAMETERS:  tContext  - pointer to text context
 *              anchorId  - unique ID for anchor
 *              reference - record reference
 *              pOffset   - offset to first paragraph
 *              image     - image reference
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void SetAnchorData(const TextContext *tContext, const Int16 anchorId, 
			  const Int16 reference, const Int16 pOffset, const Int16 image)
{
	VisibleAnchor *newAnchor;

	newAnchor = GetFreeAnchor();
	if (newAnchor == NULL) {
		FATAL("no free anchors!!\n");
		return;
	}
	currentAnchor = newAnchor;

	newAnchor->bounds.topLeft.x = tContext->cursorX;
	newAnchor->anchorId = anchorId;
	newAnchor->reference = reference;
	newAnchor->paragraphOffset = pOffset;
	newAnchor->image = image;
	newAnchor->state = ANCHOR_LIMBO;
	newAnchor->noUnderline = false;
}

/***********************************************************************
 *
 * FUNCTION:    AdjustVisibleAnchors
 *
 * DESCRIPTION: Adjust all of the visible anchors by the given amount.
 *
 * PARAMETERS:  adjustment - adjustment in pixels
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void AdjustVisibleAnchors(const Int16 adjustment)
{
	VisibleAnchor *thisAnchor = visibleAnchors;
	Int16 i;

	for (i = 0; i < MAXANCHORS; i++, thisAnchor++) {
		if (thisAnchor->state == ANCHOR_UNUSED)
			continue;

		thisAnchor->bounds.topLeft.y += adjustment;
		if (!AnchorInViewport(thisAnchor))
			thisAnchor->state = ANCHOR_UNUSED;
	}
}

/***********************************************************************
 *
 * FUNCTION:    ClearVisibleAnchors
 *
 * DESCRIPTION: Clear the visible anchors structure.
 *
 * PARAMETERS:  None 
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ClearVisibleAnchors(void)
{
	MemSet((void *) &visibleAnchors, sizeof(visibleAnchors), 0);
}

/***********************************************************************
 *
 * FUNCTION:    FindVisibleAnchor
 *
 * DESCRIPTION: Find visible anchor at given location.
 *
 * PARAMETERS:  x - the x coordinate
 *              y - the y coordinate
 *
 * RETURN:      The anchor ID, or -1 if no anchor was found at the given
 *              location
 *
 ***********************************************************************/
Int16 FindVisibleAnchor(const Int16 x, const Int16 y)
{
	Int16 i;

	for (i = 0; i < MAXANCHORS; i++)
		if ((visibleAnchors[i].state != ANCHOR_UNUSED) && 
		    (RctPtInRectangle(x, y, &(visibleAnchors[i].bounds))))
			return i;

	return NOT_FOUND;
}

/***********************************************************************
 *
 * FUNCTION:    GetVisibleReference
 *
 * DESCRIPTION: Return the reference of a visible anchor.
 *
 * PARAMETERS:  index - index of visible anchor
 *
 * RETURN:      The reference of the visible anchor
 *
 ***********************************************************************/
Int16 GetVisibleReference(const Int16 index)
{
	return visibleAnchors[index].reference;
}

/***********************************************************************
 *
 * FUNCTION:    GetVisibleOffset
 *
 * DESCRIPTION: Return the offset of a visible anchor.
 *
 * PARAMETERS:  index - index of visible anchor
 *
 * RETURN:      The offset of the visible anchor
 *
 ***********************************************************************/
Int16 GetVisibleOffset(const Int16 index)
{
	return visibleAnchors[index].paragraphOffset;
}

/***********************************************************************
 *
 * FUNCTION:    GetVisibleImage
 *
 * DESCRIPTION: Return the image reference for a visible anchor.
 *
 * PARAMETERS:  index - index of visible anchor
 *
 * RETURN:      The image reference for the visible anchor
 *
 ***********************************************************************/
Int16 GetVisibleImage(const Int16 index)
{
	return visibleAnchors[index].image;
}

/***********************************************************************
 *
 * FUNCTION:    GetVisibleImagePosition
 *
 * DESCRIPTION: Return the image coordinates for a visible anchor.
 *
 * PARAMETERS:  index - index of visible anchor
 *
 * RETURN:      The RectangleType for the visible anchor
 *
 ***********************************************************************/
RectangleType GetVisibleImagePosition(const Int16 index)
{
	return visibleAnchors[index].bounds;
}

/***********************************************************************
 *
 * FUNCTION:    HighlightAnchor
 *
 * DESCRIPTION: Set highlight status for the anchor.
 *
 * PARAMETERS:  control - control ID of the object
 *              state   - anchor state (ANCHOR_SELECTED or 
 *                                      ANCHOR_UNSELECTED)
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void HighlightAnchor(const Int16 control, const AnchorStateType state)
{
	VisibleAnchor *thisAnchor = visibleAnchors;
	Int16 anchorId = visibleAnchors[control].anchorId;
	Int16 topY = TopLeftY();
	Int16 bottomY = topY + ExtentY();
	Int16 i;

	MSG_IF(state != ANCHOR_SELECTED && state != ANCHOR_UNSELECTED, 
	       _("invalid anchor state, %d\n", state));

	for (i = 0; i < MAXANCHORS; i++, thisAnchor++) {
		if (thisAnchor->anchorId == anchorId && 
		    thisAnchor->state != ANCHOR_UNUSED && 
		    thisAnchor->state != state) {
			if (thisAnchor->bounds.topLeft.y < topY) {
				thisAnchor->bounds.extent.y -= topY - thisAnchor->bounds.topLeft.y;
				thisAnchor->bounds.topLeft.y = topY;
			} else if ((thisAnchor->bounds.topLeft.y + thisAnchor->bounds.extent.y) > bottomY) {
				thisAnchor->bounds.extent.y = bottomY - thisAnchor->bounds.topLeft.y;
			}
			WinInvertRectangle(&(thisAnchor->bounds), 0);
			thisAnchor->state = state;
		}
	}
}

/***********************************************************************
 *
 * FUNCTION:    StartAnchor
 *
 * DESCRIPTION: Initialize a new visible anchor.
 *
 * PARAMETERS:  tContext  - pointer to text context
 *              reference - record reference
 *              pOffset   - offset to first paragraph
 *              image     - image reference
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void StartAnchor(const TextContext *tContext, const Int16 reference, 
		 const Int16 pOffset, const Int16 image)
{
	SetAnchorData(tContext, nextAnchorId++, reference, pOffset, image);
}

/***********************************************************************
 *
 * FUNCTION:    ContinueAnchor
 *
 * DESCRIPTION: Handle multi-line anchors.
 *
 * PARAMETERS:  tContext   - pointer to text context
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ContinueAnchor(const TextContext *tContext)
{
	SetAnchorData(tContext, currentAnchor->anchorId, currentAnchor->reference, 
		      currentAnchor->paragraphOffset, currentAnchor->image);
}

/***********************************************************************
 *
 * FUNCTION:    RestartAnchor
 *
 * DESCRIPTION: Restart anchor.
 *
 * PARAMETERS:  tContext   - pointer to text context
 *              height     - height of line
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void RestartAnchor(const TextContext *tContext, const Int16 height)
{
	StopAnchor(tContext, height);
	SetAnchorData(tContext, currentAnchor->anchorId, currentAnchor->reference, 
		      currentAnchor->paragraphOffset, currentAnchor->image);
}

/***********************************************************************
 *
 * FUNCTION:    StopImageAnchor
 *
 * DESCRIPTION: Mark the end of a visible image anchor.
 *
 * PARAMETERS:  tContext - pointer to text context
 *              height   - height of image
 *              width    - width of image
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void StopImageAnchor(TextContext *tContext, const Int16 height, 
		     const Int16 width)
{
	currentAnchor->noUnderline = true;
	tContext->cursorX += width;

	RestartAnchor(tContext, height);

	tContext->cursorX -= width;
	currentAnchor->noUnderline = false;
}

/***********************************************************************
 *
 * FUNCTION:    StopAnchor
 *
 * DESCRIPTION: Mark the end of a visible anchor.
 *
 * PARAMETERS:  tContext   - pointer to text context
 *              height     - height of line
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void StopAnchor(const TextContext *tContext, const Int16 height)
{
	VisibleAnchor *newAnchor = currentAnchor;

	newAnchor->bounds.topLeft.y = tContext->cursorY - height;
	newAnchor->bounds.extent.x = tContext->cursorX - newAnchor->bounds.topLeft.x;
	newAnchor->bounds.extent.y = height;

	// Only underline visible anchors
	if (AnchorInViewport(newAnchor) && tContext->writeMode) {
		// No underline for images
		if (!newAnchor->noUnderline && newAnchor->bounds.extent.x != 0) {
			if (LinkVisited(newAnchor->reference)) {
				if (Prefs()->strikethrough)
					StrikeThrough(newAnchor);

				WinDrawGrayLine(newAnchor->bounds.topLeft.x, 
						tContext->cursorY - 1, 
						tContext->cursorX, 
						tContext->cursorY - 1);

			} else if (Prefs()->underlineMode) {
				WinDrawGrayLine(newAnchor->bounds.topLeft.x, 
						tContext->cursorY - 1, 
						tContext->cursorX, 
						tContext->cursorY - 1);
			} else {
					WinDrawLine(newAnchor->bounds.topLeft.x, 
					    tContext->cursorY - 1, 
						tContext->cursorX, 
					    tContext->cursorY - 1);
			}
		}
		DeleteLikeAnchors(newAnchor);
		newAnchor->state = ANCHOR_UNSELECTED;
	} else
		newAnchor->state = ANCHOR_UNUSED;	/* oooooh :( */
}
// vim: ts=4
