/*
 * $Id: bookmark.c,v 1.12 2000/10/22 20:17:32 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "bookmark.h"
#include "debug.h"
#include "document.h"
#include "documentdata.h"
#include "history.h"
#include "resourceids.h"
#include "util.h"

static UInt16 CreateBookmarkList(ListType *list) __attribute__ ((section ("plkr")));
UInt16 CreateLongBookmarkList(ListType *list) __attribute__ ((section ("plkr")));
UInt16 CreateShortBookmarkList(ListType *list) __attribute__ ((section ("plkr")));
void ReleaseBookmarkList(void) __attribute__ ((section ("plkr")));
Int16 RestoreBookmarkData(Int16 index) __attribute__ ((section ("plkr")));
void AddBookmark(Char *name) __attribute__ ((section ("plkr")));
Boolean DeleteBookmark(Int16 index) __attribute__ ((section ("plkr")));

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define NOT_FOUND -1
#define ADD_EDIT_OPTIONS true
#define NO_OPTIONS false

/***********************************************************************
 *
 *      Private Variables
 *
 ***********************************************************************/
static MemHandle nameListHandle = NULL;
static Char addBookmark[25];
static Char editBookmark[25];
static Boolean popupList =  false;

/***********************************************************************
 *
 * FUNCTION:    DrawListItem
 *
 * DESCRIPTION: Draw item in list
 *
 * PARAMETERS:  itemNum   - item number
 *              bounds    - pointer to a RectangleType structure that
 *                          specifies the bounds for the list item
 *              itemsText - pointer to an array of text strings.
 *
 * RETURN:      Number of bookmarks in list
 *
 ***********************************************************************/
static void DrawListItem(Int16 itemNum, RectangleType *bounds, Char **itemsText)
{
	FontID oldFont;

	SET_A4_FROM_A5

	if (popupList && (itemNum == 0 || itemNum == 1))
		oldFont = FntSetFont(boldFont);
	else
		oldFont = FntSetFont(stdFont);

	if (itemsText[itemNum] != NULL)
		WinDrawChars(itemsText[itemNum], StrLen(itemsText[itemNum]), bounds->topLeft.x, bounds->topLeft.y);

	FntSetFont(oldFont);

	RESTORE_A4
}

/***********************************************************************
 *
 * FUNCTION:    CreateBookmarkList
 *
 * DESCRIPTION: Populate list with bookmarks
 *
 * PARAMETERS:  list   - pointer to list
 *
 * RETURN:      Number of bookmarks in list
 *
 ***********************************************************************/
static UInt16 CreateBookmarkList(ListType *list)
{
	UInt16 entries;
	UInt16 i;
	MemHandle handle = NULL;
	Char **nameList = NULL;
	UInt8 *bookmarkPtr = NULL;
	UInt16 extraListItems;

	extraListItems = (popupList ? 2 : 0);

	if (ReturnDynamicHandle(INTERNAL_BOOKMARKS_ID, 0, &handle) != 0) {
		if (popupList)
			entries = 2;
		else {
			entries = 0;
			LstSetListChoices(list, nameList, entries);
			LstSetDrawFunction (list, DrawListItem);
			return entries;
		}
	} else {
		bookmarkPtr = (UInt8 *) MemHandleLock(handle);
		FATAL_IF(bookmarkPtr == NULL, "MemHandleLock failed\n");
		ErrFatalDisplayIf(bookmarkPtr == NULL, "CreateBookmarkList: MemHandleLock failed");

		entries = *(bookmarkPtr + 2) * 256 + *(bookmarkPtr + 3) + extraListItems;
		bookmarkPtr = (void *) bookmarkPtr + 3 * sizeof(UInt16);
	}

	// Allocate arrays for name list
	nameListHandle = MemHandleNew(entries * sizeof(*nameList));
	if (nameListHandle == NULL) {
		if (entries > extraListItems)
			MemHandleUnlock(handle);

		return 0;
	}
	nameList = MemHandleLock(nameListHandle);
	FATAL_IF(nameList == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(nameList == NULL, "CreateBookmarkList: MemHandleLock failed");

	if (popupList) {
		SysCopyStringResource(addBookmark, strAddBookmark);
		nameList[0] = addBookmark;
		SysCopyStringResource(editBookmark, strEditBookmark);
		nameList[1] = editBookmark;
	}
	for (i = extraListItems; i < entries; i++) {
		nameList[i] = (Char *) bookmarkPtr;
		bookmarkPtr = (void *) bookmarkPtr + StrLen(bookmarkPtr) + 1;
	}
	LstSetListChoices(list, nameList, entries);
	LstSetDrawFunction (list, DrawListItem);

	if (entries > 12)
		LstSetHeight(list, 12);
	else
		LstSetHeight(list, entries);

	if (entries > extraListItems)
		MemHandleUnlock(handle);

	return entries;
}

/***********************************************************************
 *
 * FUNCTION:    CreateLongBookmarkList
 *
 * DESCRIPTION: Populate list with bookmarks (including Add/Edit Bookmarks)
 *
 * PARAMETERS:  list - pointer to list
 *
 * RETURN:      Number of bookmarks in list
 *
 ***********************************************************************/
UInt16 CreateLongBookmarkList(ListType *list)
{
	popupList = ADD_EDIT_OPTIONS;
	return CreateBookmarkList(list);
}

/***********************************************************************
 *
 * FUNCTION:    CreateShortBookmarkList
 *
 * DESCRIPTION: Populate list with bookmarks (skip Add/Edit Bookmarks)
 *
 * PARAMETERS:  list - pointer to list
 *
 * RETURN:      Number of bookmarks in list
 *
 ***********************************************************************/
UInt16 CreateShortBookmarkList(ListType *list)
{
	popupList = NO_OPTIONS;
	return CreateBookmarkList(list);
}

/***********************************************************************
 *
 * FUNCTION:    ReleaseBookmarkList
 *
 * DESCRIPTION: Unlock list handle and free allocated memory
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ReleaseBookmarkList(void)
{
	if (nameListHandle) {
		MemHandleUnlock(nameListHandle);

		MemHandleFree(nameListHandle);
		nameListHandle = NULL;
	}
}

/***********************************************************************
 *
 * FUNCTION:    RestoreBookmarkData
 *
 * DESCRIPTION: Restore document data (document ID, vertical offset, 
 *              first and last paragraph) for current bookmark
 *
 * PARAMETERS:  index - index in bookmarks list
 *
 * RETURN:      The document ID (if there are no bookmarks -1 is returned)
 *
 ***********************************************************************/
Int16 RestoreBookmarkData(Int16 index)
{
	DynamicDocument *dynamic;
	BookmarkData bookmarkData;
	UInt8 *bookmarkPtr;
	MemHandle handle = NULL;
	UInt16 offset;

	if (ReturnDynamicHandle(INTERNAL_BOOKMARKS_ID, 0, &handle) != 0)
		return NOT_FOUND;

	// Skip "Add/Edit Bookmarks"
	index -= 2;

	bookmarkPtr = (UInt8 *) MemHandleLock(handle);
	FATAL_IF(bookmarkPtr == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(bookmarkPtr == NULL, "RestoreBookmarkData: MemHandleLock failed");

	offset = *(bookmarkPtr + 4) * 256 + *(bookmarkPtr + 5);
	bookmarkPtr = (void *) bookmarkPtr + offset + index * sizeof(BookmarkData);
	MemMove(&bookmarkData, bookmarkPtr, sizeof(BookmarkData));

	dynamic = (DynamicDocument *) MemHandleLock(GetDynamicHandle(bookmarkData.documentId));
	DmWrite(dynamic, OFFSETOF(DynamicDocument, verticalOffset),
		(void *) &bookmarkData, sizeof(BookmarkData) - sizeof(Int16));

	// Add to history
	AddHistoryDocument(bookmarkData.documentId);
	SetHistoryFirstParagraph(bookmarkData.firstVisibleParagraph, bookmarkData.firstParagraphY);
	SetHistoryLastParagraph(bookmarkData.lastVisibleParagraph, bookmarkData.lastParagraphY);
	SetHistoryVerticalOffset(bookmarkData.verticalOffset);

	MemPtrUnlock(dynamic);
	MemHandleUnlock(handle);

	return bookmarkData.documentId;
}

/***********************************************************************
 *
 * FUNCTION:    AddBookmark
 *
 * DESCRIPTION: Add data for bookmark
 *
 * PARAMETERS:  name - name of bookmark
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void AddBookmark(Char *name)
{
	DynamicDocument *dynamic;
	BookmarkData bookmarkData;
	MemHandle handle = NULL;
	UInt8 *bookmarkPtr;
	UInt16 entries, offset;
	UInt16 newSize, i;
	UInt32 endOfRecord;
	UInt16 nameLength = StrLen(name) + 1;

	if (ReturnDynamicHandle(INTERNAL_BOOKMARKS_ID, 0, &handle) != 0)
		if (AddBookmarkRecord(&handle) != 0)
			return;

	endOfRecord = MemHandleSize(handle);
	newSize = endOfRecord + sizeof(BookmarkData) + nameLength;
	ResizeRecord(INTERNAL_BOOKMARKS_ID, newSize, &handle);

	if (handle == NULL)
		return;

	bookmarkPtr = (UInt8 *) MemHandleLock(handle);
	FATAL_IF(bookmarkPtr == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(bookmarkPtr == NULL, "AddBookmark: MemHandleLock failed");

	entries = *(bookmarkPtr + 2) * 256 + *(bookmarkPtr + 3) + 1;
	offset = *(bookmarkPtr + 4) * 256 + *(bookmarkPtr + 5) + nameLength;
	DmWrite(bookmarkPtr, sizeof(Header), (void *) &entries, sizeof(UInt16));
	DmWrite(bookmarkPtr, sizeof(Header) + sizeof(UInt16), (void *) &offset, sizeof(UInt16));

	dynamic = (DynamicDocument *) MemHandleLock(GetDynamicDocument());
	bookmarkData.verticalOffset = dynamic->verticalOffset;
	bookmarkData.firstVisibleParagraph = dynamic->firstVisibleParagraph;
	bookmarkData.firstParagraphY = dynamic->firstParagraphY;
	bookmarkData.lastVisibleParagraph = dynamic->lastVisibleParagraph;
	bookmarkData.lastParagraphY = dynamic->lastParagraphY;
	bookmarkData.documentId = GetHistoryDocID();

	// Write new block of bookmark data
	DmWrite(bookmarkPtr, endOfRecord + nameLength, &bookmarkData, sizeof(BookmarkData));
	endOfRecord -= sizeof(BookmarkData);

	// Reshuffle old blocks with bookmark data
	for (i = 1; i < entries; i++) {
		MemMove(&bookmarkData, bookmarkPtr + endOfRecord, sizeof(BookmarkData));
		DmWrite(bookmarkPtr, endOfRecord + nameLength, &bookmarkData, sizeof(BookmarkData));
		endOfRecord -= sizeof(BookmarkData);
	}

	// Write new bookmark name
	DmStrCopy(bookmarkPtr, offset - nameLength, name);

	MemHandleUnlock(GetDynamicDocument());
	MemHandleUnlock(handle);
}

/***********************************************************************
 *
 * FUNCTION:    DeleteBookmark
 *
 * DESCRIPTION: Delete bookmark
 *
 * PARAMETERS:  index - index in bookmarks list
 *
 * RETURN:      True if list is empty
 *
 ***********************************************************************/
Boolean DeleteBookmark(Int16 index)
{
	BookmarkData bookmarkData;
	MemHandle handle = NULL;
	UInt8 *bookmarkPtr;
	UInt8 *readPtr;
	UInt16 entries, offset, tempOffset;
	UInt16 newSize, i;
	UInt32 endOfRecord;
	UInt16 nameLength, length;
	UInt16 counter;
	Char tempString[21];

	if (ReturnDynamicHandle(INTERNAL_BOOKMARKS_ID, 0, &handle) != 0)
		return true;

	bookmarkPtr = (UInt8 *) MemHandleLock(handle);
	FATAL_IF(bookmarkPtr == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(bookmarkPtr == NULL, "DeleteBookmark: MemHandleLock failed");

	entries = *(bookmarkPtr + 2) * 256 + *(bookmarkPtr + 3);

	if (entries <= 1) {
		MemHandleUnlock(handle);
		ReleaseBookmarkList();
		RemoveBookmarkRecord();
		return true;
	}
		
	offset = *(bookmarkPtr + 4) * 256 + *(bookmarkPtr + 5);
	counter = entries - index - 1;

	// Find name string for bookmark 
	readPtr = (void *) bookmarkPtr + 3 * sizeof(UInt16);
	for (i = 0; i < index; i++)
		readPtr = (void *) readPtr + StrLen(readPtr) + 1;

	nameLength = StrLen(readPtr) + 1;
	tempOffset = readPtr - bookmarkPtr;

	readPtr = (void *) readPtr + StrLen(readPtr) + 1;
	while (counter--) {
		MemMove(tempString, readPtr, StrLen(readPtr) + 1);
		length = StrLen(tempString) + 1;
		DmWrite(bookmarkPtr, tempOffset, tempString, length);
		readPtr = (void *) readPtr + length;
		tempOffset += length;
	}

	// Reshuffle blocks with bookmark data
	readPtr = (void *) bookmarkPtr + offset;
	for (i = 0; i < index; i++) {
		MemMove(&bookmarkData, readPtr, sizeof(BookmarkData));
		DmWrite(bookmarkPtr, tempOffset, &bookmarkData, sizeof(BookmarkData));
		readPtr = (void *) readPtr + sizeof(BookmarkData);
		tempOffset += sizeof(BookmarkData);
	}
	readPtr = (void *) readPtr + sizeof(BookmarkData);
	for (i = index + 1; i < entries; i++) {
		MemMove(&bookmarkData, readPtr, sizeof(BookmarkData));
		DmWrite(bookmarkPtr, tempOffset, &bookmarkData, sizeof(BookmarkData));
		readPtr = (void *) readPtr + sizeof(BookmarkData);
		tempOffset += sizeof(BookmarkData);
	}

	entries--;
	DmWrite(bookmarkPtr, sizeof(Header), (void *) &entries, sizeof(UInt16));
	offset -= nameLength;
	DmWrite(bookmarkPtr, sizeof(Header) + sizeof(UInt16), (void *) &offset, sizeof(UInt16));

	endOfRecord = MemHandleSize(handle);
	newSize = endOfRecord - sizeof(BookmarkData) - nameLength;

	MemHandleUnlock(handle);

	ResizeRecord(INTERNAL_BOOKMARKS_ID, newSize, &handle);

	return false;
}
// vim: ts=4
