/*
* $Id: categorydata.c,v 1.6 2000/10/22 20:17:32 nordstrom Exp $
*
* Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
* Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
* 
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "categorydata.h"
#include "const.h"
#include "db.h"
#include "debug.h"
#include "documentdata.h"
#include "resourceids.h"
#include "util.h"

static Err PlkrAppInfoInit(DmOpenRef dbRef) __attribute__ ((section ("plkr")));
static void UpdateCategoryData(UInt16 dst, UInt16 src, Boolean merge) __attribute__ ((section ("plkr")));
static Int16 FindRecord(Char *name, MemHandle *handle, Int16 *index, UInt16 numRecords) __attribute__ ((section ("plkr")));
static Err PlkrDBListInit(void) __attribute__ ((section ("plkr")));

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
static Char PlkrDBName[] = "PlkrDB";

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static DmOpenRef PlkrDB;

/***********************************************************************
 *
 * FUNCTION:    PlkrAppInfoInit
 *
 * DESCRIPTION: Create an app info chunk if missing.
 *
 * PARAMETERS:  dbRef  - reference to database
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
static Err PlkrAppInfoInit(DmOpenRef dbRef)
{
	UInt16 cardNo;
	MemHandle h;
	LocalID dbID;
	LocalID appInfoID;
	PlkrAppInfoType *appInfoP;
	
	if (DmOpenDatabaseInfo(dbRef, &dbID, NULL, NULL, &cardNo, NULL))
		return dmErrInvalidParam;
	if (DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, 
			   NULL, NULL, &appInfoID, NULL, NULL, NULL))
		return dmErrInvalidParam;
	if (appInfoID == NULL) {
		h = DmNewHandle (dbRef, sizeof(PlkrAppInfoType));
		if (!h) return dmErrMemError;
		
		appInfoID = MemHandleToLocalID(h);
		DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, 
						NULL, NULL, &appInfoID, NULL, NULL, NULL);
	}
	appInfoP = MemLocalIDToLockedPtr(appInfoID, cardNo);
	DmSet(appInfoP, 0, sizeof(PlkrAppInfoType), 0);
	CategoryInitialize((AppInfoPtr) appInfoP, strCatDefault);
	MemPtrUnlock(appInfoP);

	return 0;
}

/***********************************************************************
 *
 * FUNCTION:    ReturnDBInfoHandle
 *
 * DESCRIPTION: Return a handle to DB info record.
 *
 * PARAMETERS:  index - record index
 *              hand  - upon successful return, the handle to the
 *                      record, otherwise undefined
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
Err ReturnDBInfoHandle(const UInt16 index, MemHandle *hand)
{
	if (PlkrDB == NULL)
		return 1;

	FATAL_IF(hand == NULL, "NULL argument\n");
	ErrFatalDisplayIf(hand == NULL, "ReturnDBInfoHandle: NULL argument");

	*hand = DmQueryRecord(PlkrDB, index);
	if (*hand == NULL)
		return 1;
	else
		return 0;
}

/***********************************************************************
 *
 * FUNCTION:    GetCategoryName
 *
 * DESCRIPTION: Get category name
 *
 * PARAMETERS:  index - category index
 *              name  - upon successful return, the name of
 *                      the category, otherwise NULL
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
Boolean GetCategoryName(Char *name, UInt16 index)
{
	CategoryGetName(PlkrDB, index, name);
	if (name[0] == '\0')
		return false;
	else
		return true;
}

/***********************************************************************
 *
 * FUNCTION:    CategoryExists
 *
 * DESCRIPTION: Check if category already exists
 *
 * PARAMETERS:  index - category index
 *
 * RETURN:      True if category alrady exists
 *
 ***********************************************************************/
Boolean CategoryExists(UInt16 index)
{
	Char name[dmRecNumCategories];

	return GetCategoryName(name, index);
}

/***********************************************************************
 *
 * FUNCTION:    UpdateCategoryData
 *
 * DESCRIPTION: Update the category data for the records
 *
 * PARAMETERS:  dst   - destination category
 *              src   - source category
 *              merge - true if the given categories should be merged
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void UpdateCategoryData(UInt16 dst, UInt16 src, Boolean merge)
{
	MemHandle handle;
	DatabaseInfo *handlePtr;
	UInt16 i, category, categories;

	category = 1 << src;

	for (i = 0; i < DmNumRecords(PlkrDB); i++) {
		handle = DmQueryRecord(PlkrDB, i);
		if (handle == NULL) {
			break;
		}
		handlePtr = (DatabaseInfo *) MemHandleLock(handle);
		FATAL_IF(handlePtr == NULL, "MemHandleLock failed\n");
		ErrFatalDisplayIf(handlePtr == NULL, "UpdateCategoryData: MemHandleLock failed");
		categories = handlePtr->categories;
		if (categories & category) {
			categories &= ~category;
			if (merge)
				categories |= (1 << dst);
			else if (categories == 0)
				categories = 1;
			DmWrite(handlePtr, OFFSETOF(DatabaseInfo, categories), &categories, sizeof(UInt16));
		}
		MemHandleUnlock(handle);
	}
	CategorySetName(PlkrDB, src, NULL);
}

/***********************************************************************
 *
 * FUNCTION:    RemoveCategoryName
 *
 * DESCRIPTION: Remove category
 *
 * PARAMETERS:  index - category index
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void RemoveCategoryName(UInt16 index)
{
	UpdateCategoryData(0, index, false);
}

/***********************************************************************
 *
 * FUNCTION:    MergeCategories
 *
 * DESCRIPTION: Merge categories
 *
 * PARAMETERS:  dst - destination category
 *              src - source category
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void MergeCategories(UInt16 dst, UInt16 src)
{
	UpdateCategoryData(dst, src, true);
}

/***********************************************************************
 *
 * FUNCTION:    SetCategoryName
 *
 * DESCRIPTION: Set category name
 *
 * PARAMETERS:  index - category index
 *              name  - category name
 *
 * RETURN:      -1 if successful, otherwise the index for the already
 *              existing category
 *
 ***********************************************************************/
Int16 SetCategoryName(UInt16 index, Char *name)
{
	UInt8 i;
	Char tempName[dmCategoryLength];

	for (i = 0; i < dmRecNumCategories; i++) {
		MemSet((void *) tempName, dmCategoryLength, 0);
		CategoryGetName(PlkrDB, i, tempName);
		if (StrCompare(tempName, name) == 0)
			return i;
	}
	CategorySetName(PlkrDB, index, name);
	return -1;
}

/***********************************************************************
 *
 * FUNCTION:    RemoveDBRecord
 *
 * DESCRIPTION: Remove DB info record
 *
 * PARAMETERS:  index - record index
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
Err RemoveDBRecord(UInt16 index)
{
	return DmRemoveRecord(PlkrDB, index);
}

/***********************************************************************
 *
 * FUNCTION:    CompareNames
 *
 * DESCRIPTION: 
 *
 * PARAMETERS: rec1         - Pointer to the record to sort. 
 *             rec2         - Pointer to the record to sort. 
 *             other        - Any other custom information you want 
 *                            passed to the comparison function.
 *             rec1SortInfo - Pointer to SortRecordInfoType structure 
 *                            that specify unique sorting information 
 *                            for the record.
 *             rec2SortInfo - Pointer to SortRecordInfoType structure 
 *                            that specify unique sorting information 
 *                            for the record.
 *             appInfoH     - A handle to the database's application 
 *                            info block.
 *
 * RETURN:     0 if rec1 = rec2.  
 *           < 0 if rec1 < rec2.  
 *           > 0 if rec1 > rec2.  
 *
 ***********************************************************************/
Int16 CompareNames(void* rec1, void* rec2, Int16 other, SortRecordInfoPtr rec1SortInfo, 
	SortRecordInfoPtr rec2SortInfo, MemHandle appInfoH) 
{
	Int16 result;

	SET_A4_FROM_A5

	result = StrCompare(((DatabaseInfo *)rec1)->name, ((DatabaseInfo *)rec2)->name);

	RESTORE_A4

	return result;
}

/***********************************************************************
 *
 * FUNCTION:    FindRecord
 *
 * DESCRIPTION: Find record in database
 *
 * PARAMETERS:  item       - UID to search for
 *              handle     - upon successful return, the handle to the record,
 *                           otherwise undefined
 *              index      - upon successful return, the index of the record,
 *                           otherwise undefined. Pass NULL for this parameter 
 *                           if you don't want to retrieve this value.
 *              numRecords - range of records to search
 *
 * RETURN:      0 if successful
 *
 ***********************************************************************/
static Int16 FindRecord(Char *name, MemHandle *handle, Int16 *index, UInt16 numRecords)
{
	Int16 l = 0;
	Int16 r = numRecords - 1;
	Int16 x;
	Int16 result;
	DatabaseInfo *dbInfo;

	while (r >= l) {
		x = (r + l) / 2;

		*handle = DmQueryRecord(PlkrDB, x);
		if (*handle == NULL)
			break;

		dbInfo = (DatabaseInfo *) MemHandleLock(*handle);
		FATAL_IF(dbInfo == NULL, "MemHandleLock failed\n");
		ErrFatalDisplayIf(dbInfo == NULL, "FindRecord: MemHandleLock failed");

		result = StrCompare(name, dbInfo->name);
		if (result == 0)  {
			MemHandleUnlock(*handle);
			if (index)
				*index = x;
			return 0;
		}
		if (result > 0) 
			l = x + 1;
		else 
			r = x - 1;

		MemHandleUnlock(*handle);
	}
	return 1;
}

/***********************************************************************
 *
 * FUNCTION:    PlkrDBListInit
 *
 * DESCRIPTION: Populate meta database with databases. This function
 *              will find all Plucker databases and check if they
 *              already exist in the DB info database, if not it will
 *              add a new entry. If a database already exist, but the
 *              database is newer than the stored one, the info will
 *              be updated in the DB info database.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Returns 0 is successful, otherwise the error code 
 *              indicating the reason to the failure.
 *
 ***********************************************************************/
static Err PlkrDBListInit(void)
{
	MemHandle handle, dbHandle;
	DatabaseInfo *handlePtr = NULL;
	DatabaseInfo dbInfo;
	SysDBListItemType *dbIDs;
	Err err;
	UInt16 i, dbCount, index;
	UInt8 *deleteList = NULL;
	UInt16 numRecords = DmNumRecords(PlkrDB);

	if (SysCreateDataBaseList((UInt32) ViewerDBType, (UInt32) ViewerAppID, &dbCount, &dbHandle, true)) {
		dbIDs = (SysDBListItemType *) MemHandleLock(dbHandle);
		FATAL_IF(dbInfo == NULL, "MemHandleLock failed\n");
		ErrFatalDisplayIf(dbInfo == NULL, "PlkrDBListInit: MemHandleLock failed");
	} else {
		return 1;
	}
	if (numRecords != 0) {
		deleteList = (UInt8 *) MemPtrNew(numRecords * sizeof(UInt8));
		MemSet((void *) deleteList, numRecords, 1);
	}
	for (i = 0; i < dbCount; i++) {
		DmDatabaseInfo(dbIDs[i].cardNo, dbIDs[i].dbID, dbInfo.name,
				&dbInfo.attributes, NULL, &dbInfo.created, 
				NULL, NULL, NULL, NULL, NULL, NULL, NULL);

		err = FindRecord(dbInfo.name, &handle, &index, numRecords);
		if (!err) {
			deleteList[index] = 0;
			handlePtr = (DatabaseInfo *) MemHandleLock(handle);
			FATAL_IF(handlePtr == NULL, "MemHandleLock failed\n");
			ErrFatalDisplayIf(handlePtr == NULL, "PlkrDBListInit: MemHandleLock failed");

			if (handlePtr->created < dbInfo.created) {
				DmDatabaseSize(dbIDs[i].cardNo, dbIDs[i].dbID, NULL, &dbInfo.size, NULL);
				dbInfo.categories = handlePtr->categories;
				dbInfo.cardNo = dbIDs[i].cardNo;
				DmWrite((void *) handlePtr, 0, (void *) &dbInfo, sizeof(DatabaseInfo));
			}
			MemHandleUnlock(handle);
		} else {
			UInt16 dbIndex = dmMaxRecordIndex;

			handle = DmNewRecord(PlkrDB, &dbIndex, sizeof(DatabaseInfo));
			if (handle == NULL) {
				MemHandleUnlock(dbHandle);
				if (deleteList)
					MemPtrFree(deleteList);
				return DmGetLastErr();
			}
			handlePtr = (DatabaseInfo *) MemHandleLock(handle);
			FATAL_IF(handlePtr == NULL, "MemHandleLock failed\n");
			ErrFatalDisplayIf(handlePtr == NULL, "PlkrDBListInit: MemHandleLock failed");

			DmDatabaseSize(dbIDs[i].cardNo, dbIDs[i].dbID, NULL, &dbInfo.size, NULL);
			dbInfo.categories = GetDefaultCategories(dbIDs[i].dbID, dbIDs[i].cardNo);
			dbInfo.cardNo = dbIDs[i].cardNo;
			DmWrite((void *) handlePtr, 0, (void *) &dbInfo, sizeof(DatabaseInfo));

			MemHandleUnlock(handle);
			DmReleaseRecord(PlkrDB, dbIndex, true);
		}
	}
	MemHandleUnlock(dbHandle);

	index = numRecords;
	while (index > 0) {
		index--;
		if (deleteList[index]) {
			DmRemoveRecord(PlkrDB, index);
		}
	}
	if (deleteList)
		MemPtrFree(deleteList);

	DmInsertionSort(PlkrDB, (DmComparF *)CompareNames, 0);
	return 0;
}

/***********************************************************************
 *
 * FUNCTION:    CreatePlkrDatabase
 *
 * DESCRIPTION: Create database used for meta data (categories, etc.)
 *
 * PARAMETERS:  None
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
Err CreatePlkrDatabase(void)
{
	UInt16 cardNo = 0;
	Err err = 0;

	PlkrDB = DmOpenDatabaseByTypeCreator(PlkrDBType, ViewerAppID, dmModeReadWrite);
	if (!PlkrDB) {
		err = DmCreateDatabase (cardNo, PlkrDBName, ViewerAppID, PlkrDBType, false);
		if (err) {
			MSG("Couldn't create Plkr database\n");
			FrmAlert(warnInsufficientMemory);
			return err;
		}
		PlkrDB = DmOpenDatabaseByTypeCreator(PlkrDBType, ViewerAppID, dmModeReadWrite);

		err = PlkrAppInfoInit(PlkrDB);
		if (err) {
			LocalID dbID;

			DmOpenDatabaseInfo(PlkrDB, &dbID, NULL, NULL, &cardNo, NULL);
			DmCloseDatabase(PlkrDB);
			DmDeleteDatabase(cardNo, dbID);
			MSG("Couldn't initialize Plkr database [appInfo]\n");
			FrmAlert(warnInsufficientMemory);
			return err;
		}
	}
	err = PlkrDBListInit();
	if (err) {
		MSG("Couldn't initialize Plkr database [DB info]\n");
		FrmAlert(warnInsufficientMemory);
	}
	return err;
}

/***********************************************************************
 *
 * FUNCTION:    ClosePlkrDatabase
 *
 * DESCRIPTION: Close meta database.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ClosePlkrDatabase(void)
{
	if (PlkrDB == NULL)
		return;
	CloseDatabase(PlkrDB);
	PlkrDB = NULL;
}

/***********************************************************************
 *
 * FUNCTION:    GetNumOfDatabases
 *
 * DESCRIPTION: Return number of Plucker databases
 *
 * PARAMETERS:  None
 *
 * RETURN:      Number of database
 *
 ***********************************************************************/
UInt16 GetNumOfDatabases(void)
{
	return DmNumRecords(PlkrDB);
}

/***********************************************************************
 *
 * FUNCTION:    GetCategoryIndex
 *
 * DESCRIPTION: Check if category is included in list of categories
 *
 * PARAMETERS:  category - name of category
 *
 * RETURN:      Category index if the category exists otherwise
 *              dmRecNumCategories
 *
 ***********************************************************************/
UInt8 GetCategoryIndex(Char *category)
{
	UInt8 i;
	Char tempName[dmCategoryLength];

	for (i= 0; i < dmRecNumCategories; i++) {
		MemSet((void *) tempName, dmCategoryLength, 0);
		CategoryGetName(PlkrDB, i, tempName);
		if (StrCompare(tempName, category) == 0)
			break;
	}
	return i;
}
// vim: ts=4
