/*
 * $Id: categoryform.c,v 1.6 2000/10/24 05:19:04 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "categorydata.h"
#include "dbmgrform.h"
#include "debug.h"
#include "prefsdata.h"
#include "resourceids.h"
#include "util.h"

static void StoreCategories(UInt16 index, UInt16 categories) __attribute__ ((section ("plkr")));

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define CATEGORY_DELETE 0
#define CATEGORY_RENAME 1

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static UInt8 category = 0;
static UInt16 categories = 0;
static Boolean savedPrefsData = false;
static Boolean origMultipleSelect = false;
static UInt16 origCategories = 0;
static Boolean newCategory = false;
static FilterType origFilterMode = FILTER_OR;
static FieldType *fldPtr;

/***********************************************************************
 *
 * FUNCTION:    StoreCategories
 *
 * DESCRIPTION: Store category info for database
 *
 * PARAMETERS:  index      - index for database that should be updated
 *              categories - category data
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void StoreCategories(UInt16 index, UInt16 categories)
{
	MemHandle handle;
	DatabaseInfo *recordPtr;

	if (ReturnDBInfoHandle(index, &handle) == 0) {
		recordPtr = (DatabaseInfo *) MemHandleLock(handle);
		FATAL_IF(recordPtr == NULL, "MemHandleLock failed\n");
		ErrFatalDisplayIf(recordPtr == NULL, "StoreCategories: MemHandleLock failed");
		DmWrite(recordPtr, OFFSETOF(DatabaseInfo, categories), &categories, sizeof(UInt16));
		DBInfo(index)->categories = categories;
		MemHandleUnlock(handle);
	}
}

/***********************************************************************
 *
 * FUNCTION:    UpdateListCategoryData
 *
 * DESCRIPTION: Update the category data for the DB info list
 *
 * PARAMETERS:  dst   - destination category
 *              src   - source category
 *              merge - true if the given categories should be merged
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void UpdateListCategoryData(UInt16 dst, UInt16 src, Boolean merge)
{
	UInt16 i, category;

	category = 1 << src;

	for (i = 0; i < GetNumOfDatabases(); i++)  {
		if (DBInfo(i)->categories & category) {
			DBInfo(i)->categories &= ~category;
			if (merge)
				DBInfo(i)->categories |= (1 << dst);
			else if (DBInfo(i)->categories == 0)
				DBInfo(i)->categories = 1;
		}
	}
}

/***********************************************************************
 *
 * FUNCTION:    RemoveListCategoryName
 *
 * DESCRIPTION: Remove category
 *
 * PARAMETERS:  index - category index
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void RemoveListCategoryName(UInt16 index)
{
	UpdateListCategoryData(0, index, false);
}

/***********************************************************************
 *
 * FUNCTION:    MergeListCategories
 *
 * DESCRIPTION: Merge categories
 *
 * PARAMETERS:  dst - destination category
 *              src - source category
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void MergeListCategories(UInt16 dst, UInt16 src)
{
	UpdateListCategoryData(dst, src, true);
}

/***********************************************************************
 *
 * FUNCTION:    CategoryNameFormInit
 *
 * DESCRIPTION: Initialize the add bookmark form.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void CategoryNameFormInit(void)
{
	Char name[dmCategoryLength];
	FormType *nameForm = FrmGetFormPtr(frmNewCategory);
    
	GetCategoryName(name, category);
	fldPtr = GetObjectPtr(frmNewCatName);
	FldSetSelection(fldPtr, 0, 0);
	FldInsert(fldPtr, name, StrLen(name));
	FldSetSelection(fldPtr, 0, StrLen(name));
	FldSetDirty(fldPtr, false);

	FrmDrawForm(nameForm);
	FrmSetFocus(nameForm, FrmGetObjectIndex(nameForm, frmNewCatName));
}

/***********************************************************************
 *
 * FUNCTION:    CategoryNameFormHandleEvent
 *
 * DESCRIPTION: Event handler for the category name form.
 *
 * PARAMETERS:  event - pointer to an EventType structure
 *
 * RETURN:      True if the event was handled
 *
 ***********************************************************************/
Boolean CategoryNameFormHandleEvent(const EventType *event)
{
	Boolean handled = false;

	SET_A4_FROM_A5

	switch (event->eType) {
	case frmOpenEvent:
		CategoryNameFormInit();
		handled = true;
		break;

	case frmCloseEvent:
		handled = false;
		break;

	case ctlSelectEvent:
		if (event->data.ctlEnter.controlID == frmNewCatOK) {
			if (FldDirty(fldPtr)) {
				Char *name = FldGetTextPtr(fldPtr);
				if (*name) {
					Int16 dst;

					dst = SetCategoryName(category, name);
					if (dst != -1) {
						if (newCategory) {
							FrmCustomAlert(errCategoryExists, name, NULL, NULL);
							FldSetSelection(fldPtr, 0, StrLen(name));
							break;
						} else {
							if (FrmCustomAlert(confirmMergeCategory, name, NULL, NULL) == 0) {
								MergeCategories(dst, category);
								MergeListCategories(dst, category);
							}
						}
					} else {
						categories |= 1 << category;
					}
				}
			}
		} else if (event->data.ctlEnter.controlID != frmNewCatCancel) 
			break;

		FrmReturnToForm(frmCategory);
		if (event->data.ctlEnter.controlID != frmNewCatCancel)
			FrmUpdateForm(frmCategory, frmUpdateList);
		handled = true;
		break;

	default:
		handled = false;
	}

	RESTORE_A4

	return handled;
}

/***********************************************************************
 *
 * FUNCTION:    CategoryFormInit
 *
 * DESCRIPTION: Initialize the category form.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void CategoryFormInit(void)
{
	FormType *categoryForm = FrmGetFormPtr(frmCategory);
	UInt8 i;
	Char name[dmCategoryLength];
	Char *labelPtr;
    
	FrmDrawForm(categoryForm);

	if (!savedPrefsData) {
		if (SelectCategoryMode()) {
			CtlSetValue(GetObjectPtr(frmCatMultiple), Prefs()->multipleSelect);
			CtlSetValue(GetObjectPtr(frmCatOR), Prefs()->filterMode == FILTER_OR);
			CtlSetValue(GetObjectPtr(frmCatAND), Prefs()->filterMode == FILTER_AND);
			categories = Prefs()->categories;
			FrmShowObject(categoryForm, FrmGetObjectIndex(categoryForm, frmCatMultiple));
			FrmShowObject(categoryForm, FrmGetObjectIndex(categoryForm, frmCatOR));
			FrmShowObject(categoryForm, FrmGetObjectIndex(categoryForm, frmCatAND));
		} else {
			MemHandle handle;
			DatabaseInfo *recordPtr;
			// Maybe an error check?
			ReturnDBInfoHandle(ReturnLastIndex(), &handle);
			recordPtr = (DatabaseInfo *) MemHandleLock(handle);
			FATAL_IF(recordPtr == NULL, "MemHandleLock failed\n");
			ErrFatalDisplayIf(recordPtr == NULL, "CategoryFormInit: MemHandleLock failed");
			categories = recordPtr->categories;
			MemHandleUnlock(handle);

			FrmHideObject(categoryForm, FrmGetObjectIndex(categoryForm, frmCatMultiple));
			FrmHideObject(categoryForm, FrmGetObjectIndex(categoryForm, frmCatOR));
			FrmHideObject(categoryForm, FrmGetObjectIndex(categoryForm, frmCatAND));
		}
	}
	CtlSetValue(GetObjectPtr(frmCat1), categories & 0x01);
	for (i = 1; i <= dmRecNumCategories - 1; i++) {
		labelPtr = (Char *) CtlGetLabel(GetObjectPtr(frmCat1 + i));
		FrmHideObject(categoryForm, FrmGetObjectIndex(categoryForm, frmCat1 + i));
		if (GetCategoryName(name, i)) {
			StrCopy(labelPtr, name);
			FrmShowObject(categoryForm, FrmGetObjectIndex(categoryForm, frmCatOpen1 + i));
		} else {
			StrCopy(labelPtr, "");
			FrmHideObject(categoryForm, FrmGetObjectIndex(categoryForm, frmCatOpen1 + i));
			categories &= ~(1 << i);
		}
		CtlSetValue(GetObjectPtr(frmCat1 + i), categories & (1 << i));
		FrmShowObject(categoryForm, FrmGetObjectIndex(categoryForm, frmCat1 + i));
	}
	if (SelectCategoryMode())
		Prefs()->categories = categories;
	else 
		StoreCategories(ReturnLastIndex(), categories);

	if (!savedPrefsData) {
		origMultipleSelect = Prefs()->multipleSelect;
		origFilterMode = Prefs()->filterMode;
		origCategories = categories;
		savedPrefsData = true;
	}
}

/***********************************************************************
 *
 * FUNCTION:    CategoryFormHandleEvent
 *
 * DESCRIPTION: Event handler for the category form.
 *
 * PARAMETERS:  event - pointer to an EventType structure
 *
 * RETURN:      True if the event was handled
 *
 ***********************************************************************/
Boolean CategoryFormHandleEvent(const EventType *event)
{
	Boolean handled = false;
	UInt8 i;
	Char name[dmCategoryLength];

	SET_A4_FROM_A5

	switch (event->eType) {
	case ctlSelectEvent:
		switch (event->data.ctlEnter.controlID) {
		case frmCatOK:
			savedPrefsData = false;
			if (SelectCategoryMode())
				Prefs()->categories = categories;
			else if (categories == 0) {
				FrmAlert(errNoCategory);
				handled = true;
				break;
			} else
				StoreCategories(ReturnLastIndex(), categories);

			FrmGotoForm(frmDBMgr);
			FrmUpdateForm(frmDBMgr, frmUpdateTable);
			handled = true;
			break;

		case frmCatCancel:
			savedPrefsData = false;
			Prefs()->multipleSelect = origMultipleSelect;
			Prefs()->filterMode = origFilterMode;
			if (SelectCategoryMode())
				Prefs()->categories = origCategories;
			else 
				StoreCategories(ReturnLastIndex(), origCategories);

			FrmGotoForm(frmDBMgr);
			handled = true;
			break;

		case frmCatAll:
			if (SelectCategoryMode()) {
				Prefs()->multipleSelect = true;
				CtlSetValue(GetObjectPtr(frmCatMultiple), Prefs()->multipleSelect);
			}
			categories = 1;
			CtlSetValue(GetObjectPtr(frmCat1), true);
			for (i = 1; i <= dmRecNumCategories - 1; i++)
				if (GetCategoryName(name, i)) {
					CtlSetValue(GetObjectPtr(frmCat1 + i), true);
					categories |= (1 << i);
				}
			handled = true;
			break;

		case frmCatNone:
			CtlSetValue(GetObjectPtr(frmCat1), false);
			for (i = 1; i <= dmRecNumCategories - 1; i++)
				if (GetCategoryName(name, i))
					CtlSetValue(GetObjectPtr(frmCat1 + i), false);

			categories = 0;
			handled = true;
			break;

		case frmCatMultiple:
			Prefs()->multipleSelect = CtlGetValue(GetObjectPtr(frmCatMultiple));
			if (!Prefs()->multipleSelect) {
				CtlSetValue(GetObjectPtr(frmCat1), true);
				for (i = 1; i <= dmRecNumCategories - 1; i++)
					if (GetCategoryName(name, i))
						CtlSetValue(GetObjectPtr(frmCat1 + i), false);
				categories = 1;
			}
			break;

		case frmCatOR:
			Prefs()->filterMode = FILTER_OR;
			break;

		case frmCatAND:
			Prefs()->filterMode = FILTER_AND;
			break;

		case frmCat1: case frmCat2: case frmCat3: case frmCat4: 
		case frmCat5: case frmCat6: case frmCat7: case frmCat8: 
		case frmCat9: case frmCat10: case frmCat11: case frmCat12:
		case frmCat13: case frmCat14: case frmCat15: case frmCat16:
			if (SelectCategoryMode() && !Prefs()->multipleSelect) {
				CtlSetValue(GetObjectPtr(frmCat1), false);
				for (i = 1; i <= dmRecNumCategories - 1; i++)
					if (GetCategoryName(name, i))
						CtlSetValue(GetObjectPtr(frmCat1 + i), false);
				CtlSetValue(GetObjectPtr(event->data.ctlEnter.controlID), true);
			}
			category = event->data.ctlEnter.controlID - frmCat1;
			if (CategoryExists(category)) {
				if (!SelectCategoryMode() || Prefs()->multipleSelect) {
					categories ^= (1 << category);
					if (categories == 0)
						categories = (1 << category);
				} else
					categories = (1 << category);
				CtlSetValue(GetObjectPtr(event->data.ctlEnter.controlID), categories & (1 << category));
				break;
			}
			CtlSetValue(GetObjectPtr(event->data.ctlEnter.controlID), false);
			if (SelectCategoryMode() && !Prefs()->multipleSelect)
				categories = 0;
			newCategory = true;
			FrmPopupForm(frmNewCategory);
			handled = true;
			break;

		case frmCatOpen1: case frmCatOpen2: case frmCatOpen3: case frmCatOpen4: 
		case frmCatOpen5: case frmCatOpen6: case frmCatOpen7: case frmCatOpen8: 
		case frmCatOpen9: case frmCatOpen10: case frmCatOpen11: case frmCatOpen12: 
		case frmCatOpen13: case frmCatOpen14: case frmCatOpen15: case frmCatOpen16:
		{
			ListType *list = GetObjectPtr(frmCatList);
			UInt16 selection;

			LstSetPosition(list, event->screenX, event->screenY);
			selection = LstPopupList(list);

			if (selection == noListSelection)
				break;

			if (selection == CATEGORY_DELETE) {
				category = event->data.ctlEnter.controlID - frmCatOpen1;
				RemoveCategoryName(category);
				RemoveListCategoryName(category);
				categories &= ~(1 << category);
				// if no other category is selected set it to Unfiled
				if (!SelectCategoryMode() && categories == 0) {
					categories = 1;
					origCategories = 1;
				}
				FrmEraseForm(FrmGetFormPtr(frmCategory));
				CategoryFormInit();
			} else if (selection == CATEGORY_RENAME) {
				category = event->data.ctlEnter.controlID - frmCatOpen1;
				newCategory = false;
				FrmPopupForm(frmNewCategory);
			}
			handled = true;
			break;
		}

		default:	
			break;
		}
		break;

	case frmCloseEvent:
		handled = false;
		break;

	case frmOpenEvent:
		CategoryFormInit();
		handled = true;
		break;

	case frmUpdateEvent:
		if (event->data.frmUpdate.updateCode == frmUpdateList) {
			FrmEraseForm(FrmGetFormPtr(frmCategory));
			CategoryFormInit();
			handled = true;
		}
		break;

	default:
		handled = false;
	}

	RESTORE_A4

	return handled;
}

// vim: ts=4
