/*
 * $Id: control.c,v 1.27 2000/10/03 19:27:25 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "anchor.h"
#include "bookmark.h"
#include "const.h"
#include "control.h"
#include "debug.h"
#include "document.h"
#include "documentdata.h"
#include "history.h"
#include "image.h"
#include "link.h"
#include "paragraph.h"
#include "prefsdata.h"
#include "resourceids.h"
#include "search.h"

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define CONTROLCOUNT 9
#define NO_OFFSET 0
#define NOT_FOUND -1
#define ADD_BOOKMARK 0
#define EDIT_BOOKMARK 1

/***********************************************************************
 *
 *      Internal Structures
 *
 ***********************************************************************/

/* Somewhere to keep pen location information between events... */
static struct {
	Int16 penDownY;
	Int16 penDownX;
	Int16 penMoveX;
	Int16 penMoveY;
	Int16 moveFromX;
	Int16 moveFromY;
	Int16 selectedControl;
} PenData;

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static RectangleType controls[CONTROLCOUNT];
static Int16 newX;
static Int16 newY;
static Int16 imageWidth;
static Int16 imageHeight;

/***********************************************************************
 *
 * FUNCTION:    JumpToDoc
 *
 * DESCRIPTION: Jump to requested document (used by find function)
 *
 * PARAMETERS:  record - document ID
 *              offset - offset to first paragraph
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void JumpToDoc(const Int16 record, const Int16 offset)
{
	Boolean successful;
		
	successful = ViewDocument(record, true, offset);

	if (!successful) {
		ViewDocument(GetHistoryPrev(), false, NO_OFFSET);
		FrmAlert(warnInsufficientMemory);
	} else
		SetVisitedLink(record);
}

/***********************************************************************
 *
 * FUNCTION:    InitializeTbmpData
 *
 * DESCRIPTION: Initialize the data necessary for scrolling images
 *
 * PARAMETERS:  width  - width of image
 *              height - height of image
 * 
 * RETURN:      Nothing
 *
 ***********************************************************************/
void InitializeTbmpData(Int16 width, Int16 height)
{
	newX = newY = 0;
	imageWidth = width;
	imageHeight = height;
}

/***********************************************************************
 *
 * FUNCTION:    ClearPenData
 *
 * DESCRIPTION: Clear the structure holding pen data between events.
 *
 * PARAMETERS:  None
 * 
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ClearPenData(void)
{
	MemSet((void *) &PenData, sizeof(PenData), 0);
}

/***********************************************************************
 *
 * FUNCTION:    ClearControlBounds
 *
 * DESCRIPTION: Clear the structure holding control boundaries
 *
 * PARAMETERS:  None
 * 
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ClearControlBounds(void)
{
	MemSet((void *) &controls, sizeof(controls), 0);
}

/***********************************************************************
 *
 * FUNCTION:    HighlightGraphicalControl
 *
 * DESCRIPTION: Set highlight status for a graphical control object.
 *
 * PARAMETERS:  control - control ID of the object
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void HighlightGraphicalControl(const Int16 control)
{
	WinInvertRectangle(&controls[control], 0);
}

/***********************************************************************
 *
 * FUNCTION:    HighlightControl
 *
 * DESCRIPTION: Determine whether an anchor or control object should 
 *              change highlight status.
 *
 * PARAMETERS:  control - control ID of the object
 *              state   - anchor state (ANCHOR_SELECTED or
 *                                      ANCHOR_UNSELECTED)
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void HighlightControl(const Int16 control, const AnchorStateType state)
{
	if (control > 0)
		HighlightAnchor(control - 1, state);
	else if (control < 0)
		HighlightGraphicalControl(-(control + 1));
}

/***********************************************************************
 *
 * FUNCTION:    FindGraphicalControl
 *
 * DESCRIPTION: Find control object at given location.
 *
 * PARAMETERS:  x - the x coordinate
 *              y - the y coordinate
 *
 * RETURN:      Control ID, or -1 if no object was found at the given
 *              location
 *
 ***********************************************************************/
static Int16 FindGraphicalControl(const Int16 x, const Int16 y)
{
	Int16 i;

	for (i = 0; i < CONTROLCOUNT; i++)
		if (RctPtInRectangle(x, y, &(controls[i])))
			return i;

	return NOT_FOUND;
}

/***********************************************************************
 *
 * FUNCTION:    FindControlObject 
 *
 * DESCRIPTION: Find control object at pen location.
 *
 * PARAMETERS:  event - pointer to an EventType structure
 *
 * RETURN:      If the event occurred on an anchor, return value > 0
 *              (value = anchor + 1)
 *
 *              If the event occurred on a graphical control, return 
 *              value < 0 (value = -(control + 1))
 *
 *              If the event did not occur on anything, return 0
 *
 ***********************************************************************/
static Int16 FindControlObject(const EventType *event)
{
	Int16 controlObject = FindGraphicalControl(event->screenX, event->screenY);

	if (controlObject != NOT_FOUND)
		return -(controlObject + 1);

	if (event->screenY <= TopLeftY()) // We're in the header
		return 0;

	controlObject = FindVisibleAnchor(event->screenX, event->screenY);
	if (controlObject != NOT_FOUND)
		return controlObject + 1;

	return 0;
}

/***********************************************************************
 *
 * FUNCTION:    DoControlAction
 *
 * DESCRIPTION: Perform action assigned to given control object.
 *
 * PARAMETERS:  control - control value of the object
 *
 * RETURN:      Nothing
 *
 * NOTE:        If the control value > 0 it is an achor with anchor ID
 *              = control - 1
 *
 *              If the control value < 0 it is a graphical control with 
 *              control ID = -(control + 1)
 * 
 ***********************************************************************/
void DoControlAction(const Int16 control)
{
	Int16 newDocument;

	// Clear location of find pattern
	ClearFindPatternData();

	if (control > 0) {
		Int16 reference = GetVisibleReference(control - 1);
		Int16 offset = GetVisibleOffset(control - 1);
		Int16 image = GetVisibleImage(control - 1);

		if (Prefs()->linkClick)
			SndPlaySystemSound(sndClick);

		if (image) {
			RectangleType bounds = GetVisibleImagePosition(control - 1);
			ListType *list = GetObjectPtr(frmMainImageDialog);
			Int16 x, y;
			UInt16 selection;

			x = (bounds.extent.x - 60) / 2;
			if (x < 0)
				x = 0;
			y = (bounds.extent.y - 22) / 2;
			if (y < 0)
				y = 0;
			x += bounds.topLeft.x;
			y += bounds.topLeft.y;
			LstSetPosition(list, x, y);

			selection = LstPopupList(list);
			if (selection == noListSelection)
				return;
			else if (selection == 1)
				reference = image;
		}
		JumpToDoc(reference, offset);
		return;
	}

	if (control != 0)
		SndPlaySystemSound(sndClick);

	switch (-(control + 1)) {
	case HOMECONTROL:
		ViewDocument(HOME_PAGE_ID, true, NO_OFFSET);
		SetVisitedLink(HOME_PAGE_ID);
		break;

	case LEFTCONTROL:
		newDocument = GetHistoryPrev();
		if (newDocument != NOT_FOUND) {
			ViewDocument(newDocument, false, NO_OFFSET);
			SetVisitedLink(newDocument);
		}
		break;

	case RIGHTCONTROL:
		newDocument = GetHistoryNext();
		if (newDocument != NOT_FOUND) {
			ViewDocument(newDocument, false, NO_OFFSET);
			SetVisitedLink(newDocument);
		}
		break;

	case DBCONTROL:
		FrmGotoForm(frmDBMgr);
		break;

	case FINDCONTROL:
		FrmPopupForm(frmSearch);
		break;

	case AGAINCONTROL:
		SearchAgain();
		break;

	case MENUCONTROL:
		EvtEnqueueKey (menuChr, 0, 0);
		break;

	case OFFSETCONTROL:
	{
		Int16 selection;

		selection = LstPopupList(GetObjectPtr(frmMainPercentList));

		if (selection != noListSelection) {
			DynamicDocument *dynamic;
			Int16 adjustment;

			dynamic = (DynamicDocument *) MemHandleLock(GetDynamicDocument());
			FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
			ErrFatalDisplayIf(dynamic == NULL, "MemHandleLock failed");

			adjustment = -(dynamic->verticalOffset - ExtentY() + selection * (dynamic->height / 10 + 1));

			MemHandleUnlock(GetDynamicDocument());

			DoPageMove(adjustment);
		}
		break;
	}

	case BOOKMARKCONTROL:
	{
		ListType *list = GetObjectPtr(frmMainBookmarkList);
		Int16 selection;

		if (CreateLongBookmarkList(list) == 0)
			break;

		selection = LstPopupList(GetObjectPtr(frmMainBookmarkList));

		ReleaseBookmarkList();

		if (selection == ADD_BOOKMARK) {
			FrmPopupForm(frmAddBookmark);
		} else if (selection == EDIT_BOOKMARK) {
			FrmGotoForm(frmEditBookmarks);
		} else if (selection != noListSelection) {
			newDocument = RestoreBookmarkData(selection);
			if (newDocument != NOT_FOUND) {
				ViewDocument(newDocument, false, NO_OFFSET);
				SetVisitedLink(newDocument);
			}
		}
		break;
	}

	default:
		break;
	}
}

/***********************************************************************
 *
 * FUNCTION:    DoScreenAction
 *
 * DESCRIPTION: Perform action for specified screen region
 *
 * PARAMETERS:  region - screen region
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void DoScreenAction(RegionType region)
{
	switch (Prefs()->select[Prefs()->controlMode][region]) {
	case SELECT_NONE:
		break;

	case SELECT_FULL_PAGE_UP:
		DoPageMove(SCROLL_ONE_PAGE);
		break;

	case SELECT_HALF_PAGE_UP:
		DoPageMove(SCROLL_ONE_PAGE / 2);
		break;

	case SELECT_FULL_PAGE_DOWN:
		DoPageMove(-SCROLL_ONE_PAGE);
		break;

	case SELECT_HALF_PAGE_DOWN:
		DoPageMove(-SCROLL_ONE_PAGE / 2);
		break;

	case SELECT_GO_BACK:
		DoControlAction(-(LEFTCONTROL + 1));
		break;

	case SELECT_GO_FORWARD:
		DoControlAction(-(RIGHTCONTROL + 1));
		break;

	case SELECT_GO_HOME:
		DoControlAction(-(HOMECONTROL + 1));
		break;

	default:
		break;
	}
}

/***********************************************************************
 *
 * FUNCTION:    HandlePenDown
 *
 * DESCRIPTION: Respond to pen tap.
 *
 * PARAMETERS:  event - pointer to an EventType structure
 *
 * RETURN:      Nothing
 *
 * NOTE:        The screen is divided into the following regions:
 *
 *               Mode 1                 Mode 2
 *
 *               -----------------      --------+--------
 *              |    [1] 25%      |    |  [1]   |   [2]  |
 *              +-----------------+    |  25%   |   25%  |
 *              |    [2] 25%      |    |        |        |
 *              +-----------------+    +--------+--------+
 *              |    [3] 25%      |    |  [3]   |   [4]  |
 *              +-----------------+    |  25%   |   25%  |
 *              |    [4] 25%      |    |        |        |
 *               -----------------      --------+--------
 *
 ***********************************************************************/
void HandlePenDown(const EventType *event)
{
	PenData.penDownX = PenData.moveFromX = event->screenX;
	PenData.penDownY = PenData.moveFromY = event->screenY;

	if (LargeImage() && event->screenY > TopLeftY() && event->screenY < TopLeftY() + ExtentY())
		return;

	PenData.selectedControl = FindControlObject(event);

	if (PenData.selectedControl) {
		HighlightControl(PenData.selectedControl, ANCHOR_SELECTED);
	} else if ((Prefs()->controlMode != MODE3) && event->screenY > TopLeftY() && event->screenY < TopLeftY() + ExtentY()) {
		RegionType region;

		if (Prefs()->controlMode == MODE1) {
			if (PenData.penDownY <= (TopLeftY() + ExtentY() / 4))
				region = REGION1;
			else if (PenData.penDownY <= (TopLeftY() + ExtentY() / 2))
				region = REGION2;
			else if (PenData.penDownY <= (TopLeftY() + 3 * ExtentY() / 4))
				region = REGION3;
			else
				region = REGION4;
		} else {
			if (PenData.penDownY <= (TopLeftY() + ExtentY() / 2)) {
				if (PenData.penDownX <= (TopLeftX() + EXTENT_X / 2))
					region = REGION1;
				else
					region = REGION2;
			} else {
				if (PenData.penDownX <= (TopLeftX() + EXTENT_X / 2))
					region = REGION3;
				else
					region = REGION4;
			}
		}
		DoScreenAction(region);
	}
}

/***********************************************************************
 *
 * FUNCTION:    HandlePenMove
 *
 * DESCRIPTION: Respond to pen movement.
 *
 * PARAMETERS:  event - pointer to an EventType structure
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void HandlePenMove(const EventType *event)
{
	if (LargeImage() && PenData.moveFromX >= 0) {
		if (event->screenY > TopLeftY() && event->screenY < TopLeftY() + ExtentY()) {
			PenData.penMoveX = event->screenX;
			PenData.penMoveY = event->screenY;

			newX += 2 * (PenData.moveFromX - PenData.penMoveX);
			newY += 2 * (PenData.moveFromY - PenData.penMoveY);

			if (newX >= (imageWidth - MAX_EXTENT_X))
				newX = imageWidth - MAX_EXTENT_X;
			if (newX < 0)
				newX = 0;
				
			if (newY >= (imageHeight - ExtentY()))
				newY = imageHeight - ExtentY();
			if (newY < 0)
				newY = 0;
				
			PenData.moveFromX = PenData.penMoveX;
			PenData.moveFromY = PenData.penMoveY;

			DoImageMove(newX, newY);
		}
		return;
	}
	if (PenData.selectedControl >= 0) {
		PenData.penMoveY = event->screenY;
		DoPageMove(PenData.penMoveY - PenData.moveFromY);
		PenData.moveFromY = PenData.penMoveY;
		PenData.selectedControl = FindControlObject(event);
	}
	if (PenData.selectedControl > 0) {
		Int16 mouseDelta = PenData.penDownY - PenData.penMoveY;
      
		if ((mouseDelta > 3) || (mouseDelta < -3)) {
			HighlightControl(PenData.selectedControl, ANCHOR_UNSELECTED);
			PenData.selectedControl = 0;
		}
	}
}

/***********************************************************************
 *
 * FUNCTION:    HandlePenUp
 *
 * DESCRIPTION: Respond to pen release.
 *
 * PARAMETERS:  event - pointer to an EventType structure
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void HandlePenUp(const EventType *event)
{
	if (PenData.selectedControl != 0)
		HighlightControl(PenData.selectedControl, ANCHOR_UNSELECTED);

	if (PenData.selectedControl == FindControlObject(event))
		DoControlAction(PenData.selectedControl);

	ClearPenData();
	PenData.moveFromX = -1;
	PenData.moveFromY = -1;
}

/***********************************************************************
 *
 * FUNCTION:    GetControlBounds
 *
 * DESCRIPTION: Retrieve the bounds of an object.
 *
 * PARAMETERS:  form - pointer to memory block that contains the form
 *              objectId - ID of an object in the form
 *              control - control ID of the object
 *
 * RETURN:      Nothing (the object's bounds are stored in an internal 
 *              array, controls[])
 *
 * NOTE:        The objects in the form MUST have an ID equal to 1000 
 *              + control ID (they are defined in resourceids.h)
 *
 ***********************************************************************/
void GetControlBounds(const FormType *form, const UInt16 objectId, const Int16 control)
{
	FrmGetObjectBounds(form, FrmGetObjectIndex(form, objectId), &controls[control]);
}

// vim: ts=4
