/*
 * $Id: dbmgrform.c,v 1.29 2000/10/29 16:07:10 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "const.h"
#include "dbmgrform.h"
#include "debug.h"
#include "documentdata.h"
#include "emailform.h"
#include "image.h"
#include "mainform.h"
#include "prefsdata.h"
#include "resourceids.h"
#include "session.h"
#include "util.h"

static void CreateTable(void) __attribute__ ((section ("plkr")));

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define NOT_FOUND -1
#define REMOVE_COLUMN 0
#define ONE_ROW 1

#define DB_OPEN 0
#define DB_CATEGORIZE 1
#define DB_DELETE 2
#define DB_RENAME 3

#define VISIBLE_ROWS 11
#define MAX_COLUMNS 4

#define SELECTDB_COLUMN 0
#define DBNAME_COLUMN 1
#define DATE_COLUMN 2
#define SIZE_COLUMN 3

#define TABLE_WIDTH 153
#define SELECTDB_WIDTH 9
#define DBNAME_WIDTH 73
#define DATE_WIDTH 40
#define SIZE_WIDTH 26

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static DatabaseInfo *dbInfo = NULL;
static MemHandle dbInfoHandle = NULL;
static UInt16 *indexList = NULL;

static UInt16 entries = 0;
static UInt16 numberOfRows = 0;
static Int16 selectedIndex = -1;
static UInt16 firstVisibleRow = 0;

static Int16 startY = 0;
static RectangleType lastBounds;
static TableType *table;
 
static FieldType *fldPtr;
static Int16 lastIndex = 0;
static Boolean selectCategories = false;

static MemHandle docBitmapHandle = NULL;
static BitmapType *docBitmapPtr = NULL;

static Boolean newSession = true;

/***********************************************************************
 *
 * FUNCTION:    DBInfo
 *
 * DESCRIPTION: Return pointer to dbinfo structure
 *
 * PARAMETERS:  index - item number in list
 *
 * RETURN:      Pointer to dbinfo structure
 *
 ***********************************************************************/
DatabaseInfo *DBInfo(UInt16 index)
{
	return &dbInfo[index];
}

/***********************************************************************
 *
 * FUNCTION:    ReturnLastIndex
 *
 * DESCRIPTION: Return index for selected database
 *
 * PARAMETERS:  None
 *
 * RETURN:      Index
 *
 ***********************************************************************/
UInt16 ReturnLastIndex(void)
{
	return lastIndex;
}

/***********************************************************************
 *
 * FUNCTION:    SelectCategoryMode
 *
 * DESCRIPTION: Check if in select categories mode
 *
 * PARAMETERS:  None
 *
 * RETURN:      True if in select category mode
 *
 ***********************************************************************/
Boolean SelectCategoryMode(void)
{
	return selectCategories;
}

/***********************************************************************
 *
 * FUNCTION:    ReadOnlyDatabase
 *
 * DESCRIPTION: Check if database is read only
 *
 * PARAMETERS:  index - item number in list
 *
 * RETURN:      True if database is read only 
 *
 ***********************************************************************/
static Boolean ReadOnlyDatabase(Int16 index)
{
	return ((dbInfo[index].attributes & dmHdrAttrReadOnly) == dmHdrAttrReadOnly);
}

/***********************************************************************
 *
 * FUNCTION:    GetDatabaseID
 *
 * DESCRIPTION: Return ID for database
 *
 * PARAMETERS:  index - item number in list
 *
 * RETURN:      Database ID
 *
 ***********************************************************************/
static LocalID GetDatabaseID(Int16 index)
{
	return DmFindDatabase(dbInfo[index].cardNo, dbInfo[index].name);
}

/***********************************************************************
 *
 * FUNCTION:    GetDatabaseDate
 *
 * DESCRIPTION: Return creation date of database
 *
 * PARAMETERS:  index - item number in list
 *
 * RETURN:      Database creation date
 *
 ***********************************************************************/
static Char *GetDatabaseDate(Int16 index)
{
	static Char dateStr[dateStringLength];
	DateTimeType dateTime;

	TimSecondsToDateTime(dbInfo[index].created, &dateTime);

	DateToAscii(dateTime.month, dateTime.day, dateTime.year, 
		    PrefGetPreference(prefDateFormat), (Char *)&dateStr);

	return (Char *)&dateStr;
}

/***********************************************************************
 *
 * FUNCTION:    InitFieldValue
 *
 * DESCRIPTION: Initialize field values
 *
 * PARAMETERS:  bounds    - pointer to a RectangleType structure that 
 *                          specifies the bounds for the field.
 *              text      - pointer to text
 *              alignment - alignment type (rightAlign or leftAlign)
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void InitFieldValue(RectangleType *bounds, Char *text, UInt16 alignment)
{
	FieldType fld;
	FieldAttrType attr;

	MemSet(&fld, sizeof(FieldType), 0);
	FldSetBounds(&fld, bounds);

	MemSet(&attr, sizeof(FieldAttrType), 0);
	attr.usable = true;
	attr.singleLine = true;
	attr.justification = alignment;
	FldSetAttributes(&fld, &attr);

	FldSetTextPtr(&fld, text);
	FldDrawField(&fld);
}

/***********************************************************************
 *
 * FUNCTION:    UpdateScrollbar
 *
 * DESCRIPTION: Update scrollbar
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void UpdateScrollbar(void)
{
	UInt16 maxValue;

	if (numberOfRows < VISIBLE_ROWS)
		maxValue = 0;
	else
		maxValue = numberOfRows - VISIBLE_ROWS;

	SclSetScrollBar(GetObjectPtr(frmDBMgrScrollBar), firstVisibleRow, 
			0, maxValue, VISIBLE_ROWS);
}

/***********************************************************************
 *
 * FUNCTION:    ScrollTo
 *
 * DESCRIPTION: Scroll to given row
 *
 * PARAMETERS:  row - row to scroll to
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ScrollTo(Int16 row)
{
	Int16 maxScroll = numberOfRows - VISIBLE_ROWS; 

	firstVisibleRow = row;

	if (row < 0 || maxScroll < 0)
		firstVisibleRow = 0;
	else if (row > maxScroll)
		firstVisibleRow = maxScroll;

	TblMarkTableInvalid(table);
	TblRedrawTable(table);
	UpdateScrollbar();
}

/***********************************************************************
 *
 * FUNCTION:    ScrollUp
 *
 * DESCRIPTION: Scroll up
 *
 * PARAMETERS:  amount - number of rows to scroll
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ScrollUp(Int16 amount)
{
	ScrollTo(firstVisibleRow - amount);
}

/***********************************************************************
 *
 * FUNCTION:    ScrollDown
 *
 * DESCRIPTION: Scroll down
 *
 * PARAMETERS:  amount - number of rows to scroll
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ScrollDown(Int16 amount)
{
	ScrollTo(firstVisibleRow + amount);
}

/***********************************************************************
 *
 * FUNCTION:    DrawItem
 *
 * DESCRIPTION: Draw item in table
 *
 * PARAMETERS:  tbl    - pointer to table
 *              row    - row of the item (zero-based)
 *              column - column of the item (zero-based)
 *              bounds - pointer to a RectangleType structure that 
 *                       specifies the bounds for the table item.
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void DrawItem(void *tbl, Int16 row, Int16 column, RectangleType *bounds)
{
	SET_A4_FROM_A5

	WinEraseRectangle(bounds, 0);
	row += firstVisibleRow;

	if (row < numberOfRows) {
		FontID oldFont;
		Char text[dmDBNameLength];
		UInt16 index = indexList[row];
		UInt32 size;

		text[0] = '\0';

		switch (column) {
		case SELECTDB_COLUMN:
			bounds->extent.x = TABLE_WIDTH;
			WinEraseRectangle(bounds, 0);
			WinDrawBitmap(docBitmapPtr, bounds->topLeft.x + 1, bounds->topLeft.y + 1);
			if (selectedIndex == index) {
				bounds->extent.x = SELECTDB_WIDTH;
				WinInvertRectangle(bounds, 0);
			}
			break;

		case DBNAME_COLUMN:
			oldFont = FntSetFont(stdFont);
			StrNCopy(text, dbInfo[index].name, dmDBNameLength);
			TrimText(text, bounds->extent.x - 2);
			InitFieldValue(bounds, text, leftAlign);
			FntSetFont(oldFont);
			break;

		case DATE_COLUMN:
			oldFont = FntSetFont(stdFont);
			if (Prefs()->showDate && Prefs()->showSize) {
				StrNCopy(text, GetDatabaseDate(index), dmDBNameLength);
			}
			InitFieldValue(bounds, text, rightAlign);
			FntSetFont(oldFont);
			break;

		case SIZE_COLUMN:
			oldFont = FntSetFont(stdFont);
			if (Prefs()->showSize) {
				size = (dbInfo[index].size + 500) / 1024;
				if (size != 0)
					StrPrintF(text, "%ldk", size);
				else
					StrPrintF(text, "%ld", dbInfo[index].size);
			} else if (Prefs()->showDate) {
				StrNCopy(text, GetDatabaseDate(index), dmDBNameLength);
			}
			InitFieldValue(bounds, text, rightAlign);
			FntSetFont(oldFont);
			break;
		
		default:
			break;
		}
	}

	RESTORE_A4
}

/***********************************************************************
 *
 * FUNCTION:    CreateTable
 *
 * DESCRIPTION: Populate table with databases
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void CreateTable(void)
{
	MemHandle handle;
	DatabaseInfo *recordPtr;
	UInt16 i;

	entries = GetNumOfDatabases();

	if (entries == 0)
		return;

	// Allocate arrays for database info table
	indexList = (UInt16 *) MemPtrNew(entries * sizeof(UInt16));
	if (indexList == NULL)
		return;

	dbInfoHandle = MemHandleNew(entries * sizeof(DatabaseInfo));
	if (dbInfoHandle == NULL) {
		MemPtrFree(indexList);
		indexList = NULL;
		return;
	}
	dbInfo = MemHandleLock(dbInfoHandle);
	FATAL_IF(dbInfo == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dbInfo == NULL, "CreateTable: MemHandleLock failed");

	for (i = 0; i < entries; i++)  {
		if (ReturnDBInfoHandle(i, &handle) == 0) {
			recordPtr = (DatabaseInfo *) MemHandleLock(handle);
			FATAL_IF(recordPtr == NULL, "MemHandleLock failed\n");
			ErrFatalDisplayIf(recordPtr == NULL, "CreateTable: MemHandleLock failed");
			StrNCopy(dbInfo[i].name, recordPtr->name, dmDBNameLength);
			dbInfo[i].cardNo = recordPtr->cardNo;
			dbInfo[i].created = recordPtr->created;
			dbInfo[i].attributes = recordPtr->attributes;
			dbInfo[i].size = recordPtr->size;
			dbInfo[i].categories = recordPtr->categories;
			if (Prefs()->dbID == DmFindDatabase(dbInfo[i].cardNo, dbInfo[i].name))
				selectedIndex = i;
			MemHandleUnlock(handle);
		}
	}
}

/***********************************************************************
 *
 * FUNCTION:    ReleaseTable
 *
 * DESCRIPTION: Unlock table handles and free allocated memory
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ReleaseTable(void)
{
	if (indexList) {
		MemPtrFree(indexList);
		indexList = NULL;
	}
	if (dbInfoHandle) {
		MemHandleUnlock(dbInfoHandle);
		MemHandleFree(dbInfoHandle);
		dbInfoHandle = NULL;
	}
}

/***********************************************************************
 *
 * FUNCTION:    UpdateIndexList
 *
 * DESCRIPTION: Update list of databases to display
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void UpdateIndexList(void)
{
	UInt16 i, category;

	numberOfRows = 0;
	for (i = 0; i < entries; i++)  {
		category = dbInfo[i].categories & Prefs()->categories;
		if (((Prefs()->filterMode == FILTER_OR) && category) ||
		    ((Prefs()->filterMode == FILTER_AND) && (category == Prefs()->categories))) {
			indexList[numberOfRows] = i;
			numberOfRows++;
		}
	}
}

/***********************************************************************
 *
 * FUNCTION:    OpenNewDatabase
 *
 * DESCRIPTION: Open the database specified by the user.
 *
 * PARAMETERS:  selection - currently selected item in table
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
static Err OpenNewDatabase(UInt16 selection)
{
	Err err;
	UInt16 index = indexList[selection];
	LocalID dbID = GetDatabaseID(index);

	err = OpenDocumentDatabase(dbID, dbInfo[index].cardNo);
	if (err)
		return err;

	Prefs()->dbID = dbID;
	Prefs()->cardNo = dbInfo[index].cardNo;
	StrCopy(Prefs()->dbName, dbInfo[index].name);
	selectedIndex = index;
	InitSessionData();

	return 0;
}

/***********************************************************************
 *
 * FUNCTION:    DeleteDatabase
 *
 * DESCRIPTION: Delete the database specified by the user.
 *
 * PARAMETERS:  selection - currently selected item in table
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void DeleteDatabase(UInt16 selection)
{
	UInt16 index = indexList[selection];
	Char *label = dbInfo[index].name;
	UInt16 i;

	if (FrmCustomAlert(confirmDeleteDB, label, NULL, NULL) == 0)
	{
		if (DmDeleteDatabase(dbInfo[index].cardNo, GetDatabaseID(index)) == 0) {
			DeleteDynamicDatabase(label);
			RemoveDBRecord(index);

			if (selectedIndex == index) {
				selectedIndex = -1;
				StrCopy(Prefs()->dbName, "");
				Prefs()->dbID = NULL;
				Prefs()->cardNo = 0;
			}else if (selectedIndex > index)
					selectedIndex--;

			entries--;
			for (i = index; i < entries; i++)
				dbInfo[i] = dbInfo[i + 1];

			numberOfRows--;
			for (i = index; i < numberOfRows; i++)
				indexList[i] = indexList[i + 1];

			if (!entries) {
				EventType stop;
				RectangleType firstRow = { { 0, 15 }, { 160, 11 }};

				WinEraseRectangle(&firstRow, 0);
				FrmAlert(infoNoPluckerDB);

				MemSet(&stop, sizeof(EventType), 0);
				stop.eType = appStopEvent;
				EvtAddEventToQueue(&stop);

				return;
			}
			ScrollUp(ONE_ROW);
			UpdateIndexList();
			TblDrawTable(table);
		} else
			FrmAlert(errNoDeleteDB);
	}
}

/***********************************************************************
 *
 * FUNCTION:    FindRow
 *
 * DESCRIPTION: Find row in table given y position of pen
 *
 * PARAMETERS:  y - y coordinate of pen tap
 *
 * RETURN:      Row in table or NOT_FOUND if outside of available rows
 *
 ***********************************************************************/
static UInt16 FindRow(Int16 y)
{
	UInt16 row;
	RectangleType bounds;

	for (row = 0; row < VISIBLE_ROWS; row++) {
		TblGetItemBounds(table, row, 0, &bounds);
		if (RctPtInRectangle(bounds.topLeft.x + 1, y, &bounds))
			break;
	}

	if (row == VISIBLE_ROWS)
		return NOT_FOUND;
	else
		return row;
}

/***********************************************************************
 *
 * FUNCTION:    InvertCurrentRow
 *
 * DESCRIPTION: Invert row in table
 *
 * PARAMETERS:  row - row to invert
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void InvertRow(UInt16 row)
{
	RectangleType bounds;

	TblGetItemBounds(table, row, 0, &bounds);

	if (Prefs()->scrollbar == SCROLLBAR_LEFT)
		bounds.topLeft.x = SELECTDB_WIDTH + 7;
	else
		bounds.topLeft.x = SELECTDB_WIDTH;

	bounds.extent.x = TABLE_WIDTH - SELECTDB_WIDTH;
	WinInvertRectangle(&bounds, 0);
	MemMove(&lastBounds, &bounds, sizeof(bounds));
}

/***********************************************************************
 *
 * FUNCTION:    RenameDBFormInit
 *
 * DESCRIPTION: Initialize the rename DB form.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void RenameDBFormInit(void)
{
	FormType *nameForm = FrmGetFormPtr(frmRenameDB);
    
	fldPtr = GetObjectPtr(frmRenameDBField);
	FldSetSelection(fldPtr, 0, 0);
	FldInsert(fldPtr, dbInfo[lastIndex].name, StrLen(dbInfo[lastIndex].name));
	FldSetSelection(fldPtr, 0, StrLen(dbInfo[lastIndex].name));
	FldSetDirty(fldPtr, false);

	FrmDrawForm(nameForm);
	FrmSetFocus(nameForm, FrmGetObjectIndex(nameForm, frmRenameDBField));
}

/***********************************************************************
 *
 * FUNCTION:    RenameDBFormHandleEvent
 *
 * DESCRIPTION: Event handler for the rename DB form.
 *
 * PARAMETERS:  event - pointer to an EventType structure
 *
 * RETURN:      True if the event was handled
 *
 ***********************************************************************/
Boolean RenameDBFormHandleEvent(const EventType *event)
{
	Boolean handled = false;

	SET_A4_FROM_A5

	switch (event->eType) {
	case ctlSelectEvent:
		if (event->data.ctlEnter.controlID == frmRenameDBOK) {
			if (FldDirty(fldPtr)) {
				Char *text;

				text = FldGetTextPtr(fldPtr);
				if (*text) {
					Err err;

					err = RenameDatabase(text, dbInfo[lastIndex].name, dbInfo[lastIndex].cardNo);
					if (err) {
						FrmCustomAlert(errAlreadyExists, text, NULL, NULL);
						FldSetSelection(fldPtr, 0, StrLen(text));
						break;
					} else {
						StrNCopy(dbInfo[lastIndex].name, text, dmDBNameLength);
					}
				} else
					break;
			}
		} else if (event->data.ctlEnter.controlID != frmRenameDBCancel) 
			break;

		FrmReturnToForm(frmDBMgr);
		if (event->data.ctlEnter.controlID != frmRenameDBCancel)
			FrmUpdateForm(frmDBMgr, frmUpdateTable);
		handled = true;
		break;

	case frmCloseEvent:
		handled = false;
		break;

	case frmOpenEvent:
		RenameDBFormInit();
		handled = true;
		break;

	default:
		handled = false;
	}

	RESTORE_A4

	return handled;
}

/***********************************************************************
 *
 * FUNCTION:    UpdateColumns
 *
 * DESCRIPTION: Update the width of the columns depending on filter
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void UpdateColumns(void)
{
	UInt16 DBNameWidth;

	DBNameWidth = DBNAME_WIDTH;
	if (!Prefs()->showDate) {
		DBNameWidth += DATE_WIDTH;
		TblSetColumnWidth(table, DATE_COLUMN, REMOVE_COLUMN);
	} else {
		TblSetColumnWidth(table, DATE_COLUMN, DATE_WIDTH);
	}
	if (!Prefs()->showSize) {
		DBNameWidth += SIZE_WIDTH;
		if (Prefs()->showDate) {
			TblSetColumnWidth(table, DATE_COLUMN, REMOVE_COLUMN);
			TblSetColumnWidth(table, SIZE_COLUMN, DATE_WIDTH);
		} else {
			TblSetColumnWidth(table, SIZE_COLUMN, REMOVE_COLUMN);
		}
	} else {
		TblSetColumnWidth(table, SIZE_COLUMN, SIZE_WIDTH);
	}
	TblSetColumnWidth(table, DBNAME_COLUMN, DBNameWidth);
}

/***********************************************************************
 *
 * FUNCTION:    DBmgrFormInit
 *
 * DESCRIPTION: Initialize the database manager form.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void DBmgrFormInit(void)
{
	UInt16 row, column;
	FormType *dbMgrForm = FrmGetFormPtr(frmDBMgr);
	FormType *frm = FrmInitForm(frmInitCategory);
	RectangleType r;

	table = GetObjectPtr(frmDBMgrTable);
	for (column = 0; column < MAX_COLUMNS; column++) {
		TblSetCustomDrawProcedure(table, column, DrawItem);
		TblSetColumnUsable(table, column, true);
		for (row = 0; row < VISIBLE_ROWS; row++) {
			TblSetRowUsable(table, row, true);
			TblSetItemStyle(table, row, column, customTableItem);
		}
	}
	TblGetBounds(table, &r);
	if (Prefs()->scrollbar == SCROLLBAR_LEFT) {
		r.topLeft.x = 7;

                FrmSetObjectPosition(dbMgrForm, FrmGetObjectIndex(dbMgrForm, frmDBMgrScrollBar), 0, TOP_LEFT_Y + 1);
	} else {
		r.topLeft.x = 0;

                FrmSetObjectPosition(dbMgrForm, FrmGetObjectIndex(dbMgrForm, frmDBMgrScrollBar), TABLE_WIDTH, TOP_LEFT_Y + 1);
	}
	TblSetBounds(table, &r);

	FrmSetActiveForm(frm);
	FrmDrawForm(frm);

	DeleteOffScreenWindow();
	SaveSessionData();
	CloseDocumentDatabase();

	if (newSession) {
		Err err;

		newSession = false;
		err = CreatePlkrDatabase();
		if (err) {
			EventType stop;

			MemSet(&stop, sizeof(EventType), 0);
			stop.eType = appStopEvent;
			EvtAddEventToQueue(&stop);

			return;
		}
		CreateTable();
	}
	// Keep the doc bitmap locked since it is frequently used
	docBitmapHandle = DmGetResource(bitmapRsc, bmpDoc);
	docBitmapPtr = MemHandleLock(docBitmapHandle);

	UpdateIndexList();
	UpdateColumns();

	FrmEraseForm(frm);
	FrmDeleteForm(frm);
	FrmSetActiveForm(dbMgrForm);
	FrmDrawForm(dbMgrForm);

	UpdateScrollbar();

	CtlSetValue(GetObjectPtr(frmDBMgrDate), Prefs()->showDate);
	CtlSetValue(GetObjectPtr(frmDBMgrSize), Prefs()->showSize);

	Prefs()->lastForm = frmDBMgr;
}

/***********************************************************************
 *
 * FUNCTION:    DBmgrFormHandleEvent
 *
 * DESCRIPTION: Event handler for the database manager form.
 *
 * PARAMETERS:  event - pointer to an EventType structure
 *
 * RETURN:      True if the event was handled
 *
 ***********************************************************************/
Boolean DBmgrFormHandleEvent(const EventType *event)
{
	Boolean handled = false;
	RectangleType bounds;
	UInt16 currentRow;

	SET_A4_FROM_A5

	switch (event->eType) {
	case ctlSelectEvent:
		if (event->data.ctlEnter.controlID == frmDBMgrCategory) {
			selectCategories = true;
			FrmGotoForm(frmCategory);
		} else if (event->data.ctlEnter.controlID == frmDBMgrDate) {
			Prefs()->showDate = !Prefs()->showDate;
			UpdateColumns();
			TblDrawTable(table);
		} else if (event->data.ctlEnter.controlID == frmDBMgrSize) {
			Prefs()->showSize = !Prefs()->showSize;
			UpdateColumns();
			TblDrawTable(table);
		} else
			break;

		handled = true;
		break;

	case penDownEvent:
		TblGetBounds(table, &bounds);
		if (RctPtInRectangle(event->screenX, event->screenY, &bounds)) {
			ListType *list = GetObjectPtr(frmDBMgrList);
			UInt16 selection;

			startY = event->screenY;
			currentRow = FindRow(startY);

			if (currentRow != NOT_FOUND && currentRow < numberOfRows) {
				if ((event->screenX - bounds.topLeft.x) < 10) {
					startY = 0;
					LstSetPosition(list, event->screenX, event->screenY);
					selection = LstPopupList(list);

					if (selection != noListSelection) {
						if (selection == DB_DELETE) {
							DeleteDatabase(currentRow + firstVisibleRow);
						} else if (selection == DB_OPEN) {
							Err err;

							err = OpenNewDatabase(currentRow + firstVisibleRow);
							if (!err)
								FrmGotoForm(Prefs()->toolbar);
						} else if (selection == DB_RENAME) {
							lastIndex = indexList[currentRow + firstVisibleRow];
							if (ReadOnlyDatabase(lastIndex))
								FrmAlert(errReadOnly);
							else
								FrmPopupForm(frmRenameDB);
						} else if (selection == DB_CATEGORIZE) {
							lastIndex = indexList[currentRow + firstVisibleRow];
							selectCategories = false;
							FrmGotoForm(frmCategory);
						}
					}
				} else {
					InvertRow(currentRow);
				}
			}
			if (currentRow < VISIBLE_ROWS)
				handled = true;
		}
		break;

	case penUpEvent:
		TblGetBounds(table, &bounds);
		if (startY && RctPtInRectangle(event->screenX, event->screenY, &bounds)) {
			currentRow = FindRow(event->screenY);
			if (currentRow != NOT_FOUND && currentRow < numberOfRows) {
				Err err;

				SndPlaySystemSound(sndClick);
				WinInvertRectangle(&lastBounds, 0);
				MemSet(&lastBounds, sizeof(lastBounds), 0);

				err = OpenNewDatabase(currentRow + firstVisibleRow);
				if (!err)
					FrmGotoForm(Prefs()->toolbar);
			}
			if (currentRow < VISIBLE_ROWS)
				handled = true;
		}
		break;

	case penMoveEvent:
		if (startY) {
			TblGetBounds(table, &bounds);
			if (!RctPtInRectangle(event->screenX, event->screenY, &bounds)) {
				Int16 x, y;
				Boolean penDown;

				WinInvertRectangle(&lastBounds, 0);
				MemSet(&lastBounds, sizeof(lastBounds), 0);

				if (event->screenY < bounds.topLeft.y)
					do {
						ScrollUp(ONE_ROW);
						EvtGetPen(&x, &y, &penDown);
					} while (penDown && (y < bounds.topLeft.y));

				if (event->screenY > bounds.topLeft.y + bounds.extent.y)
					do {
						ScrollDown(ONE_ROW);
						EvtGetPen(&x, &y, &penDown);
					} while (penDown && (y > bounds.topLeft.y + bounds.extent.y));
			} else {
				currentRow = FindRow(event->screenY);

				if (currentRow >= numberOfRows) {
					WinInvertRectangle(&lastBounds, 0);
					MemSet(&lastBounds, sizeof(lastBounds), 0);
				} else if (abs(startY - event->screenY) >= 5) {
					WinInvertRectangle(&lastBounds, 0);
					InvertRow(currentRow);
				}
			}
		}
		handled = true;
		break;

	case sclRepeatEvent:
		ScrollTo(event->data.sclRepeat.newValue);
		handled = false;
		break;

	case keyDownEvent:
		switch (event->data.keyDown.chr) {
		case pageUpChr:
			ScrollUp(VISIBLE_ROWS);
			handled = true;
			break;

		case pageDownChr:
			ScrollDown(VISIBLE_ROWS);
			handled = true;
			break;
		}
		break;

	case frmOpenEvent:
		DBmgrFormInit();
		handled = true;
		break;

	case frmCloseEvent:
		if (docBitmapHandle) {
			MemHandleUnlock(docBitmapHandle);
			DmReleaseResource(docBitmapHandle);
			docBitmapHandle = NULL;
		}
		startY = 0;
		handled = false;
		break;

	case frmUpdateEvent:
		if (event->data.frmUpdate.updateCode == frmUpdateTable) {
			UpdateIndexList();
			UpdateColumns();
			// if list has "shrunk" scroll to top of list
			if (firstVisibleRow > numberOfRows) {
				ScrollTo(0);
			} else {
				TblDrawTable(table);
				UpdateScrollbar();
			}

			handled = true;
		} else if (event->data.frmUpdate.updateCode == frmUpdateDoc) {
			if (docBitmapHandle) {
				MemHandleUnlock(docBitmapHandle);
				DmReleaseResource(docBitmapHandle);
				docBitmapHandle = NULL;
			}
			FrmEraseForm(FrmGetFormPtr(frmDBMgr));
			DBmgrFormInit();
			handled = true;
		}
		break;

	case menuEvent:
		// Refresh display to get rid of command text in bottom left corner
		MenuEraseStatus(NULL);
		handled = HandleCommonMenuItems(event->data.menu.itemID);
		break;

	default:
		handled = false;
	}

	RESTORE_A4

	return handled;
}

// vim: ts=4
