/*
 * $Id: document.c,v 1.38 2000/10/22 20:17:32 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "anchor.h"
#include "const.h"
#include "control.h"
#include "debug.h"
#include "document.h"
#include "documentdata.h"
#include "emailform.h"
#include "externalform.h"
#include "history.h"
#include "image.h"
#include "link.h"
#include "mainform.h"
#include "paragraph.h"
#include "prefsdata.h"
#include "resourceids.h"
#include "screen.h"
#include "search.h"
#include "uncompress.h"
#include "util.h"

static void InitViewportBounds(void) __attribute__ ((section ("plkr")));
static void UpdateScrollbar(DynamicDocument *dynamic) __attribute__ ((section ("plkr")));
static void SetParagraphHeight(Int16 index, Int16 height) __attribute__ ((section ("plkr")));
static void SetLastVisibleParagraph(const Int16 index, const Int16 y) __attribute__ ((section ("plkr")));
static void SetFirstVisibleParagraph(const Int16 index, const Int16 y) __attribute__ ((section ("plkr")));
static void DrawDocument(Document *document) __attribute__ ((section ("plkr")));
static void ScrollDocumentUp(Document *document) __attribute__ ((section ("plkr")));
static void ScrollDocumentDown(Document *document, const Int16 height) __attribute__ ((section ("plkr")));
static void AdjustVerticalOffset(Document *document, Int16 adjustment) __attribute__ ((section ("plkr")));
static Int16 CalculateDocumentHeight(Document *document, Int16 pOffset) __attribute__ ((section ("plkr")));
static void HandleNamedAnchor(Document *document, Int16 pOffset) __attribute__ ((section ("plkr")));
static Boolean ViewHtmlDocument(Document *document, Boolean newPage, Int16 pOffset) __attribute__ ((section ("plkr")));

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define NO_SCROLLBAR 0

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static Int16 minVerticalOffset;
static MemHandle currentDocument = NULL;
static MemHandle dynamicDocument = NULL;
static RectangleType viewportBounds;

/***********************************************************************
 *
 * FUNCTION:    ViewportBounds
 *
 * DESCRIPTION: Return pointer to viewport
 *
 * PARAMETERS:  None
 *
 * RETURN:      Pointer to viewport
 *
 ***********************************************************************/
RectangleType *ViewportBounds(void)
{
	return &viewportBounds;
}

/***********************************************************************
 *
 * FUNCTION:    InitViewportBounds
 *
 * DESCRIPTION: Initialize boundaries for viewport
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void InitViewportBounds(void)
{
	viewportBounds.topLeft.x = TopLeftX();
	viewportBounds.topLeft.y = TopLeftY();
	viewportBounds.extent.x = EXTENT_X;
	viewportBounds.extent.y = ExtentY();
}

/***********************************************************************
 *
 * FUNCTION:    UpdateScrollbar
 *
 * DESCRIPTION: Update the scrollbar
 *
 * PARAMETERS:  dynamic - pointer to (dynamic) document
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void UpdateScrollbar(DynamicDocument *dynamic)
{
	Int16 maxValue;

	if (Prefs()->scrollbar == SCROLLBAR_NONE || dynamic->height < viewportBounds.extent.y)
		maxValue = NO_SCROLLBAR;
	else
		maxValue = dynamic->height - viewportBounds.extent.y;

	SclSetScrollBar(GetObjectPtr(frmMainScrollBar), -(dynamic->verticalOffset), 0, 
			maxValue, SCROLL_ONE_PAGE);

	SclDrawScrollBar(GetObjectPtr(frmMainScrollBar));
}

/***********************************************************************
 *
 * FUNCTION:    SetParagraphHeight
 *
 * DESCRIPTION: Set the height of the paragraph.
 *
 * PARAMETERS:  index  - paragraph index
 *              height - paragraph height 
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void SetParagraphHeight(Int16 index, Int16 height)
{
	DynamicDocument *dynamic = MemHandleLock(dynamicDocument);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "SetParagraphHeight: MemHandleLock failed");

	DmWrite((void *) dynamic,
		sizeof(DynamicDocument) + index * sizeof(DynamicParagraph),
		(void *) &height, 
		sizeof(Int16));

	MemHandleUnlock(dynamicDocument);
}

/***********************************************************************
 *
 * FUNCTION:    SetWorking
 *
 * DESCRIPTION: Indicate that the application is working.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void SetWorking(void)
{
	FormType *mainForm = FrmGetFormPtr(Prefs()->toolbar);

	if (Prefs()->toolbar == TOOLBAR_NONE)
		return;

	FrmHideObject(mainForm, FrmGetObjectIndex(mainForm, bmpLeft));
	FrmHideObject(mainForm, FrmGetObjectIndex(mainForm, bmpRight));
	FrmHideObject(mainForm, FrmGetObjectIndex(mainForm, bmpHome));

	FrmShowObject(mainForm, FrmGetObjectIndex(mainForm, bmpWait));
}

/***********************************************************************
 *
 * FUNCTION:    SetIdle
 *
 * DESCRIPTION: Indicate that the application is idle.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void SetIdle(void)
{
	FormType *mainForm = FrmGetFormPtr(Prefs()->toolbar);

	if (Prefs()->toolbar == TOOLBAR_NONE)
		return;

	FrmHideObject(mainForm, FrmGetObjectIndex(mainForm, bmpWait));

	FrmShowObject(mainForm, FrmGetObjectIndex(mainForm, bmpLeft));
	FrmShowObject(mainForm, FrmGetObjectIndex(mainForm, bmpRight));
	FrmShowObject(mainForm, FrmGetObjectIndex(mainForm, bmpHome));
}

/***********************************************************************
 *
 * FUNCTION:    ShowVerticalOffset
 *
 * DESCRIPTION: Show vertical offset of document.
 *
 * PARAMETERS:  document - current document
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ShowVerticalOffset(Document *document)
{
	Char offText[5];
	Int32 percent = 0;
	DynamicDocument *dynamic;

	if (Prefs()->toolbar == TOOLBAR_NONE)
		return;

	dynamic = MemHandleLock(dynamicDocument);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "ShowVerticalOffset: MemHandleLock failed");

	if (dynamic->height > 0) {
		percent = -dynamic->verticalOffset + ExtentY();
		percent *= 100;
		percent /= dynamic->height; 

		if (percent > 100)
			percent = 100;
	}
	MemHandleUnlock(dynamicDocument);

	StrPrintF(offText, "%ld%%", percent);
	CtlSetLabel(GetObjectPtr(frmMainPercentPopup), offText);
}

/***********************************************************************
 *
 * FUNCTION:    SetLastVisibleParagraph
 *
 * DESCRIPTION: Store information about the last visible paragraph.
 *
 * PARAMETERS:  index - index for paragraph in document record
 *              y     - paragraph's y coordinate within document
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void SetLastVisibleParagraph(const Int16 index, const Int16 y)
{
	DynamicDocument *dynamic = MemHandleLock(dynamicDocument);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "SetLastVisibleParagraph: MemHandleLock failed");

	DmWrite((void *) dynamic, OFFSETOF(DynamicDocument, lastVisibleParagraph), 
		(void *) &index, sizeof(Int16));

	DmWrite((void *) dynamic, OFFSETOF(DynamicDocument, lastParagraphY), 
		(void *) &y, sizeof(Int16));

	MemHandleUnlock(dynamicDocument);
	SetHistoryLastParagraph(index, y);
}

/***********************************************************************
 *
 * FUNCTION:    SetFirstVisibleParagraph
 *
 * DESCRIPTION: Store information about the first visible paragraph.
 *
 * PARAMETERS:  index - index for paragraph in document record
 *              y     - paragraph's y coordinate within document
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void SetFirstVisibleParagraph(const Int16 index, const Int16 y)
{
	DynamicDocument *dynamic = MemHandleLock(dynamicDocument);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "SetFirstVisibleParagraph: MemHandleLock failed");

	DmWrite((void *) dynamic, OFFSETOF(DynamicDocument, firstVisibleParagraph), 
		(void *) &index, sizeof(Int16));

	DmWrite((void *) dynamic, OFFSETOF(DynamicDocument, firstParagraphY), 
		(void *) &y, sizeof(Int16));

	MemHandleUnlock(dynamicDocument);
	SetHistoryFirstParagraph(index, y);
}

/***********************************************************************
 *
 * FUNCTION:    InitializeTextContext
 *
 * DESCRIPTION: Initialize text context.
 *
 * PARAMETERS:  tContext - pointer to text context
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void InitializeTextContext(TextContext *tContext)
{
	tContext->style = DEFAULTSTYLE;
	tContext->cursorX = viewportBounds.topLeft.x;
	tContext->cursorY = viewportBounds.topLeft.y;
	tContext->writeMode = true;
	tContext->activeAnchor = false;
	tContext->patternOffset = -1;
}

/***********************************************************************
 *
 * FUNCTION:    DrawDocument
 *
 * DESCRIPTION: Draw the document at most until the end of the visible 
 *              document.
 *
 * PARAMETERS:  document - pointer to document record
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void DrawDocument(Document *document)
{
	Paragraph *paragraph;
	DynamicParagraph *dynamicParagraph;
	UInt16 numOfParagraphs;
	Boolean needNewParagraph;
	Int16 firstVisiblePixel, lastVisiblePixel;
	Int16 documentCursor, lastDocumentCursor;
	Int16 index;
	TextContext thisContext;
	DynamicDocument *dynamic = MemHandleLock(dynamicDocument);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "DrawDocument: MemHandleLock failed");

	InitializeTextContext(&thisContext);

	index = dynamic->firstVisibleParagraph;
	paragraph = ((void *) document) + sizeof(Document) + index * sizeof(Paragraph);
	dynamicParagraph = ((void *) dynamic) + sizeof(DynamicDocument) + index * sizeof(DynamicParagraph);
	numOfParagraphs = document->paragraphs - index;
	firstVisiblePixel = -(dynamic->verticalOffset);
	lastVisiblePixel = viewportBounds.extent.y - dynamic->verticalOffset;
	documentCursor = dynamic->firstParagraphY;
	thisContext.cursorY = documentCursor + dynamic->verticalOffset + viewportBounds.topLeft.y;

	needNewParagraph = false;

	WinSetClip(&viewportBounds);

	while (numOfParagraphs--) {
		lastDocumentCursor = documentCursor;

		DrawParagraph(&thisContext, paragraph, document);

		documentCursor += dynamicParagraph->height;

		/*
		   Save pointers to the first and last visible paragraphs. It is
		   possible that this is the same paragraph, so be careful.

		   Also note that the very first paragraph in the document is
		   the initial "firstParagraph", so we don't need code to
		   deal with that special case.
		 */
		if (documentCursor < firstVisiblePixel)
			needNewParagraph = true;
		else if (needNewParagraph) {
			SetFirstVisibleParagraph(index, lastDocumentCursor);
			needNewParagraph = false;
		}
		if (documentCursor > lastVisiblePixel) {
			SetLastVisibleParagraph(index, lastDocumentCursor);
			break;
		}
		paragraph = ((void *) paragraph) + sizeof(Paragraph);
		dynamicParagraph = ((void *) dynamicParagraph) + sizeof(DynamicParagraph);
		index++;
	}
	MemHandleUnlock(dynamicDocument);
	WinResetClip();
}

/***********************************************************************
 *
 * FUNCTION:    ScrollDocumentUp
 *
 * DESCRIPTION: Scroll document up.
 *
 * PARAMETERS:  document - pointer to document record
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ScrollDocumentUp(Document *document)
{
	DynamicParagraph *dynamicParagraph;
	Int16 firstVisiblePixel;
	Int16 visibleParagraphPixel;
	Int16 index;
	DynamicDocument *dynamic = MemHandleLock(dynamicDocument);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "ScrollDocumentUp: MemHandleLock failed");

	firstVisiblePixel = -(dynamic->verticalOffset);
	index = dynamic->firstVisibleParagraph;

	/*
	   Work out the new "first visible" paragraph by moving down
	 */
	dynamicParagraph = ((void *) dynamic) + sizeof(DynamicDocument) + index * sizeof(DynamicParagraph);
	visibleParagraphPixel = dynamic->firstParagraphY + dynamicParagraph->height;

	while (visibleParagraphPixel < firstVisiblePixel) {
		index++;
		dynamicParagraph = ((void *) dynamicParagraph) + sizeof(DynamicParagraph);
		visibleParagraphPixel += dynamicParagraph->height;
	}
	visibleParagraphPixel -= dynamicParagraph->height;

	SetFirstVisibleParagraph(index, visibleParagraphPixel);

	MemHandleUnlock(dynamicDocument);
	DrawDocument(document);
}

/***********************************************************************
 *
 * FUNCTION:    ScrollDocumentDown
 *
 * DESCRIPTION: Scroll document down.
 *
 * PARAMETERS:  document - pointer to document record
 *              height   - height of scrolled area
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ScrollDocumentDown(Document *document, const Int16 height)
{
	DynamicParagraph *dynamicParagraph;
	Int16 firstVisiblePixel, lastVisiblePixel;
	Int16 visibleParagraphPixel;
	Int16 index;
	DynamicDocument *dynamic = MemHandleLock(dynamicDocument);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "ScrollDocumentDown: MemHandleLock failed");

	firstVisiblePixel = -(dynamic->verticalOffset);
	index = dynamic->firstVisibleParagraph;

	/*
	   Work out the new "first visible" paragraph by searching up.
	 */
	dynamicParagraph = ((void *) dynamic) + sizeof(DynamicDocument) + index * sizeof(DynamicParagraph);
	visibleParagraphPixel = dynamic->firstParagraphY;

	while (visibleParagraphPixel > firstVisiblePixel) {
		index--;
		dynamicParagraph = ((void *) dynamicParagraph) - sizeof(DynamicParagraph);
		visibleParagraphPixel -= dynamicParagraph->height;
	}
	SetFirstVisibleParagraph(index, visibleParagraphPixel);

	lastVisiblePixel = viewportBounds.extent.y - dynamic->verticalOffset;
	index = dynamic->lastVisibleParagraph;

	/*
	   Work out the new "last visible" paragraph by searching up.
	 */
	dynamicParagraph = ((void *) dynamic) + sizeof(DynamicDocument) + index * sizeof(DynamicParagraph);
	visibleParagraphPixel = dynamic->lastParagraphY;

	while (visibleParagraphPixel > lastVisiblePixel) {
		index--;
		dynamicParagraph = ((void *) dynamicParagraph) - sizeof(DynamicParagraph);
		visibleParagraphPixel -= dynamicParagraph->height;
	}
	SetLastVisibleParagraph(index, visibleParagraphPixel);

	MemHandleUnlock(dynamicDocument);
	DrawDocument(document);
}

/***********************************************************************
 *
 * FUNCTION:    AdjustVerticalOffset
 *
 * DESCRIPTION: Adjust vertical offset of document.
 *
 * PARAMETERS:  document  - pointer to document record
 *              ajustment - adjustment in pixels
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void AdjustVerticalOffset(Document *document, Int16 adjustment)
{
	RectangleType vacatedRectangle;
	DynamicDocument newDocument;
	Int16 adjustAmount;
	DynamicDocument *dynamic = MemHandleLock(dynamicDocument);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "AdjustVerticalOffset: MemHandleLock failed");

	// Scrolling of Tbmp documents is handled by another function...
	if (dynamic->height == 0) {
		MemHandleUnlock(dynamicDocument);
		return;
	}
	/*
	   Adjust the document
	 */
	newDocument = *dynamic;
	if (newDocument.verticalOffset + adjustment >= 0) {
		if (newDocument.verticalOffset >= 0) {
			MemHandleUnlock(dynamicDocument);
			return;
		} else
			adjustment = -(newDocument.verticalOffset);
	} else if (newDocument.verticalOffset + adjustment < minVerticalOffset) {
		if (newDocument.verticalOffset <= minVerticalOffset) {
			adjustment = 0;
		} else
			adjustment = minVerticalOffset - newDocument.verticalOffset;
	}
	adjustAmount = (adjustment > 0) ? adjustment : -adjustment;
	newDocument.verticalOffset += adjustment;

	SetHistoryVerticalOffset(newDocument.verticalOffset);

	DmWrite((void *) dynamic, OFFSETOF(DynamicDocument, verticalOffset), 
		(void *) &(newDocument.verticalOffset), sizeof(Int16));

	/*
	   Scroll the window
	 */
	vacatedRectangle.topLeft.x = 0;
	vacatedRectangle.topLeft.y = 0;
	vacatedRectangle.extent.x = 0;
	vacatedRectangle.extent.y = 0;

	WinSetClip(&viewportBounds);
	WinScrollRectangle(&viewportBounds, (adjustment < 0) ? winUp : winDown, 
			   adjustAmount, &vacatedRectangle);
	WinEraseRectangle(&vacatedRectangle, 0);

	/*
	   Adjust the on-screen anchors. Be careful to do this BEFORE drawing
	   the new anchors!
	 */
	AdjustVisibleAnchors(adjustment);

	/*
	   This code also draws the new anchors.
	 */
	if (adjustment < 0)
		ScrollDocumentUp(document);
	else
		ScrollDocumentDown(document, adjustment);

	UpdateScrollbar(&newDocument);

	MemHandleUnlock(dynamicDocument);
	ShowVerticalOffset(document);

	WinResetClip();
}


/***********************************************************************
 *
 * FUNCTION:    GetSearchAdjustment
 *
 * DESCRIPTION: Find the necessary adjustment for the search pattern
 *              to be visible in the document.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Adjustment to make search pattern visible.
 *
 ***********************************************************************/
Int16 GetSearchAdjustment(void)
{
	Document *document;
	DynamicDocument *dynamic;
	Paragraph *paragraph;
	UInt16 numOfParagraphs;
	TextContext thisContext;
	Int16 adjustment;

	document = (Document *) MemHandleLock(currentDocument);
	FATAL_IF(document == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(document == NULL, "GetSearchAdjustment: MemHandleLock failed");

	numOfParagraphs = document->paragraphs;

	InitializeTextContext(&thisContext);
	thisContext.writeMode = false;
	thisContext.patternOffset = 0;

	paragraph = ((void *) document) + sizeof(Document);

	while (!thisContext.patternOffset && numOfParagraphs--) {
		DrawParagraph(&thisContext, paragraph, document);

		paragraph = ((void *) paragraph) + sizeof(Paragraph);
	}

	dynamic = MemHandleLock(dynamicDocument);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "GetSearchAdjustment: MemHandleLock failed");

	adjustment = -(dynamic->verticalOffset + thisContext.patternOffset);

	MemHandleUnlock(dynamicDocument);
	MemHandleUnlock(currentDocument);

	return adjustment;
}

/***********************************************************************
 *
 * FUNCTION:    CalculateDocumentHeight
 *
 * DESCRIPTION: Calculate the total height of the document and store it 
 *              in the document header.
 *
 * PARAMETERS:  document - pointer to document record
 *              pOffset  - offset to first paragraph
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static Int16 CalculateDocumentHeight(Document *document, Int16 pOffset)
{
	Paragraph *paragraph;
	UInt16 numOfParagraphs = document->paragraphs;
	Int16 paragraphHeight;
	TextContext thisContext;
	Int16 documentHeight = 0;
	Int16 lastCursor;
	Int16 index = 0;

	InitializeTextContext(&thisContext);

	if (pOffset)
		thisContext.writeMode = false;

	paragraph = ((void *) document) + sizeof(Document);

	WinSetClip(&viewportBounds);

	while (numOfParagraphs--) {
		if (!pOffset--)
			InitializeTextContext(&thisContext);

		lastCursor = thisContext.cursorY;

		DrawParagraph(&thisContext, paragraph, document);

		paragraphHeight = thisContext.cursorY - lastCursor;
		documentHeight += paragraphHeight;

		SetParagraphHeight(index, paragraphHeight);

		paragraph = ((void *) paragraph) + sizeof(Paragraph);
		index++;
	}
	WinResetClip();

	return documentHeight;
}

/***********************************************************************
 *
 * FUNCTION:    HandleNamedAnchor
 *
 * DESCRIPTION: Find the named anchor paragraph.
 *
 * PARAMETERS:  document - pointer to document record
 *              pOffset  - offset to first paragraph
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void HandleNamedAnchor(Document *document, Int16 pOffset)
{
	DynamicParagraph *dynamicParagraph;
	Int16 visibleParagraphPixel;
	Int16 heightDiff;
	Int16 verticalOffset;
	Int16 i;
	Int16 index;
	DynamicDocument *dynamic = MemHandleLock(dynamicDocument);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "HandleNamedAnchor: MemHandleLock failed");

	index = dynamic->firstVisibleParagraph;
	dynamicParagraph = ((void *) dynamic) + sizeof(DynamicDocument) + index * sizeof(DynamicParagraph);
	visibleParagraphPixel = dynamic->firstParagraphY;
	heightDiff = dynamic->height - viewportBounds.extent.y;

	WinEraseRectangle(&viewportBounds, 0);

	if (heightDiff <= 0) {
		MemHandleUnlock(dynamicDocument);
		return;
	}

	// Find named anchor paragraph
	for (i = 0; i < pOffset; i++) {
		visibleParagraphPixel += dynamicParagraph->height;
		dynamicParagraph = ((void *) dynamicParagraph) + sizeof(DynamicParagraph);
		index++;
	}

	verticalOffset = -visibleParagraphPixel;

	if (heightDiff < visibleParagraphPixel) {
		Int16 firstVisiblePixel = dynamic->height - viewportBounds.extent.y;

		dynamicParagraph = ((void *) dynamicParagraph) - sizeof(DynamicParagraph);
		visibleParagraphPixel -= dynamicParagraph->height;
		index--;

		while (visibleParagraphPixel > firstVisiblePixel) {
			dynamicParagraph = ((void *) dynamicParagraph) - sizeof(DynamicParagraph);
			visibleParagraphPixel -= dynamicParagraph->height;
			index--;
		}
		verticalOffset = viewportBounds.extent.y - dynamic->height;
	}
	SetFirstVisibleParagraph(index, visibleParagraphPixel);

	SetHistoryVerticalOffset(verticalOffset);

	DmWrite((void *) dynamic, OFFSETOF(DynamicDocument, verticalOffset), 
		(void *) &verticalOffset, sizeof(Int16));

	MemHandleUnlock(dynamicDocument);
}

/***********************************************************************
 *
 * FUNCTION:    ViewHtmlDocument
 *
 * DESCRIPTION: View Plucker HTML document.
 *
 * PARAMETERS:  document - pointer to document record
 *              newPage  - true if the page is not from the history
 *              pOffset  - offset to first paragraph
 *
 * RETURN:      False if the document could not be displayed because
 *              of insufficient memory
 *
 ***********************************************************************/
static Boolean ViewHtmlDocument(Document *document, Boolean newPage, 
				Int16 pOffset)
{
	DynamicDocument *dynamic;

	if (document->type == DOCTYPE_PHTML_COMPRESSED) {
		if (Uncompress(document) == NULL)
			return false;
	}

	dynamic = MemHandleLock(dynamicDocument);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "ViewHtmlDocument: MemHandleLock failed");

	if (newPage || dynamic->height == 0) {
		DynamicDocument documentData = *dynamic;

		documentData.verticalOffset = 0;

		DmWrite((void *) dynamic, OFFSETOF(DynamicDocument, verticalOffset),
			(void *) &(documentData.verticalOffset), sizeof(Int16));

		SetHistoryVerticalOffset(documentData.verticalOffset);
		SetFirstVisibleParagraph(0, 0);
	}
	if (dynamic->height == 0) {
		DynamicDocument documentData = *dynamic;

		SetWorking();
		documentData.height = CalculateDocumentHeight(document, pOffset);
		SetIdle();

		DmWrite((void *) dynamic, OFFSETOF(DynamicDocument, height), 
			(void *) &(documentData.height), sizeof(Int16));
	}
	if (pOffset)
		HandleNamedAnchor(document, pOffset);

	UpdateScrollbar(dynamic);
	ShowVerticalOffset(document);

	minVerticalOffset = viewportBounds.extent.y - dynamic->height;

	MemHandleUnlock(dynamicDocument);

	DrawDocument(document);

	return true;
}

/***********************************************************************
 *
 * FUNCTION:    DoPageMove
 *
 * DESCRIPTION: Move page up/down.
 *
 * PARAMETERS:  pixels - number of pixels to move (positive values move 
 *                                                 up, negative down)
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void DoPageMove(const Int16 pixels)
{
	Document *document = (Document *) MemHandleLock(currentDocument);
	FATAL_IF(document == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(document == NULL, "DoPageMove: MemHandleLock failed");

	InitViewportBounds();

	AdjustVerticalOffset(document, pixels);

	MemHandleUnlock(currentDocument);
}

/***********************************************************************
 *
 * FUNCTION:    GetDynamicHandle
 *
 * DESCRIPTION: Return a handle to the dynamic data.
 *
 * PARAMETERS:  document - pointer to document record
 *
 * RETURN:      Handle to the record
 *
 ***********************************************************************/
MemHandle GetDynamicHandle(const Int16 documentId)
{
	MemHandle dynamic;
	Document *document;
	UInt16 numOfParagraphs;

	document = (Document *) MemHandleLock(GetDocumentHandle(documentId));
	numOfParagraphs = document->paragraphs;
	MemPtrUnlock(document);

	if (ReturnDynamicHandle(documentId, numOfParagraphs, &dynamic) != 0)
			ErrDisplay("GetDynamicHandle: fatal error");

	return dynamic;
}

/***********************************************************************
 *
 * FUNCTION:    GetDocumentHandle
 *
 * DESCRIPTION: Return a handle to a document.
 *
 * PARAMETERS:  documentId - document ID
 *
 * RETURN:      Handle to the document
 *
 ***********************************************************************/
MemHandle GetDocumentHandle(const Int16 documentId)
{
	MemHandle document;

	if (ReturnDocumentHandle(documentId, &document) != 0)
		if (ReturnDocumentHandle(PLUCKER_LINKS_ID, &document) != 0)
			document = NULL;

	return document;
}

/***********************************************************************
 *
 * FUNCTION:    ViewDocument
 *
 * DESCRIPTION: View document (or image)
 *
 * PARAMETERS:  documentId - document ID
 *              newPage    - indicates if the page is from the history
 *                           or not
 *              pOffset    - offset to first paragraph
 *
 * RETURN:      False if the document/image could not be displayed 
 *              because of insufficient memory
 *
 ***********************************************************************/
Boolean ViewDocument(const Int16 documentId, const Boolean newPage, Int16 pOffset)
{
	Boolean status = false;
	Document *document;

	MSG(_("document #%d\n", documentId));

	ClearVisibleAnchors();
	InitViewportBounds();
	DeleteOffScreenWindow();
	WinEraseRectangle(&viewportBounds, 0);

	currentDocument = GetDocumentHandle(documentId);
	if (currentDocument == NULL) {
		SetLinkIndex(documentId);
		FrmGotoForm(frmExternalLinks);

		return true;
	}

	document = (Document *) MemHandleLock(currentDocument);
	FATAL_IF(document == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(document == NULL, "ViewDocument: MemHandleLock failed");

	switch (document->type) {
	case DOCTYPE_PHTML:		/* Plucker-HTML */
	case DOCTYPE_PHTML_COMPRESSED:	/* Compressed Plucker-HTML */
		if (newPage)
			AddHistoryDocument(documentId);

		dynamicDocument = GetDynamicHandle(documentId);
		FATAL_IF(dynamicDocument == NULL, "NULL document\n");
		ErrFatalDisplayIf(dynamicDocument == NULL, "ViewDocument: NULL document");

		status = ViewHtmlDocument(document, newPage, pOffset);
		break;

	case DOCTYPE_TBMP:		/* PalmOS bitmap */
	case DOCTYPE_TBMP_COMPRESSED:	/* Compressed PalmOS bitmap */
		if (newPage)
			AddHistoryDocument(documentId);

		dynamicDocument = GetDynamicHandle(documentId);
		FATAL_IF(dynamicDocument == NULL, "NULL document\n");
		ErrFatalDisplayIf(dynamicDocument == NULL, "ViewDocument: NULL document");

		status = ViewTbmpDocument(document, newPage);
		break;

	case DOCTYPE_MAILTO:		/* Mailto tag info */
		SetMailto(documentId);
		FrmGotoForm(frmEmail);
		status = true;
		break;

	case DOCTYPE_LINK_INDEX:	/* Link info */
		SetLinkIndex(documentId);
		FrmGotoForm(frmExternalLinks);
		status = true;
		break;

	default:
		ErrDisplay("ViewDocument: unknown document type");
		break;
	}
	MemHandleUnlock(currentDocument);
	return status;
}

/***********************************************************************
 *
 * FUNCTION:    ResetDocumentHeight
 *
 * DESCRIPTION: Set height of document to zero
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ResetDocumentHeight(void)
{
	Int16 height = 0;
	DynamicDocument *dynamic = MemHandleLock(dynamicDocument);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "ResetDocumentHeight: MemHandleLock failed");
	
	DmWrite((void *) dynamic, OFFSETOF(DynamicDocument, height), 
		(void *) &height, sizeof(Int16));

	MemHandleUnlock(dynamicDocument);
}

/***********************************************************************
 *
 * FUNCTION:    GetCurrentDocument
 *
 * DESCRIPTION: Return handle to current document
 *
 * PARAMETERS:  None
 *
 * RETURN:      Handle to current document
 *
 ***********************************************************************/
MemHandle GetCurrentDocument(void)
{
	return currentDocument;
}

/***********************************************************************
 *
 * FUNCTION:    GetDynamicDocument
 *
 * DESCRIPTION: Return handle to dynamic document
 *
 * PARAMETERS:  None
 *
 * RETURN:      Handle to dynamic document
 *
 ***********************************************************************/
MemHandle GetDynamicDocument(void)
{
	return dynamicDocument;
}

// vim: ts=4
