/*
 * $Id: documentdata.c,v 1.26 2000/10/28 11:27:50 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "categorydata.h"
#include "const.h"
#include "db.h"
#include "debug.h"
#include "document.h"
#include "documentdata.h"
#include "history.h"
#include "image.h"
#include "os.h"
#include "paragraph.h"
#include "prefsdata.h"
#include "resourceids.h"
#include "uncompress.h"
#include "util.h"

UInt16 GetDefaultCategories(LocalID dbID, UInt16 cardNo) __attribute__ ((section ("plkr")));

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define REFERENCE_RECORD_SIZE 64 + sizeof(Header)
#define SESSION_RECORD_SIZE sizeof(History) + sizeof(Header)
#define BOOKMARK_RECORD_SIZE 2 * sizeof(UInt16) + sizeof(Header)

#define BIT_ON 1

#define INDEX_RECORD 1

/***********************************************************************
 *
 *      Internal Structures
 *
 ***********************************************************************/
typedef enum DBTypeEnum { DBTYPE_UNKNOWN = 0,
			  DBTYPE_DOC     = 1,
			  DBTYPE_ZLIB    = 2
			} DBType;

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static DmOpenRef ViewerDB;
static DmOpenRef DynamicDB;

static UInt16 reservedRecords[MAX_RESERVED];
static UInt16 internalRecords[MAX_INTERNAL];

/***********************************************************************
 *
 * FUNCTION:    ClearRecordIDs
 *
 * DESCRIPTION: Clear the arrays holding record IDs
 *
 * PARAMETERS:  None
 * 
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ClearRecordIDs(void)
{
	MemSet((void *) &reservedRecords, MAX_RESERVED * sizeof(UInt16), 0);
	MemSet((void *) &internalRecords, MAX_INTERNAL * sizeof(UInt16), 0);
}

/***********************************************************************
 *
 * FUNCTION:    DocumentRecord
 *
 * DESCRIPTION: Check if referenced record is a document record
 *
 * PARAMETERS:  id - unique ID to check
 *
 * RETURN:      True if it is a document record
 *
 * NOTE:	    A record is defined as a document record if it is not 
 *              one of the reserved internal records (except for the
 *              home page)
 *
 ***********************************************************************/
static Boolean DocumentRecord(UInt16 id)
{
	UInt16 i;

	if (id == 0)
		return false;

	for (i = 1; i < MAX_INTERNAL; i++)
		if (id == internalRecords[i])
			return false;
	return true;
}

/***********************************************************************
 *
 * FUNCTION:    GetReservedID
 *
 * DESCRIPTION: Return the reserved record's unique ID
 *
 * PARAMETERS:  name - record "name"
 *
 * RETURN:      Unique ID
 *
 ***********************************************************************/
UInt16 GetReservedID(UInt16 name)
{
	if (name < MAX_RESERVED)
		return reservedRecords[name];
	else
		return 0;
}

/***********************************************************************
 *
 * FUNCTION:    GetInternalID
 *
 * DESCRIPTION: Return the internal record's unique ID
 *
 * PARAMETERS:  name - record "name"
 *
 * RETURN:      Unique ID
 *
 ***********************************************************************/
UInt16 GetInternalID(UInt16 name)
{
	if (name < MAX_INTERNAL)
		return internalRecords[name];
	else
		return 0;
}

/***********************************************************************
 *
 * FUNCTION:    CompareUID
 *
 * DESCRIPTION: 
 *
 * PARAMETERS: rec1         - Pointer to the record to sort. 
 *             rec2         - Pointer to the record to sort. 
 *             other        - Any other custom information you want 
 *                            passed to the comparison function.
 *             rec1SortInfo - Pointer to SortRecordInfoType structure 
 *                            that specify unique sorting information 
 *                            for the record.
 *             rec2SortInfo - Pointer to SortRecordInfoType structure 
 *                            that specify unique sorting information 
 *                            for the record.
 *             appInfoH     - A handle to the database's application 
 *                            info block.
 *
 * RETURN:     0 if rec1 = rec2.  
 *           < 0 if rec1 < rec2.  
 *           > 0 if rec1 > rec2.  
 *
 ***********************************************************************/
Int16 CompareUID(void* rec1, void* rec2, Int16 other, 
		 SortRecordInfoPtr rec1SortInfo, 
		 SortRecordInfoPtr rec2SortInfo, 
		 MemHandle appInfoH) 
{
	Int16 result;

	SET_A4_FROM_A5

	if (((Header*)rec1)->uid < ((Header*)rec2)->uid)
		result = -1;
	else if (((Header*)rec1)->uid > ((Header*)rec2)->uid)
		result = 1;
	else
		result = 0;

	RESTORE_A4

	return result;
}

/***********************************************************************
 *
 * FUNCTION:    FindRecord
 *
 * DESCRIPTION: Find record in database
 *
 * PARAMETERS:  dbRef  - reference to database
 *              item   - UID to search for
 *              handle - upon successful return, the handle to the record,
 *                       otherwise undefined
 *              index  - upon successful return, the index of the record,
 *                       otherwise undefined. Pass NULL for this parameter 
 *                       if you don't want to retrieve this value.
 *
 * RETURN:      0 if successful
 *
 ***********************************************************************/
static Int16 FindRecord(DmOpenRef dbRef, Int16 item, MemHandle *handle, Int16 *index)
{
	Int16 l = 0;
	Int16 r = DmNumRecords(dbRef) - 1;
	Int16 x;
	Header *header;

	while (r >= l) {
		x = (r + l) / 2;

		*handle = DmQueryRecord(dbRef, x);
		if (*handle == NULL)
			break;

		header = (Header *) MemHandleLock(*handle);
		FATAL_IF(header == NULL, "MemHandleLock failed\n");
		ErrFatalDisplayIf(header == NULL, "FindRecord: MemHandleLock failed");

		if (item == header->uid)  {
			MemHandleUnlock(*handle);
			if (index)
				*index = x;
			return 0;
		}
		if (item < header->uid) 
			r = x - 1;
		else 
			l = x + 1;

		MemHandleUnlock(*handle);
	}
	return 1;
}

/***********************************************************************
 *
 * FUNCTION:    InitializeDynamicRecord
 *
 * DESCRIPTION: Initialize the dynamic data.
 *
 * PARAMETERS:  handle - handle to record
 *              size   - record size
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void InitializeDynamicRecord(MemHandle handle, UInt16 size)
{
	DynamicDocument *dynamic;
	UInt16 docDepth = Prefs()->screenDepth;

	dynamic = MemHandleLock(handle);
	FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(dynamic == NULL, "InitializeDynamicRecord: MemHandleLock failed");

	DmSet(dynamic, sizeof(Header), size - sizeof(Header), 0);
	DmWrite((void *) dynamic, OFFSETOF(DynamicDocument, depth), 
		(void *) &docDepth, sizeof(UInt16));

	MemHandleUnlock(handle);
}

/***********************************************************************
 *
 * FUNCTION:    InitializeSessionRecord
 *
 * DESCRIPTION: Initialize the session data.
 *
 * PARAMETERS:  handle - handle to record
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void InitializeSessionRecord(MemHandle handle)
{
	void *sessionPtr;

	InitHistory();

	sessionPtr = MemHandleLock(handle);
	FATAL_IF(sessionPtr == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(sessionPtr == NULL, "InitializeSessionRecord: MemHandleLock failed");

	SaveHistory(sessionPtr);

	MemHandleUnlock(handle);
}

/***********************************************************************
 *
 * FUNCTION:    InitializeReferenceRecord
 *
 * DESCRIPTION: Initialize the show images or link data.
 *
 * PARAMETERS:  handle - handle to show images or link record
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void InitializeReferenceRecord(MemHandle handle)
{
	void *referencePtr;

	referencePtr = MemHandleLock(handle);
	FATAL_IF(referencePtr == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(referencePtr == NULL, "InitializeReferenceRecord: MemHandleLock failed");

	DmSet(referencePtr, sizeof(Header), REFERENCE_RECORD_SIZE - sizeof(Header), 0);

	MemHandleUnlock(handle);
}

/***********************************************************************
 *
 * FUNCTION:    InitializeBookmarkRecord
 *
 * DESCRIPTION: Initialize the bookmarks record
 *
 * PARAMETERS:  handle - handle to record
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void InitializeBookmarkRecord(MemHandle handle)
{
	void *bookmarkPtr;
	UInt16 offset = BOOKMARK_RECORD_SIZE;

	bookmarkPtr = MemHandleLock(handle);
	FATAL_IF(bookmarkPtr == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(bookmarkPtr == NULL, "InitializeBookmarkRecord: MemHandleLock failed");

	DmSet(bookmarkPtr, sizeof(Header), BOOKMARK_RECORD_SIZE - sizeof(Header), 0);
	DmWrite(bookmarkPtr, sizeof(Header) + sizeof(UInt16), (void *) &offset, sizeof(UInt16));

	MemHandleUnlock(handle);
}

/***********************************************************************
 *
 * FUNCTION:    InitializeBookmarkRecord
 *
 * DESCRIPTION: Initialize the bookmarks record
 *
 * PARAMETERS:  handle - handle to record
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void InitializeIndexRecord(MemHandle handle)
{
	void *indexPtr;
	UInt16 records = MAX_INTERNAL;

	indexPtr = MemHandleLock(handle);
	FATAL_IF(indexPtr == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(indexPtr == NULL, "InitializeIndexRecord: MemHandleLock failed");

	DmWrite(indexPtr, OFFSETOF(DynamicIndex, records), (void*) &records, sizeof(UInt16));
	DmWrite(indexPtr, OFFSETOF(DynamicIndex, internal), (void *) &internalRecords, MAX_INTERNAL * sizeof(UInt16));

	MemHandleUnlock(handle);
}

/***********************************************************************
 *
 * FUNCTION:    AddRecord
 *
 * DESCRIPTION: Add record to database
 *
 * PARAMETERS:  id   - unique ID of record to create
 *              size - record size
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
static Err AddRecord(UInt16 id, UInt16 size)
{
	Err err;
	MemHandle handle;
	MemPtr handlePtr;
	UInt16 dbIndex = dmMaxRecordIndex;

	handle = DmNewRecord(DynamicDB, &dbIndex, size);
	if (handle == NULL)
		return DmGetLastErr();

	handlePtr = MemHandleLock(handle);
	FATAL_IF(handlePtr == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(handlePtr == NULL, "AddRecord: MemHandleLock failed");

	DmWrite(handlePtr, 0, (void *)&id, sizeof(Header));
	MemHandleUnlock(handle);

	if (id == INDEX_RECORD)
		InitializeIndexRecord(handle);
	else if (DocumentRecord(id))
		InitializeDynamicRecord(handle, size);
	else if (id == internalRecords[_SESSION_DATA])
		InitializeSessionRecord(handle);
	else if (id == internalRecords[_SHOW_IMAGES] || id == internalRecords[_LINK_TABLE])
		InitializeReferenceRecord(handle);
	else if (id == internalRecords[_INTERNAL_BOOKMARKS])
		InitializeBookmarkRecord(handle);

	err = DmReleaseRecord(DynamicDB, dbIndex, true);

	DmInsertionSort(DynamicDB, (DmComparF *)CompareUID, 0);

	return err;
}

/***********************************************************************
 *
 * FUNCTION:    GetIndexData
 *
 * DESCRIPTION: Check the compression type and retrieve IDs for 
 *              reserved records
 *
 * PARAMETERS:  None
 *
 * RETURN:      The compression type
 *
 ***********************************************************************/
static DBType GetIndexData(void)
{
	MemHandle hand;
	Err err;
	DBType dbType = DBTYPE_UNKNOWN;

	err = FindRecord(ViewerDB, INDEX_RECORD, &hand, NULL);

	if (!err) {
		Index *indexRecord;
		UInt16 id, name, *reserved;
		UInt16 i;

		indexRecord = (Index *) MemHandleLock(hand);
		FATAL_IF(indexRecord == NULL, "MemHandleLock failed\n");
		ErrFatalDisplayIf(indexRecord == NULL, "CheckDBType: MemHandleLock failed");

		dbType = indexRecord->version;
		reserved = indexRecord->reserved;

		for (i = 0; i < indexRecord->records; i++) {
			name = *reserved;
			id = *(reserved + 1);
			if (name < MAX_RESERVED)
				reservedRecords[name] = id;
			reserved = (void*)reserved + 2 * sizeof(UInt16);
		}
		MemHandleUnlock(hand);

		if (dbType != DBTYPE_ZLIB && dbType != DBTYPE_DOC)
			dbType = DBTYPE_UNKNOWN;
	}
	return dbType;
}

/***********************************************************************
 *
 * FUNCTION:    GetInternalIndexData
 *
 * DESCRIPTION: Retrieve ID's for internal records
 *
 * PARAMETERS:  None
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
static Err GetInternalIndexData(void)
{
	MemHandle hand;
	Err err;

	err = FindRecord(DynamicDB, INDEX_RECORD, &hand, NULL);

	if (!err) {
		DynamicIndex *indexRecord;
		UInt16 id, *internal;
		UInt16 i;

		indexRecord = (DynamicIndex *) MemHandleLock(hand);
		FATAL_IF(indexRecord == NULL, "MemHandleLock failed\n");
		ErrFatalDisplayIf(indexRecord == NULL, "CheckDBType: MemHandleLock failed");

		internal = indexRecord->internal;

		for (i = 0; i < indexRecord->records; i++) {
			id = *internal;
			internalRecords[i] = id;
			internal = (void*)internal + sizeof(UInt16);
		}
		MemHandleUnlock(hand);
	}
	return err;
}

/***********************************************************************
 *
 * FUNCTION:    InitializeDynamicDatabase
 *
 * DESCRIPTION: Initialize the dynamic database
 *
 * PARAMETERS:  None
 *
 * RETURN:      The result from the database call
 *
 * NOTE:        Adds records for session handling, link handling 
 *              and image handling.
 * 
 ***********************************************************************/
static Err InitializeDynamicDatabase(void)
{
	Err err;
	Header *header;
	MemHandle handle;
	UInt16 i, maxID;
	UInt16 lastIndex = DmNumRecords(ViewerDB) - 1;

	handle = DmQueryRecord(ViewerDB, lastIndex);
	if (handle == NULL)
		return 1;

	header = (Header *) MemHandleLock(handle);
	FATAL_IF(header == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(header == NULL, "InitializeDynamicDatabase: MemHandleLock failed");
	maxID = header->uid;
	MemHandleUnlock(handle);

	internalRecords[_HOME_PAGE] = reservedRecords[HOME_PAGE];
	for (i = 1; i < MAX_INTERNAL; i++)
		internalRecords[i] = maxID + i;
	err = AddRecord(INDEX_RECORD, (MAX_INTERNAL + 2) * sizeof(UInt16));
	err |= AddRecord(internalRecords[_SHOW_IMAGES], REFERENCE_RECORD_SIZE);
	err |= AddRecord(internalRecords[_SESSION_DATA], SESSION_RECORD_SIZE);
	err |= AddRecord(internalRecords[_LINK_TABLE], REFERENCE_RECORD_SIZE);

	return err;
}

/***********************************************************************
 *
 * FUNCTION:    OpenDynamicDatabase
 *
 * DESCRIPTION: Open dynamic database.
 *
 * PARAMETERS:  name - the name of the associated document database
 *              date - the date the associated database was created
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
static Err OpenDynamicDatabase(Char *name, UInt32 date)
{
	UInt32 dynamicDate = 0;
	UInt16 cardNo = 0;
	Err err = 0;
	LocalID dynamicID;
	Char dynamicName[dmDBNameLength];

	StrCopy(dynamicName, "Plkr-");
	StrNCat(dynamicName, name, dmDBNameLength);
	dynamicID = DmFindDatabase(cardNo, dynamicName);
	if (dynamicID) {
		DmDatabaseInfo(cardNo, dynamicID, NULL, NULL, NULL, &dynamicDate, 
			       NULL, NULL, NULL, NULL, NULL, NULL, NULL);

		if (dynamicDate < date)
			DmDeleteDatabase(cardNo, dynamicID);
		else {
			DynamicDB = DmOpenDatabase(cardNo, dynamicID, dmModeReadWrite);
			GetInternalIndexData();
			return 0;
		}
	}
	err = DmCreateDatabase(cardNo, dynamicName, ViewerAppID, DynamicDBType, false);
	if (err) {
		MSG("Couldn't create dynamic database\n");
		FrmAlert(warnInsufficientMemory);
		return err;
	}
	dynamicID = DmFindDatabase(cardNo, dynamicName);
	DynamicDB = DmOpenDatabase(cardNo, dynamicID, dmModeReadWrite);

	err = InitializeDynamicDatabase();
	if (err) {
		CloseDocumentDatabase();
		MSG("Couldn't initialize dynamic database\n");
		FrmAlert(warnInsufficientMemory);
	}
	return err;
}

/***********************************************************************
 *
 * FUNCTION:    OpenDocumentDatabase
 *
 * DESCRIPTION: Open document database.
 *
 * PARAMETERS:  dbID - the database ID of the database
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
Err OpenDocumentDatabase(const LocalID dbID, UInt16 cardNo)
{
	MemHandle hand;
	UInt32 date = 0;
	UInt16 version;
	Err err;
	DBType dbType;
	Char name[dmDBNameLength];

	ClearRecordIDs();
	ResetImageData();

	if (dbID == NULL || (err = OpenDatabase(&ViewerDB, dbID, cardNo, &version, name, &date))) {
		MSG("no PluckerDB installed\n");
		FrmAlert(infoNoPluckerDB);
		return dmErrCantOpen;
	}
	dbType = GetIndexData();
	if (dbType == DBTYPE_ZLIB) {
		if (ZLibSupported())
			uncompressFunc = (MemHandle (*)(Document*)) &UnZip;
		else {
			CloseDocumentDatabase();
			MSG("ZLib compressed databases not supported\n");
			FrmCustomAlert(warnNoZLibSupport, name, NULL, NULL);
			return 1;
		}
	} else if (dbType == DBTYPE_DOC)
		uncompressFunc = (MemHandle (*)(Document*)) &UnDoc;
	else {
		CloseDocumentDatabase();
		MSG("Not a valid Plucker database\n");
		FrmCustomAlert(errUnknownType, name, NULL, NULL);
		return 1;
	}
	if (reservedRecords[HOME_PAGE])
		err = FindRecord(ViewerDB, reservedRecords[HOME_PAGE], &hand, NULL);
	else
		err = 1;

	if (err != 0) {
		CloseDocumentDatabase();
		MSG("couldn't open/initialize database\n");
		FrmCustomAlert(warnBrokenDatabase, name, NULL, NULL);
	} else
		OpenDynamicDatabase(name, date);

	return err;
}

/***********************************************************************
 *
 * FUNCTION:    CloseDocumentDatabase
 *
 * DESCRIPTION: Close document database.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void CloseDocumentDatabase(void)
{
	if (ViewerDB == NULL)
		return;
	CloseDatabase(ViewerDB);
	ViewerDB = NULL;

	if (DynamicDB == NULL)
		return;
	CloseDatabase(DynamicDB);
	DynamicDB = NULL;
}

/***********************************************************************
 *
 * FUNCTION:    DeleteDynamicDatabase
 *
 * DESCRIPTION: Delete dynamic database
 *
 * PARAMETERS:  name - the name of the associated document database
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void DeleteDynamicDatabase(Char *name)
{
	UInt16 cardNo = 0;
	LocalID dbID;
	Char dynamicName[dmDBNameLength];

	if (DynamicDB != NULL)
		return;

	StrCopy(dynamicName, "Plkr-");
	StrNCat(dynamicName, name, dmDBNameLength);

	dbID = DmFindDatabase(cardNo, dynamicName);
	if (dbID)
		DmDeleteDatabase(cardNo, dbID);
}

/***********************************************************************
 *
 * FUNCTION:    RenameDatabase
 *
 * DESCRIPTION: Rename database and its associated dynamic database
 *
 * PARAMETERS:  newName   - the new name of the document database
 *              oldName   - the old name of the document database
 *              oldCardNo - card that the old database resides on
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
Err RenameDatabase(Char *newName, Char *oldName, UInt16 oldCardNo)
{
	Char dynNewName[dmDBNameLength];
	Char dynOldName[dmDBNameLength];
	LocalID dbID;
	UInt16 cardNo = 0;
	Err err = 1;

	// If a database with the given name doesn't already exists, rename it
	dbID = DmFindDatabase(oldCardNo, newName);
	if (!dbID) {
		dbID = DmFindDatabase(oldCardNo, oldName);
		err = DmSetDatabaseInfo(oldCardNo, dbID, newName, NULL, NULL, NULL, 
								NULL, NULL, NULL, NULL, NULL, NULL, NULL); 
	}
	if (err)
		return err;

	// Remove any "left-overs"
	StrCopy(dynNewName, "Plkr-");
	StrNCat(dynNewName, newName, dmDBNameLength);
	dbID = DmFindDatabase(cardNo, dynNewName);
	if (dbID)
		err = DmDeleteDatabase(cardNo, dbID);

	if (err)
		return err;

	// Rename dynamic database (if it exists)
	StrCopy(dynOldName, "Plkr-");
	StrNCat(dynOldName, oldName, dmDBNameLength);
	dbID = DmFindDatabase(cardNo, dynOldName);
	if (dbID)
		err = DmSetDatabaseInfo(cardNo, dbID, dynNewName, 
								NULL, NULL, NULL, NULL, NULL, 
								NULL, NULL, NULL, NULL, NULL); 

	return err;
}

/***********************************************************************
 *
 * FUNCTION:    ReturnDynamicHandle
 *
 * DESCRIPTION: Return a handle to a dynamic record.
 *
 * PARAMETERS:  id   - unique ID to search for
 *              num  - number of paragraphs in record (used to 
 *                     initialize a new record)
 *              hand - upon successful return, the handle to the 
 *                     record, otherwise undefined
 *
 * RETURN:      The result from the database call
 *
 * NOTE:        If the bit depth for the document is different from
 *              the current screen depth the document height will be
 *              set to 0, so that it will be re-calculated when the
 *              document is rendered
 *
 ***********************************************************************/
Err ReturnDynamicHandle(UInt16 id, const UInt16 num, MemHandle *hand)
{
	Err err;

	FATAL_IF(hand == NULL, "NULL argument\n");
	ErrFatalDisplayIf(hand == NULL, "ReturnDynamicHandle: NULL argument");

	if (DynamicDB == NULL || id == 0)
		return 1;

	err = FindRecord(DynamicDB, id, hand, NULL);

	if (!err && DocumentRecord(id)) {
		DynamicDocument *dynamic;

		dynamic = MemHandleLock(*hand);
		if (dynamic->depth != Prefs()->screenDepth) {
			Int16 height = 0;
			UInt16 docDepth = Prefs()->screenDepth;

			DmWrite((void *) dynamic, OFFSETOF(DynamicDocument, height), 
				(void *) &height, sizeof(Int16));
			DmWrite((void *) dynamic, OFFSETOF(DynamicDocument, depth), 
				(void *) &docDepth, sizeof(UInt16));
		}
		MemHandleUnlock(*hand);
	} else if (err && id != internalRecords[_INTERNAL_BOOKMARKS]) {
		UInt16 size = sizeof(DynamicDocument) + num * sizeof(DynamicParagraph);
		err = AddRecord(id, size);
		if (!err)
			err = FindRecord(DynamicDB, id, hand, NULL);
	}
	return err;
}

/***********************************************************************
 *
 * FUNCTION:    ReturnDocumentHandle
 *
 * DESCRIPTION: Return a handle to a record.
 *
 * PARAMETERS:  id   - unique ID to search for
 *              hand - upon successful return, the handle to the record,
 *                     otherwise undefined
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
Err ReturnDocumentHandle(const UInt16 id, MemHandle *hand)
{
	Err err;

	if (ViewerDB == NULL || id == 0)
		return 1;

	FATAL_IF(hand == NULL, "NULL argument\n");
	ErrFatalDisplayIf(hand == NULL, "ReturnDocumentHandle: NULL argument");

	err = FindRecord(ViewerDB, id, hand, NULL);

	return err;
}

/***********************************************************************
 *
 * FUNCTION:    ReturnNextDocumentHandle
 *
 * DESCRIPTION: Return a handle to next document record.
 *
 * PARAMETERS:  index  - pointer to index of a known record
 *              hand   - upon successful return, the handle to the record,
 *                       otherwise undefined
 *              uniqID - record's unique ID
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
Err ReturnNextDocumentHandle(UInt16 *index, MemHandle *hand, UInt16 *uniqID)
{
	Document *document;
	UInt16 lastIndex = DmNumRecords(ViewerDB) - 1;

	FATAL_IF(hand == NULL, "NULL argument\n");
	ErrFatalDisplayIf(hand == NULL, "ReturnNextDocumentHandle: NULL argument");

	if (*index == 0)
			*index = 1;

	while (1) {
		if (*index <= lastIndex)
			*hand = DmQueryRecord(ViewerDB, *index);
		else
			return 1;

		if (*hand == NULL)
			return 1;

		document = (Document *) MemHandleLock(*hand);
		FATAL_IF(document == NULL, "MemHandleLock failed\n"); 
		ErrFatalDisplayIf(document == NULL, "ReturnNextDocumentHandle: MemHandleLock failed"); 

		if (document->type == DOCTYPE_PHTML || 
			document->type == DOCTYPE_PHTML_COMPRESSED) {
			*uniqID = document->uid;
			MemHandleUnlock(*hand);
			return 0;
		}
		MemHandleUnlock(*hand);
		(*index)++;
	}
}

/***********************************************************************
 *
 * FUNCTION:    RemoveDynamicRecord
 *
 * DESCRIPTION: Remove dynamic record
 *
 * PARAMETERS:  id   - unique ID to remove
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
static Err RemoveDynamicRecord(const UInt16 id)
{
	Err err;
	MemHandle hand;
	UInt16 dbIndex;

	if (DynamicDB == NULL)
		return 1;

	err = FindRecord(DynamicDB, id, &hand, &dbIndex);
	if (!err)
		err = DmRemoveRecord(DynamicDB, dbIndex);

	return err;
}

/***********************************************************************
 *
 * FUNCTION:    ResetAllDynamicDocuments
 *
 * DESCRIPTION: Reset values for all dynamic documents (i.e. remove them)
 *
 * PARAMETERS:  None
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
/*
Err ResetAllDynamicDocuments(void)
{
	Header *header;
	Err err = 0;
	UInt16 dbIndex;
	UInt16 id;
	MemHandle hand;

	if (DynamicDB == NULL)
		return 1;

	dbIndex = DmNumRecords(DynamicDB);

	while (dbIndex--) {
		hand = DmQueryRecord(DynamicDB, dbIndex);
		if (hand == NULL)
			break;

		header = (Header *) MemHandleLock(hand);
		id = header->uid;
		MemHandleUnlock(hand);

		if (DocumentRecord(id))
			err |= DmRemoveRecord(DynamicDB, dbIndex);
	}
	ResetHistory();
	return err;
}
*/

/***********************************************************************
 *
 * FUNCTION:    GetNumOfRecords
 *
 * DESCRIPTION: Return number of records in database
 *
 * PARAMETERS:  None
 *
 * RETURN:      Number of records in database
 *
 ***********************************************************************/
UInt16 GetNumOfRecords(void)
{
	return DmNumRecords(ViewerDB);
}

/***********************************************************************
 *
 * FUNCTION:    ResizeRecord
 *
 * DESCRIPTION: Resize record
 *
 * PARAMETERS:  id   - unique ID to search for
 *              size - size of new record
 *              hand - upon successful return, the handle to the record,
 *                     otherwise undefined
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
Err ResizeRecord(const UInt16 id, const UInt16 size, MemHandle *hand)
{
	Err err;
	UInt16 dbIndex;

	if (id == 0)
		return 1;

	err = FindRecord(DynamicDB, id, hand, &dbIndex);
	if (!err) {
		*hand = DmResizeRecord(DynamicDB, dbIndex, size);
		DmInsertionSort(DynamicDB, (DmComparF *)CompareUID, 0);
	}
	return err;
}

/***********************************************************************
 *
 * FUNCTION:    CheckSize
 *
 * DESCRIPTION: Check that the size of the record is enough
 *              to hold the required data otherwise resize it.
 *
 * PARAMETERS:  id        - the record ID to check
 *              reference - reference to the bit to check
 *              handle    - handle to record
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void CheckSize(const UInt16 id, Int16 reference, MemHandle handle)
{
	UInt16 size = reference / 8 + 1 + sizeof(Header);
	UInt16 origSize = MemHandleSize(handle);

	if (origSize < size) {
		void *recPtr;

		ResizeRecord(id, size, &handle);

		recPtr = MemHandleLock(handle);
		DmSet(recPtr, origSize, size - origSize, 0);
		MemHandleUnlock(handle);
	}
}

/***********************************************************************
 *
 * FUNCTION:    GetBitStatus
 *
 * DESCRIPTION: Get the status for a specific bit in the given record
 *
 * PARAMETERS:  id        - the record ID to check
 *              reference - reference to the bit to check
 *
 * RETURN:      The bit status (on = true, off = false)
 *
 ***********************************************************************/
Boolean GetBitStatus(const UInt16 id, Int16 reference)
{
	MemHandle handle = NULL;
	Char *recordPtr;
	Int16 value;

	if (ReturnDynamicHandle(id, 0, &handle) != 0)
		return false;

	CheckSize(id, reference, handle);

	recordPtr = (Char *) MemHandleLock(handle);
	FATAL_IF(recordPtr == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(recordPtr == NULL, "GetBitStatus: MemHandleLock failed");

	recordPtr = (void*)recordPtr + sizeof(Header);

	reference -= 1;
	value = (*(recordPtr + reference / 8) >> reference % 8) & 0x01;

	MemHandleUnlock(handle);

	return (value == BIT_ON);
}

/***********************************************************************
 *
 * FUNCTION:    SetBitStatus
 *
 * DESCRIPTION: Set the status of a bit in the specified record
 *
 * PARAMETERS:  id        - the record ID to set
 *              reference - the reference to set
 *              setStatus - indicates whether the bit should be set 
 *                          or unset
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void SetBitStatus(const UInt16 id, Int16 reference, Boolean setStatus)
{
	MemHandle handle = NULL;
	Char *recordPtr;
	Char bitStatus, byte, bit;
	Int16 byteOffset;

	if (ReturnDynamicHandle(id, 0, &handle) != 0)
		return;

	CheckSize(id, reference, handle);

	recordPtr = (Char *) MemHandleLock(handle);
	FATAL_IF(recordPtr == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(recordPtr == NULL, "SetBitStatus: MemHandleLock failed");

	reference -= 1;
	byteOffset = reference / 8 + sizeof(Header);

	byte = *(recordPtr + byteOffset);
	bit = 0x01 << reference % 8;

	if (setStatus)
		bitStatus = byte | bit;
	else
		bitStatus = byte & ~bit;

	DmWrite(recordPtr, byteOffset, (void *) &bitStatus, 1);

	MemHandleUnlock(handle);
}

/***********************************************************************
 *
 * FUNCTION:    AddBookmarkRecord
 *
 * DESCRIPTION: Add bookmark record to database
 *
 * PARAMETERS:  hand - upon successful return, the handle to the record,
 *                     otherwise undefined
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
Err AddBookmarkRecord(MemHandle *hand)
{
	Err err;

	err = AddRecord(internalRecords[_INTERNAL_BOOKMARKS], BOOKMARK_RECORD_SIZE);

	if (!err)
		err = ReturnDynamicHandle(internalRecords[_INTERNAL_BOOKMARKS], 0, hand);

	return err;
}

/***********************************************************************
 *
 * FUNCTION:    RemoveBookmarkRecord
 *
 * DESCRIPTION: Remove bookmark record from database
 *
 * PARAMETERS:  None
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
Err RemoveBookmarkRecord(void)
{
	return RemoveDynamicRecord(internalRecords[_INTERNAL_BOOKMARKS]);
}

/***********************************************************************
 *
 * FUNCTION:    GetDefaultCategories
 *
 * DESCRIPTION: Retrieve the default category from the document database
 *
 * PARAMETERS:  dbID   - the database ID of the database
 *              cardNo - the card the database is located on
 *
 * RETURN:      If the category record is found and the requested 
 *              category is defined it will return the category, 
 *              otherwise 1 (i.e. Unfiled)
 *
 ***********************************************************************/
UInt16 GetDefaultCategories(LocalID dbID, UInt16 cardNo)
{
	UInt16 categories = 0;
	MemHandle handle;
	Err err;

	if (ViewerDB != NULL)
		return 1;
	ViewerDB = DmOpenDatabase(cardNo, dbID, dmModeReadOnly);
	if (ViewerDB == NULL)
		return 1;

	GetIndexData();
	err = FindRecord(ViewerDB, CATEGORY_ID, &handle, NULL);
	if (err)
		categories = 1;
	else {
		Document *categoryRecord;
		Char *category;
		UInt8 index;

		categoryRecord = (Document *) MemHandleLock(handle);
		FATAL_IF(categoryRecord == NULL, "MemHandleLock failed\n");
		ErrFatalDisplayIf(categoryRecord == NULL, "GetDefaultCategories: MemHandleLock failed");

		category = (Char *) (((void *) categoryRecord) + sizeof(Document));
		index = GetCategoryIndex(category);
		if (index < dmRecNumCategories)
			categories |= (1 << index);

		MemHandleUnlock(handle);
	}
	CloseDatabase(ViewerDB);
	ViewerDB = NULL;

	if (categories == 0)
		return 1;
	else 
		return categories;
}

// vim: ts=4
