/*
 * $Id: emailform.c,v 1.14 2000/10/22 20:17:32 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "const.h"
#include "debug.h"
#include "document.h"
#include "documentdata.h"
#include "emailform.h"
#include "prefsdata.h"
#include "resourceids.h"
#include "screen.h"
#include "util.h"

static void CheckMem(void) __attribute__ ((section ("plkr")));
static void ScrollMessage(Int16 lines) __attribute__ ((section ("plkr")));
static void SetScrollLabel(FieldPtr field, Int16 id) __attribute__ ((section ("plkr")));
static void UpdateScrollers(void) __attribute__ ((section ("plkr")));
static void InsertText(FieldType *field, Char *text) __attribute__ ((section ("plkr")));
static void GetMailto(void) __attribute__ ((section ("plkr")));
static void SetObjectPosition(UInt16 id, Boolean firstColumn) __attribute__ ((section ("plkr")));
static void EmailInitForm(void) __attribute__ ((section ("plkr")));
void DoMail(void) __attribute__ ((section ("plkr")));
void SetMailto(Int16 index) __attribute__ ((section ("plkr")));

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static Char toAddress[] = "plucker-dev@rubberchicken.org";
static FieldType *toField;
static FieldType *ccField;
static FieldType *subjectField;
static FieldType *messageField;
static FieldType *currentField = NULL;
static ScrollBarType *scrollBar;
static Int16 mailIdx = 0;
static UInt32 oldDepth = 0;

/***********************************************************************
 *
 * FUNCTION:    CheckMem
 *
 * DESCRIPTION: Quick and dirty check for enough memory to send a message.
 *              (5000 bytes for body 150 for headers and some more for palm)
 *
 * PARAMETERS:  none
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void CheckMem(void)
{
	MemHandle test = NULL;

	if(!(test = MemHandleNew(7000)))
		FrmAlert(warnLowEmailMem);
	if(test)
		MemHandleFree(test);
}

/***********************************************************************
 *
 * FUNCTION:    ScrollMessage
 *
 * DESCRIPTION: Scrolls the message field.
 *
 * PARAMETERS:  lines - lines to scroll
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ScrollMessage(Int16 lines)
{
	if (lines < 0)
		FldScrollField(messageField, -lines, winUp);
	else if (lines > 0)
		FldScrollField(messageField, lines, winDown);
}

/***********************************************************************
 *
 * FUNCTION:    SetScrollLabel
 *
 * DESCRIPTION: Set label/image for scroll arrows
 *
 * PARAMETERS:  field - pointer for field
 *              id    - object ID
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void SetScrollLabel(FieldPtr field, Int16 id)
{
	Char direction[2];

	if(FldScrollable(field, winUp))
		StrCopy(direction, "\010"); // UP
	else if(FldScrollable(field, winDown))
		StrCopy(direction, "\007"); // DOWN
	else
		StrCopy(direction, "\022"); // NONE

	CtlSetLabel(GetObjectPtr(id), direction);
}

/***********************************************************************
 *
 * FUNCTION:    UpdateScrollers
 *
 * DESCRIPTION: Updates the scrollbar and scroll arrows.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void UpdateScrollers(void)
{
	UInt16 scrollPos, textHeight, fieldHeight, maxValue;

	SetScrollLabel(toField, frmEmailToArrow);
	SetScrollLabel(ccField, frmEmailCcArrow);
	SetScrollLabel(subjectField, frmEmailSubjectArrow);

	FldGetScrollValues(messageField, &scrollPos, &textHeight, &fieldHeight);
	if(textHeight > fieldHeight)
		maxValue = textHeight - fieldHeight;
	else if(scrollPos)
		maxValue = scrollPos;
	else
		maxValue = 0;
	SclSetScrollBar(scrollBar, scrollPos, 0, maxValue, 6);
}

/***********************************************************************
 *
 * FUNCTION:    InsertText
 *
 * DESCRIPTION: Insert text in to a textfield and scroll to the top.
 *
 * PARAMETERS:  field	- field to insert into
 *				text	- text to insert
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void InsertText(FieldType *field, Char *text)
{
	FldInsert(field, text, StrLen(text));
	while (FldScrollable(field, winUp))
		FldScrollField(field, 1, winUp);

}

/***********************************************************************
 *
 * FUNCTION:    GetMailto
 *
 * DESCRIPTION: Get pointer into the Mailto database.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void GetMailto(void)
{
	MemHandle handle = NULL;
	MailtoData *mail;
	void *pointer;

	// Special case for dev-list
	if(mailIdx == 0) {
		InsertText(toField, toAddress);
		return;
	}

	if(ReturnDocumentHandle(mailIdx, &handle)) {
		InsertText(toField, toAddress);
		FrmAlert(errBadMailto);
		return;
	}

	pointer = MemHandleLock(handle);
	FATAL_IF(pointer == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(pointer == NULL, "GetMailto: MemHandleLock failed");
	pointer += sizeof(Document);
	mail = (MailtoData *)pointer;

	if(mail->to_off)
		InsertText(toField, pointer + mail->to_off);
	if(mail->cc_off)
		InsertText(ccField, pointer + mail->cc_off);
	if(mail->subject_off)
		InsertText(subjectField, pointer + mail->subject_off);
	if(mail->message_off)
		InsertText(messageField, pointer + mail->message_off);

	MemHandleUnlock(handle);
}

/***********************************************************************
 *
 * FUNCTION:    SetObjectPosition
 *
 * DESCRIPTION: Set position of object in form
 *
 * PARAMETERS:  id          - object ID
 *              firstColumn - object should be placed in first column
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void SetObjectPosition(UInt16 id, Boolean firstColumn)
{
	UInt16 index, x, y;
	FormType *form = FrmGetFormPtr(frmEmail);

	index = FrmGetObjectIndex(form, id);
	FrmGetObjectPosition(form, index, &x, &y);

	if (firstColumn)
		FrmSetObjectPosition(form, index, 0, y);
	else
		FrmSetObjectPosition(form, index, x + 7, y);
}

/***********************************************************************
 *
 * FUNCTION:    EmailInitForm
 *
 * DESCRIPTION: Initialize the email form.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void EmailInitForm(void)
{
	FormType *form = FrmGetFormPtr(frmEmail);

	toField = GetObjectPtr(frmEmailTo);
	ccField = GetObjectPtr(frmEmailCc);
	subjectField = GetObjectPtr(frmEmailSubject);
	messageField = GetObjectPtr(frmEmailMessage);
	scrollBar = GetObjectPtr(frmEmailScrollBar);

	GetMailto();

	if(Prefs()->scrollbar == SCROLLBAR_LEFT) {
		SetObjectPosition(frmEmailToLabel, false);
		SetObjectPosition(frmEmailTo, false);
		SetObjectPosition(frmEmailToArrow, true);
		SetObjectPosition(frmEmailCcLabel, false);
		SetObjectPosition(frmEmailCc, false);
		SetObjectPosition(frmEmailCcArrow, true);
		SetObjectPosition(frmEmailSubjectLabel, false);
		SetObjectPosition(frmEmailSubject, false);
		SetObjectPosition(frmEmailSubjectArrow, true);
		SetObjectPosition(frmEmailMessage, false);
		SetObjectPosition(frmEmailScrollBar, true);
	}

	if (Prefs()->screenDepth != BLACKWHITE) {
		oldDepth = Prefs()->screenDepth;
		Prefs()->screenDepth = BLACKWHITE;
		SetScreenMode();
	}
	FrmDrawForm(form);
	UpdateScrollers();
}

/***********************************************************************
 *
 * FUNCTION:    SetMailto
 *
 * DESCRIPTION: Set current address in the Mailto database.
 *
 * PARAMETERS:  index - Index of record
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void SetMailto(Int16 index)
{
	mailIdx = index;
}

/***********************************************************************
 *
 * FUNCTION:    EmailFormHandleEvent
 *
 * DESCRIPTION: Event handler for the email form.
 *
 * PARAMETERS:  event - pointer to an EventType structure
 *
 * RETURN:      True if the event was handled
 *
 ***********************************************************************/
Boolean EmailFormHandleEvent(const EventType *event)
{
	Boolean result = false;

	SET_A4_FROM_A5

	switch(event->eType) {
	case frmOpenEvent:
		CheckMem();
		EmailInitForm();
		result = true;
		break;

	case frmCloseEvent:
		if (oldDepth) {
			Prefs()->screenDepth = oldDepth;
			SetScreenMode();
			oldDepth = 0;
		}
		result = false;
		break;

	case ctlSelectEvent:
		switch(event->data.ctlEnter.controlID) {
		case frmEmailSend:
			if(!StrLen(FldGetTextPtr(toField))) {
				FrmAlert(warnNoTo);
				result = true;
				break;
			}
			DoMail();		// Fall through

		case frmEmailCancel:
			FrmGotoForm(Prefs()->lastForm);
			result = true;
			break;

		case frmEmailToArrow:
			if(FldScrollable(toField, winUp))
				FldScrollField(toField, 1, winUp);
			else if(FldScrollable(toField, winDown))
				FldScrollField(toField, 1, winDown);
			UpdateScrollers();
			break;

		case frmEmailCcArrow:
			if(FldScrollable(ccField, winUp))
				FldScrollField(ccField, 1, winUp);
			else if(FldScrollable(ccField, winDown))
				FldScrollField(ccField, 1, winDown);
			UpdateScrollers();
			break;

		case frmEmailSubjectArrow:
			if(FldScrollable(subjectField, winUp))
				FldScrollField(subjectField, 1, winUp);
			else if(FldScrollable(subjectField, winDown))
				FldScrollField(subjectField, 1, winDown);
			UpdateScrollers();
			break;
		}
		break;

	case fldEnterEvent:
		currentField = event->data.fldEnter.pField;
		break;

	case fldChangedEvent:
			UpdateScrollers();
		break;

	case keyDownEvent:
		switch(event->data.keyDown.chr) {
		case pageUpChr:
			if(FldScrollable(currentField, winUp))
				FldScrollField(currentField, 1, winUp);
			UpdateScrollers();
			break;

		case pageDownChr:
			if(FldScrollable(currentField, winDown))
				FldScrollField(currentField, 1, winDown);
			UpdateScrollers();
			break;

		case chrLineFeed:
			result = (currentField != messageField);
			break;	// no <CR> in other fields
		}
		break;

	case sclRepeatEvent:
		ScrollMessage(event->data.sclRepeat.newValue - event->data.sclRepeat.value);
		break;

	default:
		result = false;
	}

	RESTORE_A4

	return result;
}

/***********************************************************************
 *
 * FUNCTION:    DoMail
 *
 * DESCRIPTION: Send mail specified in the email form
 *
 * PARAMETERS:  none
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void DoMail()
{
	LocalID dbID = 0;
	Err err;
	UInt32 result;
	UInt16 cardNo;
	DmSearchStateType searchState;
	MailAddRecordParamsType message;

	MemSet(&message, sizeof(message), 0);

	message.secret = false;
	message.signature = true;
	message.confirmRead = false;
	message.confirmDelivery = false;
	message.priority = mailPriorityNormal;
	message.to = FldGetTextPtr(toField);
	message.cc = FldGetTextPtr(ccField);
	message.subject = FldGetTextPtr(subjectField);
	message.body = FldGetTextPtr(messageField);

	DmGetNextDatabaseByTypeCreator(true, &searchState, sysFileTApplication, 
				       'mail', true, &cardNo, &dbID);
	ErrNonFatalDisplayIf(!dbID, "DoMail: Could not find Mail Application");
	if (!dbID)
		return;		// Mail app not found. Must be m100!

	err = SysAppLaunch(cardNo, dbID, 0, sysAppLaunchCmdAddRecord, (void *)&message, &result);
	ErrNonFatalDisplayIf(err, "DoMail: Could not launch Mail Application");
}
// vim: ts=4
