/*
 * $Id: history.c,v 1.15 2000/10/03 19:27:29 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "debug.h"
#include "document.h"
#include "documentdata.h"
#include "history.h"
#include "util.h"

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define NOT_FOUND -1

/***********************************************************************
 *
 *      Private Variables
 *
 ***********************************************************************/
static History history;

/***********************************************************************
 *
 * FUNCTION:    RollHistoryIndex
 *
 * DESCRIPTION: Return next history index.
 *
 * PARAMETERS:  index - pointer to history index (upon return, it will 
 *                      contain the next history index)
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void RollHistoryIndex(Int16 *index)
{
	(*index)++;
	if (*index >= HISTORYSIZE)
		*index = 0;
}

/***********************************************************************
 *
 * FUNCTION:    RestoreDocData
 *
 * DESCRIPTION: Restore document data (vertical offset, first and last
 *              paragraph) for current document from history
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void RestoreDocData(void)
{
	DynamicDocument *dynamic;

	dynamic = (DynamicDocument *) MemHandleLock(GetDynamicHandle(
		history.documents[history.currentDocument].documentId));

	DmWrite(dynamic, OFFSETOF(DynamicDocument, verticalOffset), 
		(void *) &history.documents[history.currentDocument], 
		sizeof(HistoryData) - sizeof(Int16));

	MemPtrUnlock(dynamic);
}

/***********************************************************************
 *
 * FUNCTION:    InitHistory
 *
 * DESCRIPTION: Initialize the history structure.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void InitHistory(void)
{
        MemSet((void *) &history, sizeof(History), 0);
        history.addHistory = false;
        history.documents[0].documentId = HOME_PAGE_ID;
}

/***********************************************************************
 *
 * FUNCTION:    ReadHistory
 *
 * DESCRIPTION: Read history from session record
 *
 * PARAMETERS:  recordPtr - pointer to session record
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ReadHistory(void *recordPtr)
{
	recordPtr = (void*) recordPtr + sizeof(Header);

	MemMove((void *) &history, recordPtr, sizeof(History));
}

/***********************************************************************
 *
 * FUNCTION:    SaveHistory
 *
 * DESCRIPTION: Save history in session record
 *
 * PARAMETERS:  recordPtr - pointer to session record
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void SaveHistory(void *recordPtr)
{
	DmWrite(recordPtr, sizeof(Header), (void *) &history, sizeof(History));
}

/***********************************************************************
 *
 * FUNCTION:    AddHistoryDocument
 *
 * DESCRIPTION: Add document to history.
 *
 * PARAMETERS:  documentId - document ID
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void AddHistoryDocument(const Int16 documentId)
{
	if (!history.addHistory) {
		history.addHistory = true;
		return;
	}

	RollHistoryIndex(&history.currentDocument);
	history.lastDocument = history.currentDocument;

	if (history.currentDocument == history.firstDocument)
		RollHistoryIndex(&history.firstDocument);

	history.documents[history.currentDocument].documentId = documentId;
}

/***********************************************************************
 *
 * FUNCTION:    ResetHistory
 *
 * DESCRIPTION: Reset vertical offset and data for first and last 
 *              paragraph in the history structure
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ResetHistory(void)
{
	UInt8 i;

	for (i = 0; i < HISTORYSIZE; i++) {
		history.documents[i].verticalOffset = 0;
		history.documents[i].firstVisibleParagraph = 0;
		history.documents[i].firstParagraphY = 0;
		history.documents[i].lastVisibleParagraph = 0;
		history.documents[i].lastParagraphY = 0;
	}

}

/***********************************************************************
 *
 * FUNCTION:    SetHistoryFirstParagraph
 *
 * DESCRIPTION: Store data for first paragraph in current document from
 *		history
 *
 * PARAMETERS:  offset - offset to paragraph in document record
 *              y      - paragraph's y coordinate within document
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void SetHistoryFirstParagraph(const Int16 offset, const Int16 y)
{
	history.documents[history.currentDocument].firstVisibleParagraph = offset;
	history.documents[history.currentDocument].firstParagraphY = y;
}

/***********************************************************************
 *
 * FUNCTION:    SetHistoryLastParagraph
 *
 * DESCRIPTION: Store data for last paragraph in current document from
 *		history
 *
 * PARAMETERS:  offset - offset to paragraph in document record
 *              y      - paragraph's y coordinate within document
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void SetHistoryLastParagraph(const Int16 offset, const Int16 y)
{
	history.documents[history.currentDocument].lastVisibleParagraph = offset;
	history.documents[history.currentDocument].lastParagraphY = y;
}

/***********************************************************************
 *
 * FUNCTION:    SetHistoryVerticalOffset
 *
 * DESCRIPTION: Store vertical offset for current document from	history
 *
 * PARAMETERS:  offset - vertical offset in document
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void SetHistoryVerticalOffset(const Int16 offset)
{
	history.documents[history.currentDocument].verticalOffset = offset;
}

/***********************************************************************
 *
 * FUNCTION:    GetAddHistory
 *
 * DESCRIPTION: Get the add history status
 *
 * PARAMETERS:  None
 *
 * RETURN:      Add history status
 *
 ***********************************************************************/
const Boolean GetAddHistory(void)
{
	return history.addHistory;
}

/***********************************************************************
 *
 * FUNCTION:    GetHistoryDocID
 *
 * DESCRIPTION: Get the current document ID from the history.
 *
 * PARAMETERS:  None
 *
 * RETURN:      The document ID
 *
 ***********************************************************************/
const Int16 GetHistoryDocID(void)
{
	return history.documents[history.currentDocument].documentId;
}

/***********************************************************************
 *
 * FUNCTION:    GetHistoryNext
 *
 * DESCRIPTION: Get the next document from the history.
 *
 * PARAMETERS:  None
 *
 * RETURN:      The document ID (if there is no next document in the 
 *              history -1 is returned)
 *
 ***********************************************************************/
const Int16 GetHistoryNext(void)
{
	if (history.currentDocument == history.lastDocument)
		return NOT_FOUND;

	RollHistoryIndex(&history.currentDocument);

	RestoreDocData();

	return history.documents[history.currentDocument].documentId;
}

/***********************************************************************
 *
 * FUNCTION:    GetHistoryPrev
 *
 * DESCRIPTION: Get the previous document from the history.
 *
 * PARAMETERS:  None
 *
 * RETURN:      The document ID (if there is no previous document in the 
 *              history -1 is returned)
 *
 ***********************************************************************/
const Int16 GetHistoryPrev(void)
{
	if (history.currentDocument == history.firstDocument)
		return NOT_FOUND;

	if (--history.currentDocument < 0)
		history.currentDocument = HISTORYSIZE - 1;

	RestoreDocData();

	return history.documents[history.currentDocument].documentId;
}
// vim: ts=4
