/*
 * $Id: image.c,v 1.23 2000/10/03 19:27:29 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "const.h"
#include "control.h"
#include "debug.h"
#include "document.h"
#include "documentdata.h"
#include "history.h"
#include "image.h"
#include "mainform.h"
#include "os.h"
#include "prefsdata.h"
#include "resourceids.h"
#include "screen.h"
#include "uncompress.h"
#include "util.h"

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define MAX_IMAGE_SIZE 480000 // 60k * 8

/***********************************************************************
 *
 *      Internal Structures
 *
 ***********************************************************************/
typedef enum ShowImagesEnum{
	IMAGE_LIMBO = 0,
	ALL_IMAGES = 1,
	NO_IMAGES = 2
} ShowImagesType;

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static UInt16 lastImageID = 0;
static ShowImagesType imageStatus = IMAGE_LIMBO;

static WinHandle osWindow = NULL;
static Boolean largeImage = false;
static UInt32 oldDepth = 0;

/***********************************************************************
 *
 * FUNCTION:    ResetImageData
 *
 * DESCRIPTION: Clear the static image data
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ResetImageData(void)
{
	lastImageID = 0;
	imageStatus = IMAGE_LIMBO;
}

/***********************************************************************
 *
 * FUNCTION:    NotSupportedImageCompression
 *
 * DESCRIPTION: Determine if Palm OS image compression can be used by
 *              the device
 *
 * PARAMETERS:  imagePtr - pointer to image structure
 *
 * RETURN:      True if image compression is not supported
 *
 ***********************************************************************/
Boolean NotSupportedImageCompression(BitmapType *imagePtr)
{
	UInt32 osVersion = GetOSVersion();

	if (imageStatus != IMAGE_LIMBO)
		return (imageStatus == NO_IMAGES);

	if (imagePtr->flags.compressed) {
		if ((sysGetROMVerMajor(osVersion) == 0x03) && (sysGetROMVerMinor(osVersion) >= 2)) {
			if ((imagePtr->compressionType == BitmapCompressionTypeScanLine) 
			|| ((imagePtr->compressionType == BitmapCompressionTypeRLE) && 
			    (sysGetROMVerMinor(osVersion) >= 5)))
				imageStatus = ALL_IMAGES;
			else {
				imageStatus = NO_IMAGES;
			}
		} else {
			imageStatus = NO_IMAGES;
		}
	} else {
		imageStatus = ALL_IMAGES;
	}
	return (imageStatus == NO_IMAGES);
}

/***********************************************************************
 *
 * FUNCTION:    ShowNoImages
 *
 * DESCRIPTION: Determine if images should not be rendered 
 *
 * PARAMETERS:  None
 *
 * RETURN:      True if no images should be rendered
 *
 ***********************************************************************/
Boolean ShowNoImages(void)
{
	return (imageStatus == NO_IMAGES);
}

/***********************************************************************
 *
 * FUNCTION:    DrawInlineImage
 *
 * DESCRIPTION: Draw a bitmap within a document.
 *
 * PARAMETERS:  reference - the bitmap document's record number
 *              tContext  - pointer to text context
 *              width     - upon return, set to the width of the image
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void DrawInlineImage(const Int16 reference, const TextContext *tContext, Int16 *width)
{
	MemHandle imageHandle;
	MemHandle uncompressHandle = NULL;
	Document *imageDocument;
	BitmapType *imagePtr;
	Char imageText[12];

	if ((imageStatus == NO_IMAGES) || (!ShowImages(GetHistoryDocID()) && tContext->writeMode)) {
		StrCopy(imageText, "[img]");
		*width = FntCharsWidth(imageText, StrLen(imageText));
		WinDrawChars(imageText, StrLen(imageText), 
			     tContext->cursorX,
			     tContext->cursorY - FntCharHeight());
		return;
	}
	imageHandle = GetDocumentHandle(reference);
	imageDocument = (Document *) MemHandleLock(imageHandle);
	FATAL_IF(imageDocument == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(imageDocument == NULL, "DrawInlineImage: MemHandleLock failed");

	switch (imageDocument->type) {

	case DOCTYPE_TBMP_COMPRESSED:
		if (reference == lastImageID)
			uncompressHandle = GetUncompressImageHandle();
		else {
			uncompressHandle = Uncompress(imageDocument);
			lastImageID = reference;
		}
		if (uncompressHandle == NULL) {
			if (tContext->writeMode) {
				StrCopy(imageText, "[img]");
				*width = FntCharsWidth(imageText, StrLen(imageText));

				WinDrawChars(imageText, StrLen(imageText), 
		 			    tContext->cursorX,
					     tContext->cursorY - FntCharHeight());
			}
			MemHandleUnlock(imageHandle);
			return;
		}
		imagePtr = (BitmapType *) MemHandleLock(uncompressHandle);
		break;

	case DOCTYPE_TBMP:
		imagePtr = ((void *) imageDocument) + sizeof(Document);
		break;

	default:
		MemHandleUnlock(imageHandle);
		return;
	}

	*width = imagePtr->width;

	if (tContext->writeMode) {
		if (TooHighBitDepth(imagePtr->pixelSize) || !ShowImages(GetHistoryDocID())) {
			StrPrintF(imageText, "[img %dbpp]", imagePtr->pixelSize);
			*width = FntCharsWidth(imageText, StrLen(imageText));
			WinDrawChars(imageText, StrLen(imageText), 
				     tContext->cursorX,
				     tContext->cursorY - FntCharHeight());
		} else {
			WinDrawBitmap(imagePtr, 
				      tContext->cursorX, 
				      tContext->cursorY - imagePtr->height);
		}
	}

	if (imageDocument->type == DOCTYPE_TBMP_COMPRESSED)
		MemHandleUnlock(uncompressHandle);

	MSG(_("image #%d\n", reference));

	MemHandleUnlock(imageHandle);
}

/***********************************************************************
 *
 * FUNCTION:    GetImageMetrics
 *
 * DESCRIPTION: Get the height and width of the image.
 *
 * PARAMETERS:  reference - the bitmap document's record number
 *              width     - upon return, set to the width of the image
 *              height    - upon return, set to the height of the image
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void GetImageMetrics(const Int16 reference, Int16 *width, Int16 *height)
{
	MemHandle imageHandle;
	MemHandle uncompressHandle = NULL;
	Document *imageDocument;
	BitmapType *imagePtr;
	Char imageText[12];

	FATAL_IF(width == NULL || height == NULL, "NULL argument\n");
	ErrFatalDisplayIf(width == NULL || height == NULL, "GetImageMetrics: NULL argument");

	if ((imageStatus == NO_IMAGES) || !ShowImages(GetHistoryDocID())) {
		StrCopy(imageText, "[img]");
		*width = FntCharsWidth(imageText, StrLen(imageText));
		*height = FntCharHeight();
		return;
	}
	imageHandle = GetDocumentHandle(reference);
	imageDocument = (Document *) MemHandleLock(imageHandle);
	FATAL_IF(imageDocument == NULL, "MemHandleLock failed\n");
	ErrFatalDisplayIf(imageDocument == NULL, "GetImageMetrics: MemHandleLock failed");

	switch (imageDocument->type) {

	case DOCTYPE_TBMP_COMPRESSED:
		if (reference == lastImageID)
			uncompressHandle = GetUncompressImageHandle();
		else {
			uncompressHandle = Uncompress(imageDocument);
			lastImageID = reference;
		}
		if (uncompressHandle == NULL) {
			StrCopy(imageText, "[img]");
			*width = FntCharsWidth(imageText, StrLen(imageText));
			*height = FntCharHeight();
			MemHandleUnlock(imageHandle);
			return;
		}

		imagePtr = (BitmapType *) MemHandleLock(uncompressHandle);
		break;

	case DOCTYPE_TBMP:
		imagePtr = ((void *) imageDocument) + sizeof(Document);
		break;

	default:
		MemHandleUnlock(imageHandle);
		return;
	}

	if (NotSupportedImageCompression(imagePtr)) {
		FrmAlert(warnImageCompression);
		StrCopy(imageText, "[img]");
		*width = FntCharsWidth(imageText, StrLen(imageText));
		*height = FntCharHeight();
	}
	else if (TooHighBitDepth(imagePtr->pixelSize) || !ShowImages(GetHistoryDocID())) {
		StrPrintF(imageText, "[img %dbpp]", imagePtr->pixelSize);
		*width = FntCharsWidth(imageText, StrLen(imageText));
		*height = FntCharHeight();
	} else {
		*width = imagePtr->width;
		*height = imagePtr->height;
	}

	if (imageDocument->type == DOCTYPE_TBMP_COMPRESSED)
		MemHandleUnlock(uncompressHandle);

	MSG(_("width = %d, height = %d\n", *width, *height));

	MemHandleUnlock(imageHandle);
}

/***********************************************************************
 *
 * FUNCTION:    ShowImages
 *
 * DESCRIPTION: Determine if images should be rendered in this document
 *
 * PARAMETERS:  reference - the record number
 *
 * RETURN:      True if images should be rendered
 *
 ***********************************************************************/
Boolean ShowImages(Int16 reference)
{
	return !GetBitStatus(SHOW_IMAGES_ID, reference);
}

/***********************************************************************
 *
 * FUNCTION:    ShowImagesOn
 *
 * DESCRIPTION: Turn on image rendering for this document
 *
 * PARAMETERS:  reference - the record number
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ShowImagesOn(Int16 reference)
{
	SetBitStatus(SHOW_IMAGES_ID, reference, false);
}

/***********************************************************************
 *
 * FUNCTION:    ShowImagesOff
 *
 * DESCRIPTION: Turn off image rendering for this document
 *
 * PARAMETERS:  reference - the record number
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ShowImagesOff(Int16 reference)
{
	SetBitStatus(SHOW_IMAGES_ID, reference, true);
}

/***********************************************************************
 *
 * FUNCTION:    ViewTbmpDocument
 *
 * DESCRIPTION: View bitmap.
 *
 * PARAMETERS:  document - pointer to document record
 *              newPage  - true if the page is from the history
 *
 * RETURN:      False if the image could not be displayed because
 *              of insufficient memory
 *
 ***********************************************************************/
Boolean ViewTbmpDocument(Document *document, const Boolean newPage)
{
	Err error;
	UInt32 reqMem;
	BitmapType *bitmap;
	WinHandle screenWindow = WinGetDrawWindow();
	MemHandle uncompressHandle = NULL;
	RectangleType imageBounds = { { 0, 0 }, { MAX_EXTENT_X, ViewportBounds()->extent.y } };
	RectangleType screenArea = { { 0, ViewportBounds()->topLeft.y }, { MAX_EXTENT_X, ViewportBounds()->extent.y } };

	WinEraseRectangle(&screenArea, 0);

	if (ShowNoImages())
		return true;

	if (document->type == DOCTYPE_TBMP)
		bitmap = ((void *) document) + sizeof(Document);
	else {
		uncompressHandle = Uncompress(document);

		if (uncompressHandle == NULL)
			return false;

		bitmap = (BitmapType *) MemHandleLock(uncompressHandle);
		FATAL_IF(bitmap == NULL, "MemHandleLock failed\n");
		ErrFatalDisplayIf(bitmap == NULL, "ViewTbmpDocument: MemHandleLock failed");
	}

	if (NotSupportedImageCompression(bitmap)) {
		FrmAlert(warnImageCompression);
		if (uncompressHandle)
			MemHandleUnlock(uncompressHandle);
		return true;
	}
	// Make sure the image is not too big for the used screen depth
	reqMem = (UInt32)bitmap->width * (UInt32)bitmap->height * bitmap->pixelSize;
	if (reqMem > MAX_IMAGE_SIZE) {
		FrmAlert(warnLowImageMem);
		if (uncompressHandle)
			MemHandleUnlock(uncompressHandle);
		return true;
	}
	if (Prefs()->screenDepth != bitmap->pixelSize) {
		oldDepth = Prefs()->screenDepth;
		Prefs()->screenDepth = bitmap->pixelSize;
		SetScreenMode();
		FrmEraseForm(FrmGetFormPtr(Prefs()->toolbar));
		MainFormInit();
	}
	if (TooHighBitDepth(bitmap->pixelSize)) {
		FrmAlert(infoTooHighBitDepth);
		if (uncompressHandle)
			MemHandleUnlock(uncompressHandle);
		return true;
	}

	osWindow = WinCreateOffscreenWindow(bitmap->width, bitmap->height, screenFormat , &error);

	if (osWindow == NULL) {
		FrmAlert(warnLowImageMem);
		if (uncompressHandle)
			MemHandleUnlock(uncompressHandle);
		return true;
	}
	WinSetDrawWindow(osWindow);
	WinDrawBitmap(bitmap, 0, 0);

	InitializeTbmpData(bitmap->width, bitmap->height);
	largeImage = true;

	if (uncompressHandle)
		MemHandleUnlock(uncompressHandle);

	WinSetDrawWindow(screenWindow);
	WinCopyRectangle(osWindow, screenWindow, &imageBounds, 0, ViewportBounds()->topLeft.y, winPaint);

	return true;
}

/***********************************************************************
 *
 * FUNCTION:    DoImageMove
 *
 * DESCRIPTION: Move image up/down/left/right.
 *
 * PARAMETERS:  x - new x coordinate
 *              y - new y coordinate
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void DoImageMove(Int16 x, Int16 y)
{
	WinHandle screenWindow = WinGetDrawWindow();
	RectangleType imageBounds = { { x, y }, { MAX_EXTENT_X, ViewportBounds()->extent.y } };

	WinCopyRectangle(osWindow, screenWindow, &imageBounds, 0, ViewportBounds()->topLeft.y, winPaint);
}

/***********************************************************************
 *
 * FUNCTION:    LargeImage
 *
 * DESCRIPTION: Check if a large image is displayed
 *
 * PARAMETERS:  None
 *
 * RETURN:      True if Tbmp document is displayed
 *
 ***********************************************************************/
Boolean LargeImage(void)
{
	return largeImage;
}

/***********************************************************************
 *
 * FUNCTION:    DeleteOffScreenWindow
 *
 * DESCRIPTION: Delete offscreen window
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void DeleteOffScreenWindow(void)
{
	if (osWindow) {
		RectangleType imageBounds = { { 0, ViewportBounds()->topLeft.y }, { MAX_EXTENT_X, ViewportBounds()->extent.y } };

		WinEraseRectangle(&imageBounds, 0);

		WinDeleteWindow(osWindow, false);
		osWindow = NULL;
	}
	if (oldDepth) {
		Prefs()->screenDepth = oldDepth;
		SetScreenMode();
		oldDepth = 0;
	}
	largeImage = false;
}

// vim: ts=4
