/*
 * $Id: mainform.c,v 1.26 2000/10/03 19:27:30 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "bookmark.h"
#include "const.h"
#include "control.h"
#include "debug.h"
#include "document.h"
#include "emailform.h"
#include "history.h"
#include "image.h"
#include "link.h"
#include "mainform.h"
#include "paragraph.h"
#include "prefsdata.h"
#include "prefsform.h"
#include "resourceids.h"
#include "resultform.h"
#include "search.h"
#include "util.h"

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define NO_OFFSET 0

const UInt16 ToolbarButtonBitmaps[] = { 0, bmpBookmark, BarInfoBitmap, bmpSettings, bmpSettings };
const UInt16 ToolbarButtonCommands[] = { 0, mGoAddBookmark, mOptionsDetails, mOptionsPref, mOptionsControl };

/***********************************************************************
 *
 * FUNCTION:    HandleCommonMenuItems
 *
 * DESCRIPTION: Handle common menu items
 *
 * PARAMETERS:  itemID - ID for selected menu item
 *
 * RETURN:      True if event is handled
 *
 ***********************************************************************/
Boolean HandleCommonMenuItems(UInt16 itemID)
{
	FormType *form;
	Boolean handled = false;

	switch (itemID) {
	case mOptionsAbout:
		form = FrmInitForm(frmAbout);
		FrmDoDialog(form);
		FrmDeleteForm(form);
		handled = true;
		break;

	case mOptionsContact:
		form = FrmInitForm(frmContact);
		FrmDoDialog(form);
		FrmDeleteForm(form);

		form = FrmInitForm(frmContact2);
		if(FrmDoDialog(form) == frmContact2Send) {
			SetMailto(0);
			FrmGotoForm(frmEmail);
		}
		FrmDeleteForm(form);

		handled = true;
		break;

	case mOptionsPref:
		FrmPopupForm(frmPref);
		handled = true;
		break;

	case mOptionsControl:
		FrmPopupForm(frmControl);
		handled = true;
		break;

	default:
		break;
	}
	return handled;
}

/***********************************************************************
 *
 * FUNCTION:    HandleMenuEvent
 *
 * DESCRIPTION: Handle a menu event
 *
 * PARAMETERS:  itemID - ID for selected menu item
 *
 * RETURN:      True if event is handled
 *
 ***********************************************************************/
static Boolean HandleMenuEvent(UInt16 itemID)
{
	Boolean handled = false;

	if (HandleCommonMenuItems(itemID))
		return true;
	else switch (itemID) {
		case mGoHome:
		case mGoBack:
		case mGoForward:
			DoControlAction(mGoHome - itemID - 1);
			handled = true;
			break;

		case mGoSearch:
			FrmPopupForm(frmSearch);
			handled = true;
			break;

		case mGoSearchAgain:
			SearchAgain();
			handled = true;
			break;

		case mGoAddBookmark:
			FrmPopupForm(frmAddBookmark);
			handled = true;
			break;

		case mGoEditBookmarks:
			FrmGotoForm(frmEditBookmarks);
			handled = true;
			break;

		case mViewTopToolbar:
			Prefs()->toolbar = TOOLBAR_TOP;
			FrmGotoForm(TOOLBAR_TOP);
			handled = true;
			break;

		case mViewBottomToolbar:
			Prefs()->toolbar = TOOLBAR_BOTTOM;
			FrmGotoForm(TOOLBAR_BOTTOM);
			handled = true;
			break;

		case mViewNoToolbar:
		{
			DynamicDocument *dynamic;
			Int16 adjustment;

			dynamic = (DynamicDocument *) MemHandleLock(GetDynamicDocument());
			FATAL_IF(dynamic == NULL, "MemHandleLock failed\n");
			ErrFatalDisplayIf(dynamic == NULL, "MemHandleLock failed");
			if (dynamic->height <= TOP_LEFT_Y + EXTENT_Y)
				adjustment = -dynamic->verticalOffset;
			else
				adjustment = 0;
			MemHandleUnlock(GetDynamicDocument());

			DoPageMove(adjustment);
			Prefs()->toolbar = TOOLBAR_NONE;
			FrmGotoForm(TOOLBAR_NONE);
			handled = true;
			break;
		}

		case mOptionsDBMgr:
			FrmGotoForm(frmDBMgr);
			handled = true;
			break;

		case mOptionsDetails:
			FrmPopupForm(frmDetails);
			handled = true;
			break;
		
		default:
			break;
	}
	return handled;
}

/***********************************************************************
 *
 * FUNCTION:    MainFormInit
 *
 * DESCRIPTION: Initialize the main form.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void MainFormInit(void)
{
	FormType *mainForm = FrmGetFormPtr(Prefs()->toolbar);
	ScrollBarType *scrollbar;
	UInt16 scrollbarIndex = FrmGetObjectIndex(mainForm, frmMainScrollBar);

	scrollbar = FrmGetObjectPtr(mainForm, scrollbarIndex);

	if (Prefs()->scrollbar == SCROLLBAR_LEFT)
		FrmSetObjectPosition (mainForm, scrollbarIndex, 0, TopLeftY() + 2);
	else
		FrmSetObjectPosition (mainForm, scrollbarIndex, TopLeftX() + EXTENT_X, TopLeftY() + 2);

	FrmDrawForm(mainForm);

	if (Prefs()->toolbar != TOOLBAR_NONE) {
		GetControlBounds(mainForm, bmpMenu, MENUCONTROL);
		GetControlBounds(mainForm, bmpBookmark, BOOKMARKCONTROL);
		GetControlBounds(mainForm, frmMainPercentPopup, OFFSETCONTROL);
		GetControlBounds(mainForm, bmpDbase, DBCONTROL);
		GetControlBounds(mainForm, bmpFind, FINDCONTROL);
		GetControlBounds(mainForm, bmpAgain, AGAINCONTROL);
		GetControlBounds(mainForm, bmpLeft, LEFTCONTROL);
		GetControlBounds(mainForm, bmpHome, HOMECONTROL);
		GetControlBounds(mainForm, bmpRight, RIGHTCONTROL);

		CtlSetLabel(GetObjectPtr(frmMainPercentPopup), "0%");
	} else {
		ClearControlBounds();
	}
	Prefs()->lastForm = Prefs()->toolbar;
}

/***********************************************************************
 *
 * FUNCTION:    MainFormHandleEvent
 *
 * DESCRIPTION: Event handler for the main form.
 *
 * PARAMETERS:  event - pointer to an EventType structure
 *
 * RETURN:      True if the event was handled
 *
 ***********************************************************************/
Boolean MainFormHandleEvent(const EventType *event)
{
	Boolean successful = false;
	Boolean handled = false;

	SET_A4_FROM_A5

	switch (event->eType) {
	case frmOpenEvent:
	{
		Int16 reference;

		MainFormInit();

		/* Start with last visited page or home page at the very first activation */
		reference = GetHistoryDocID();
		SetVisitedLink(reference);
		successful = ViewDocument(reference, !GetAddHistory(), NO_OFFSET);

		if (!successful) {
			FrmAlert(warnInsufficientMemory);
			FrmGotoForm(frmDBMgr);
		}
		handled = true;
		break;
	}

	case frmGotoEvent:
		SetFindPatternData(event->data.frmGoto.matchPos, event->data.frmGoto.matchLen);
		JumpToDoc(event->data.frmGoto.recordNum, NO_OFFSET);
		DoPageMove(GetSearchAdjustment());

		ClearFindPatternData();

		handled = true;
		break;

	case frmUpdateEvent:
		if (event->data.frmUpdate.updateCode == frmUpdateDoc) {
			FrmEraseForm(FrmGetFormPtr(Prefs()->toolbar));
			MainFormInit();

			successful = ViewDocument(GetHistoryDocID(), !GetAddHistory(), NO_OFFSET);
			if (!successful) {
				FrmAlert(warnInsufficientMemory);
				FrmGotoForm(frmDBMgr);
			}

			handled = true;
		} else if (event->data.frmUpdate.updateCode == frmShowResult) {
			DoPageMove(GetSearchAdjustment());

			ViewDocument(GetHistoryDocID(), !GetAddHistory(), NO_OFFSET);
			handled = true;
		}
		break;

	case menuCmdBarOpenEvent:

		MenuCmdBarAddButton(menuCmdBarOnLeft, bmpRight, menuCmdBarResultMenuItem, mGoForward, NULL);
		MenuCmdBarAddButton(menuCmdBarOnLeft, bmpHome, menuCmdBarResultMenuItem, mGoHome, NULL);
		MenuCmdBarAddButton(menuCmdBarOnLeft, bmpLeft, menuCmdBarResultMenuItem, mGoBack, NULL);
		if (Prefs()->toolbarButton != 0) {
			MenuCmdBarAddButton(menuCmdBarOnLeft, ToolbarButtonBitmaps[Prefs()->toolbarButton], 
					    menuCmdBarResultMenuItem, ToolbarButtonCommands[Prefs()->toolbarButton], NULL);
		}
		MenuCmdBarAddButton(menuCmdBarOnLeft, bmpDbase, menuCmdBarResultMenuItem, mOptionsDBMgr, NULL);

		handled = false;
		break;

	case menuEvent:

		// Refresh display to get rid of command text in bottom left corner
		MenuEraseStatus(NULL);
		handled = HandleMenuEvent(event->data.menu.itemID);
		break;

	case sclRepeatEvent:
		DoPageMove(event->data.sclRepeat.value - event->data.sclRepeat.newValue);
		break;

	case keyDownEvent:
		switch (event->data.keyDown.chr) {
		case pageUpChr:
			DoPageMove(SCROLL_ONE_PAGE);
			handled = true;
			break;

		case pageDownChr:
			DoPageMove(-SCROLL_ONE_PAGE);
			handled = true;
			break;
		}
		break;

	case penDownEvent:
		MSG(_("pen down at (%d,%d)\n", event->screenX, event->screenY));

		// Tap in document area (i.e. not in header or on scrollbar) or 
		// tap in active area of header (i.e. title or graphical buttons)
		if (!LargeImage() && event->screenY > TopLeftY() && event->screenY < (TopLeftY() + ExtentY())) { // We're in the document
			switch(Prefs()->scrollbar) {
			case SCROLLBAR_LEFT:
				if (event->screenX > TOP_LEFT_X + 7) {
					HandlePenDown(event);
					handled = true;
				}
				break;

			default:
				if (event->screenX < EXTENT_X) {
					HandlePenDown(event);
					handled = true;
				}
				break;
			}
		} else {        // We're in the header
			HandlePenDown(event);
			handled = true;
		}
		break;

	case penMoveEvent:
		if (Prefs()->controlMode == MODE3 || LargeImage())
			HandlePenMove(event);
		handled = true;
		break;

	case penUpEvent:
		MSG(_("pen up at (%d,%d)\n", event->screenX, event->screenY));
		HandlePenUp(event);
		handled = true;
		break;

	default:
		handled = false;
	}

	RESTORE_A4

	return handled;
}
// vim: ts=4
