/*
 * $Id: os.c,v 1.15 2000/10/03 19:27:30 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 *
 * OS-specific stuff for plucker.
 */

#include "debug.h"
#include "os.h"
#include "resourceids.h"
#include "screen.h"

#include "SysZLib.h"

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
static FontID os3Styles[MAXSTYLES] =
{
	stdFont, largeBoldFont, largeBoldFont, largeFont,
	largeFont, boldFont, boldFont, boldFont
};

static FontID os2Styles[MAXSTYLES] =
{
	stdFont, boldFont, boldFont, boldFont,
	boldFont, stdFont, stdFont, stdFont
};

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static FontID *Styles;
static UInt32 osVersion;
static Boolean zlibStatus = false;

/***********************************************************************
 *
 * FUNCTION:    GetStyle
 *
 * DESCRIPTION: Return font style
 *
 * PARAMETERS:  style - numerical representation for font style
 *
 * RETURN:      Font ID
 *
 ***********************************************************************/
FontID GetStyle(const Int16 style)
{
	return Styles[style];
}

/***********************************************************************
 *
 * FUNCTION:    GetOSVersion
 *
 * DESCRIPTION: Return OS version
 *
 * PARAMETERS:  None
 *
 * RETURN:      OS version
 *
 ***********************************************************************/
UInt32 GetOSVersion(void)
{
	return osVersion;
}

/***********************************************************************
 *
 * FUNCTION:    ZLibSupported
 *
 * DESCRIPTION: Return status for ZLib support
 *
 * PARAMETERS:  None
 *
 * RETURN:      True if ZLib is supported
 *
 ***********************************************************************/
Boolean ZLibSupported(void)
{
	return zlibStatus;
}

/***********************************************************************
 *
 * FUNCTION:    OS3
 *
 * DESCRIPTION: PalmOS 3.x specific operations.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void OS3(void)
{
	Styles = os3Styles;

	if (sysGetROMVerMinor(osVersion) < 5)
		depthCheck = (Boolean (*)(UInt32)) &TooHighBitDepth_Default;
	else
		depthCheck = (Boolean (*)(UInt32)) &TooHighBitDepth_35;
	
	screenMode = (void (*)(void)) &SetScreenModeOS3;
	restoreMode = (Boolean (*)(void)) &RestoreScreenModeOS3;

	if ((SysLibFind("Z.lib", &ZLibRef) == 0) || (SysLibLoad('libr', 'ZLib', &ZLibRef) == 0)) {
		ZLibOpen(ZLibRef);
		zlibStatus = true;
	} else {
		zlibStatus = false;
		ZLibRef = NULL;
	}
}

/***********************************************************************
 *
 * FUNCTION:    OS2
 *
 * DESCRIPTION: PalmOS 2.x specific operations.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void OS2(void)
{
	Styles = os2Styles;

	depthCheck = (Boolean (*)(UInt32)) &TooHighBitDepth_Default;
	screenMode = (void (*)(void)) &SetScreenModeOS2;
	restoreMode = (Boolean (*)(void)) &RestoreScreenModeOS2;
}

/***********************************************************************
 *
 * FUNCTION:    OS_Start
 *
 * DESCRIPTION: Handle start of OS specific features.
 *
 * PARAMETERS:  None
 *
 * RETURN:      sysErrRomIncompatible if started on a PalmOS 1.0,
 *              otherwise 0.
 *
 ***********************************************************************/
Err OS_Start(void)
{
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &osVersion);

	switch (sysGetROMVerMajor(osVersion)) {
	case 0x01:
		FATAL(_("wrong ROM version detected (%lx)\n", osVersion));
		FrmAlert(infoWrongROMVersion);
		return sysErrRomIncompatible;

	case 0x02:
		OS2();
		break;

	case 0x03:
		OS3();
		break;

	default:
		OS2();
	}

	return 0;
}

/***********************************************************************
 *
 * FUNCTION:    OS_Stop
 *
 * DESCRIPTION: Handle stop of OS specific features.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void OS_Stop(void)
{
	ZLTeardown;
}
// vim: ts=4
