/*
 * $Id: paragraph.c,v 1.35 2000/10/23 18:11:36 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "anchor.h"
#include "const.h"
#include "debug.h"
#include "document.h"
#include "documentdata.h"
#include "history.h"
#include "image.h"
#include "os.h"
#include "paragraph.h"
#include "prefsdata.h"
#include "search.h"
#include "uncompress.h"
#include "util.h"

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define NO_OFFSET 0
#define NO_IMAGE 0

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/

// Used to keep track of multi-line anchors
static Boolean multilineAnchor;

// Used to keep track of multi-line italic
static Boolean multilineItalic;

// Used to keep track of multi-line underlining
static Boolean multilineUnderline;

// Used to know if margins should be added to the
// current line or only the following lines
static Boolean addMarginToCurrent;

// Used to keep track of new line functions
static Boolean newLine;

// Used to keep track of horizontal rules
static Boolean horizontalRule;

// Used to keep track of images
static Boolean image;

// Used to keep track of the search pattern
static Int16 patternCount;
static Int16 findPatternPos;
static Int16 findPatternLen;

// Used for full justification
static Int16 spaceCount = 0;	// How many spaces in current line
static Int16 bigSpace = 0;	// How many get 1 extra
static Int16 littleSpace = 0;	// Extra pixels in each

/***********************************************************************
 *
 * FUNCTION:    SetFindPatternData
 *
 * DESCRIPTION: Set position and length of find pattern
 *
 * PARAMETERS:  pos - position of find pattern
 *              len - length of find pattern
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void SetFindPatternData(const Int16 pos, const Int16 len)
{
	findPatternPos = pos;
	findPatternLen = len;
}

/***********************************************************************
 *
 * FUNCTION:    ClearFindPatternData
 *
 * DESCRIPTION: Clear position and length of find pattern
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ClearFindPatternData(void)
{
	findPatternPos = -1;
	findPatternLen = -1;
}

/***********************************************************************
 *
 * FUNCTION:    GetParagraphOffset
 *
 * DESCRIPTION: Find offset to paragraph in text stream.
 *
 * PARAMETERS:  document  - pointer to document
 *              paragraph - pointer to paragraph
 *
 * RETURN:      Paragraph offset
 *
 ***********************************************************************/
static Int16 GetParagraphOffset(Document *document, Paragraph *paragraph)
{
	Int16 pOffset = 0;
	Paragraph *p = ((void *) document) + sizeof(Document);

	while (p < paragraph) {
		pOffset += p->size;
		p = ((void *) p) + sizeof(Paragraph);
	}

	return pOffset;
}

/***********************************************************************
 *
 * FUNCTION:    AlignText
 *
 * DESCRIPTION: Align text/image
 *
 * PARAMETERS:  tContext - pointer to text context
 *              pContext - pointer to paragraph context
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void AlignText(TextContext *tContext, ParagraphContext *pContext)
{
	Int16 temp = pContext->maxPixels - pContext->linePixels;

	bigSpace = littleSpace = 0;

	if (pContext->type == ALIGNMENT_CENTER)
		tContext->cursorX += temp / 2;
	else if (pContext->type == ALIGNMENT_RIGHT)
		tContext->cursorX += temp;
	else if (pContext->type == ALIGNMENT_JUSTIFY && !multilineAnchor && spaceCount > 0 && temp < pContext->maxPixels / 4) {
		littleSpace = temp / spaceCount;	// each space gets pixels
		bigSpace = temp % spaceCount;		// this many get 1 extra
	}
}

/***********************************************************************
 *
 * FUNCTION:    CharIsSpace
 *
 * DESCRIPTION: Check if given character is a non-visible character.
 *
 * PARAMETERS:  c - character to check
 *
 * RETURN:      True if the given character is non-visible
 *
 ***********************************************************************/
static Boolean CharIsSpace(Char c)
{
	return (c <= ' ') && (c >= 0);
}

/***********************************************************************
 *
 * FUNCTION:    SetStyle
 *
 * DESCRIPTION: Set new font style.
 *
 * PARAMETERS:  pContext - pointer to paragraph context
 *              style    - new font style
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void SetStyle(ParagraphContext *pContext, Int16 style)
{
	if ((style >= MINSTYLES) && (style < MAXSTYLES)) {
		FntSetFont(GetStyle(style));
		pContext->fontHeight = FntLineHeight();
	}
}

/***********************************************************************
 *
 * FUNCTION:    StartItalic
 *
 * DESCRIPTION: Initialize a new block of italic text.
 *
 * PARAMETERS:  tContext - pointer to text context
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void StartItalic(TextContext *tContext)
{
	tContext->italic.topLeft.x = tContext->cursorX;
}

/***********************************************************************
 *
 * FUNCTION:    ContinueItalic
 *
 * DESCRIPTION: Handle several lines with italic text.
 *
 * PARAMETERS:  tContext - pointer to text context
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ContinueItalic(TextContext *tContext)
{
	StartItalic(tContext);
}

/***********************************************************************
 *
 * FUNCTION:    StopItalic
 *
 * DESCRIPTION: Mark the end of a block of italic text.
 *
 * PARAMETERS:  tContext - pointer to text context
 *              height   - height of line
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void StopItalic(TextContext *tContext, const Int16 height)
{
	if (tContext->writeMode) {
		tContext->italic.topLeft.y = tContext->cursorY - height;
		tContext->italic.extent.x = tContext->cursorX - tContext->italic.topLeft.x + 1;
		tContext->italic.extent.y = height / 2 + 1;

		// Move upper half of text 1 pixel to the right -> italic font
		WinScrollRectangle(&tContext->italic, winRight, 1, &tContext->italic);
		WinEraseRectangle(&tContext->italic, 0);
	}
}

/***********************************************************************
 *
 * FUNCTION:    StartUnderline
 *
 * DESCRIPTION: Initialize a new block of underlined text.
 *
 * PARAMETERS:  tContext - pointer to text context
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void StartUnderline(void)
{
	if (Prefs()->underlineMode)
		WinSetUnderlineMode(solidUnderline);
}

/***********************************************************************
 *
 * FUNCTION:    ContinueUnderline
 *
 * DESCRIPTION: Handle several lines with underlined text.
 *
 * PARAMETERS:  tContext - pointer to text context
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ContinueUnderline(void)
{
	StartUnderline();
}

/***********************************************************************
 *
 * FUNCTION:    StopUnderline
 *
 * DESCRIPTION: Mark the end of a block of underlined text.
 *
 * PARAMETERS:  tContext - pointer to text context
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void StopUnderline()
{
	if (Prefs()->underlineMode)
		WinSetUnderlineMode(noUnderline);
}

/***********************************************************************
 *
 * FUNCTION:    DrawHorizontalRule
 *
 * DESCRIPTION: Draw a horizontal rule.
 *
 * PARAMETERS:  tContext - pointer to text context
 *              height   - height of horizontal rule
 *              width    - width of horizontal rule
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void DrawHorizontalRule(TextContext *tContext, Int16 height, Int16 width)
{
	if (tContext->writeMode) {
		RectangleType hr;

		hr.topLeft.x = tContext->cursorX;
		hr.topLeft.y = tContext->cursorY - height;
		hr.extent.x = width;
		hr.extent.y = height;

		WinDrawRectangle(&hr, 0);
	}
}

/***********************************************************************
 *
 * FUNCTION:    InitializeParagraphContext
 *
 * DESCRIPTION: Initialize paragraph context with data from a paragraph.
 *
 * PARAMETERS:  pContext  - pointer to paragraph context
 *              paragraph - pointer to paragraph
 *              document  - pointer to document
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void InitializeParagraphContext(ParagraphContext *pContext, Paragraph *paragraph, 
				       Document *document)
{
	if (document->type == DOCTYPE_PHTML_COMPRESSED)
		pContext->position = (UInt8 *) MemHandleLock(GetUncompressDocHandle());
	else
		pContext->position = ((void *) document) + sizeof(Document) + document->paragraphs * sizeof(Paragraph);

	pContext->position += GetParagraphOffset(document, paragraph);
	pContext->last = pContext->position + paragraph->size;
	pContext->function = NULL;
	pContext->left = 0;
	pContext->right = 0;
	pContext->maxPixels = EXTENT_X;
	pContext->linePixels = 0;
	pContext->type = ALIGNMENT_LEFT;
	SetStyle(pContext, DEFAULTSTYLE);

	if (document->type == DOCTYPE_PHTML_COMPRESSED)
		MemHandleUnlock(GetUncompressDocHandle());
}

/***********************************************************************
 *
 * FUNCTION:    HandleAnchor
 * 
 * DESCRIPTION: Handle anchor data
 *
 * PARAMETERS:  pContext - pointer to paragraph context
 *              tContext - pointer to text context
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void HandleAnchor(ParagraphContext *pContext, TextContext *tContext)
{
	ParagraphFunction function = *(pContext->function);
	UInt8 *functionArgs = pContext->function + 1;
	Int16 reference = (*functionArgs) * 256 + *(functionArgs + 1);
	Int16 paragraphOffset = NO_OFFSET;
	Int16 imageref = NO_IMAGE;

	switch (function) {
	case ANCHOR_START:
		break;

	case NAMED_ANCHOR_START:
		paragraphOffset = *(functionArgs + 2) * 256 + *(functionArgs + 3);
		break;

	case IMAGE_ANCHOR_START:
		imageref = *(functionArgs + 2) * 256 + *(functionArgs + 3);
		break;

	default:
		return;	
	}	
	StartAnchor(tContext, reference, paragraphOffset, imageref);
}

/***********************************************************************
 *
 * FUNCTION:    HandleFunction
 * 
 * DESCRIPTION: Handle a function.
 *
 * PARAMETERS:  pContext      - pointer to paragraph context
 *              tContext      - pointer to text context
 *              functionWidth - upon return, set to the width of a 
 *                              handled inline image or the size of a
 *                              horizontal rule (will be 0 for other 
 *                              functions)
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void HandleFunction(ParagraphContext *pContext, TextContext *tContext, 
			   Int16 *functionWidth)
{
	ParagraphFunction function = *(pContext->function);
	UInt8 *functionArgs = pContext->function + 1;
	Int16 reference;
	Int16 width;
	Int16 widthPercentage;

	*functionWidth = 0;

	switch (function) {
	case ANCHOR_START:
	case NAMED_ANCHOR_START:
	case IMAGE_ANCHOR_START:
		if (tContext == NULL)
			break;

		tContext->activeAnchor = true;
		multilineAnchor = true;
		HandleAnchor(pContext, tContext);
		break;

	case ANCHOR_END:
		if (tContext == NULL)
			break;

		tContext->activeAnchor = false;
		multilineAnchor = false;
		StopAnchor(tContext, pContext->fontHeight);
		break;

	case SET_STYLE:
		SetStyle(pContext, (Int16) *functionArgs);
		break;

	case INLINE_IMAGE:
		image = true;
		reference = (*functionArgs) * 256 + *(functionArgs + 1);
		if (tContext == NULL) {
			GetImageMetrics(reference, functionWidth, &(pContext->fontHeight));
		} else if (tContext->activeAnchor) {
			RestartAnchor(tContext, pContext->fontHeight);
			DrawInlineImage(reference, tContext, functionWidth);
			StopImageAnchor(tContext, pContext->fontHeight, *functionWidth);
		} else
			DrawInlineImage(reference, tContext, functionWidth);
		break;

	case SET_MARGIN:
		if (addMarginToCurrent) {
			*functionWidth = (*functionArgs - pContext->left);
			if (tContext == NULL)
				*functionWidth += *(functionArgs + 1) - pContext->right;
		}
		pContext->left = *functionArgs;
		pContext->right = *(functionArgs + 1);
		break;

	case ALIGNMENT:
		if (tContext == NULL)
			pContext->type = *functionArgs;
		break;

	case NEWLINE:
		if (tContext == NULL)
			newLine = true;
		break;

	case HRULE:
		pContext->fontHeight = *functionArgs;
		width = *(functionArgs + 1);
		widthPercentage = *(functionArgs + 2);

		// Check values and assign default value if necessary 
		if (!pContext->fontHeight)
			pContext->fontHeight = DEFAULT_HRULE_SIZE;

		if (width)
			*functionWidth = width;
		else if (widthPercentage)
			*functionWidth = pContext->maxPixels * widthPercentage / 100;
		else
			*functionWidth = pContext->maxPixels;

		if (tContext == NULL) {
			horizontalRule = true;
			break;
		}
		DrawHorizontalRule(tContext, pContext->fontHeight, *functionWidth);

		// Restore the current font height
		pContext->fontHeight = FntLineHeight();

		break;

	case ITALIC_START:
		if (tContext == NULL)
			break;

		multilineItalic = true;
		StartItalic(tContext);
		break;

	case ITALIC_END:
		if (tContext == NULL)
			break;

		multilineItalic = false;
		StopItalic(tContext, pContext->fontHeight);
		break;

	case UNDERLINE_START:
		if (tContext == NULL)
			break;

		multilineUnderline = true;
		StartUnderline();
		break;

	case UNDERLINE_END:
		if (tContext == NULL)
			break;

		multilineUnderline = false;
		StopUnderline();
		break;
	}
}

/***********************************************************************
 *
 * FUNCTION:    GetNextToken
 *
 * DESCRIPTION: Get next token from paragraph context.
 *
 * PARAMETERS:  pContext  - pointer to paragraph context
 *              nextToken - token value
 *
 * RETURN:      The token type:
 *
 *                TOKEN_CHARACTER      character
 *                TOKEN_PARAGRAPH_END  end-of-paragraph
 *                TOKEN_FUNCTION       function. Get the details from 
 *                                     ParagraphContext.function
 *
 ***********************************************************************/
static TokenType GetNextToken(ParagraphContext *pContext, Char *nextToken)
{
	Char nextChar;
	Int16 offset;

	if (pContext->position >= pContext->last)
		return TOKEN_PARAGRAPH_END;

	nextChar = *((pContext->position)++);

	if (nextChar) {
		*nextToken = nextChar;
		return TOKEN_CHARACTER;
	}
	pContext->function = pContext->position;
	offset = 1 + (*(pContext->position) & 0x07);
	pContext->position += offset;

	patternCount += offset;

	return TOKEN_FUNCTION;
}

/***********************************************************************
 *
 * FUNCTION:    GetLineMetrics
 *
 * DESCRIPTION: Return the number of characters that will fit in a line.
 *
 * PARAMETERS:  pContext         - pointer to paragraph context
 *              skipLeadingSpace - Whether initial non-visible characters
 *                                 should be skipped
 *              length           - upon return, set to number of characters
 *                                 that will fit in the line
 *              height           - upon return, set to the height of the 
 *                                 line
 *
 * RETURN:      Nothing
 *
 * NOTE:        This function will return a length of zero if called 
 *              while at EOP
 *
 ***********************************************************************/
static void GetLineMetrics(ParagraphContext *pContext, Boolean skipLeadingSpace, 
			   Int16 *length, Int16 *height)
{
	Char nextToken;
	TokenType nextTokenType;
	Int16 tokenCount = 0;
	Int16 lastSpace = 0;
	Int16 linePixels = 0;
	Int16 lastSpacePixels = 0;
	Int16 charWidth;
	Char *startPosition = pContext->position;
	FontID initialFontStyle = FntGetFont();
	Int16 initialLeftMargin = pContext->left;
	Int16 initialRightMargin = pContext->right;

	*height = 0;
	spaceCount = 0;

	addMarginToCurrent = true;

	while (1) {
		nextTokenType = GetNextToken(pContext, &nextToken);

		if (nextTokenType == TOKEN_PARAGRAPH_END)
			break;
		else if (nextTokenType == TOKEN_FUNCTION) {
			HandleFunction(pContext, NULL, &charWidth);

			if (newLine) {
				newLine = false;
				tokenCount++;
				break;
			}
			if (horizontalRule) {
				horizontalRule = false;
				linePixels += charWidth;
				*height = pContext->fontHeight;
				tokenCount++;
				break;
			}
			if (image) {
				image = false;
				if (pContext->fontHeight > *height)
					*height = pContext->fontHeight;

				skipLeadingSpace = false;
			}
			linePixels += charWidth;
			if (linePixels > pContext->maxPixels)
				break;

			tokenCount++;
			continue;
		}
		addMarginToCurrent = false;

		if (skipLeadingSpace && CharIsSpace(nextToken))
			continue;

		skipLeadingSpace = false;

		if (nextToken == ' ') {
			spaceCount++;		// Count the spaces, leading ones are skipped above
			lastSpacePixels = linePixels;
			lastSpace = tokenCount;	// This is the count *before* the space
		}
		charWidth = FntCharWidth(nextToken);
		linePixels += charWidth;

		if (linePixels > pContext->maxPixels) {
			if (lastSpacePixels < pContext->maxPixels / 2) {
				lastSpacePixels = pContext->maxPixels;
				lastSpace = tokenCount;
			} else if (spaceCount)
				spaceCount--;

			break;
		}

		if (FntLineHeight() > *height)
			*height = pContext->fontHeight = FntLineHeight();

		tokenCount++;
	}
	if (pContext->fontHeight != *height)
		pContext->fontHeight = *height;

	pContext->position = startPosition;

	if (spaceCount && (linePixels == lastSpacePixels))
		spaceCount--;

	if ((linePixels > pContext->maxPixels) && (lastSpace != 0)) {
		pContext->linePixels = lastSpacePixels;
		*length = lastSpace;
	} else {
		pContext->linePixels = linePixels;
		*length = tokenCount;
	}
	pContext->maxPixels += (initialLeftMargin - pContext->left) + 
			       (initialRightMargin - pContext->right);

	// Restore font style and margins
	FntSetFont(initialFontStyle);
	pContext->left = initialLeftMargin;
	pContext->right = initialRightMargin;

	return;
}

/***********************************************************************
 *
 * FUNCTION:    WriteLine
 * 
 * DESCRIPTION: Draw characters from a paragraph onto the display.
 *
 * PARAMETERS:  pContext         - pointer to paragraph context
 *              tContext         - pointer to text context
 *              characters       - number of characters to draw
 *              skipLeadingSpace - Whether initial non-visible characters
 *                                 should be skipped
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void WriteLine(ParagraphContext *pContext, TextContext *tContext, 
		      Int16 characters, Boolean skipLeadingSpace)
{
	Boolean invertPattern = false;
	Char chars[characters + 1];
	Int16 count = 0;
	Int16 storeChar = 0;
	TokenType nextTokenType;
	Char nextChar;
	Int16 imageWidth;
	const Int16 findPatternStart = findPatternPos;
	const Int16 findPatternStop = findPatternStart + findPatternLen - 1;

	addMarginToCurrent = true;

	AlignText(tContext, pContext);

	while (count < characters) {
		nextTokenType = GetNextToken(pContext, &nextChar);

		if (nextTokenType == TOKEN_PARAGRAPH_END)
			break;
		else if (nextTokenType == TOKEN_FUNCTION) {
			if (storeChar > 0) {
				if (tContext->writeMode) {
					if (invertPattern)
						WinDrawInvertedChars(chars, storeChar, tContext->cursorX, 
											tContext->cursorY - FntCharHeight());
					else
						WinDrawChars(chars, storeChar, tContext->cursorX, 
									tContext->cursorY - FntCharHeight());
				}
				tContext->cursorX += FntCharsWidth(chars, storeChar);
				storeChar = 0;
			}
			HandleFunction(pContext, tContext, &imageWidth);
			tContext->cursorX += imageWidth;

			count++;
			patternCount++;
			if (image)
				skipLeadingSpace = false;
			continue;
		}
		addMarginToCurrent = false;

		patternCount++;

		if (skipLeadingSpace && CharIsSpace(nextChar))
			continue;

		skipLeadingSpace = false;

		if (patternCount == findPatternStart) {
			if (storeChar > 0) {
				if (tContext->writeMode) {
					WinDrawChars(chars, storeChar, tContext->cursorX, 
						     tContext->cursorY - FntCharHeight());
				}
				tContext->cursorX += FntCharsWidth(chars, storeChar);
				storeChar = 0;
			}
			invertPattern = true;
			tContext->patternOffset = tContext->cursorY - FntCharHeight() - TopLeftY();
		} 

		chars[storeChar++] = nextChar;
		count++;

		if (patternCount == findPatternStop) {
			if (tContext->writeMode) {
				WinDrawInvertedChars(chars, storeChar, tContext->cursorX, 
						     tContext->cursorY - FntCharHeight());
			}
			invertPattern = false;
			tContext->cursorX += FntCharsWidth(chars, storeChar);
			storeChar = 0;
		}

		if(pContext->type == ALIGNMENT_JUSTIFY && nextChar == ' ') {
			if (storeChar > 0) {
				if (tContext->writeMode) {
					WinDrawChars(chars, storeChar, tContext->cursorX, 
						     tContext->cursorY - FntCharHeight());
				}
				tContext->cursorX += FntCharsWidth(chars, storeChar);
				storeChar = 0;
			}
			if(bigSpace) {
				bigSpace--;
				tContext->cursorX++;
			}
			tContext->cursorX += littleSpace;
		}
	}
	if (storeChar > 0) {
		if (tContext->writeMode) {
			if (invertPattern)
				WinDrawInvertedChars(chars, storeChar, tContext->cursorX, 
						     tContext->cursorY - FntCharHeight());
			else
				WinDrawChars(chars, storeChar, tContext->cursorX, 
					     tContext->cursorY - FntCharHeight());
		}
		tContext->cursorX += FntCharsWidth(chars, storeChar);
	}
}

/***********************************************************************
 *
 * FUNCTION:    DrawParagraph
 *
 * DESCRIPTION: Draw a paragraph using the given text context.
 *
 * PARAMETERS:  tContext  - pointer to text context
 *              paragraph - pointer to paragraph
 *              document  - pointer to document
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void DrawParagraph(TextContext *tContext, Paragraph *paragraph, Document *document)
{
	ParagraphContext pContext;
	Int16 length;
	Int16 height;
	Int16 storeCount;

	InitializeParagraphContext(&pContext, paragraph, document);
	tContext->cursorX = TopLeftX();

	multilineAnchor = false;
	multilineItalic = false;
	multilineUnderline = false;
	newLine = false;
	horizontalRule = false;
	image = false;
	patternCount = GetParagraphOffset(document, paragraph) - 1;

	/* Check for extra paragraph spacing */
	tContext->cursorY += (paragraph->attributes & 0x07) * DEFAULT_PARAGRAPH_SPACING;

	if (document->type == DOCTYPE_PHTML_COMPRESSED)
		MemHandleLock(GetUncompressDocHandle());

	while (1) {
		storeCount = patternCount;
		GetLineMetrics(&pContext, true, &length, &height);
		patternCount = storeCount;

		if (length <= 0)
			break;

		tContext->cursorY += height;

		// Continue any anchor started on the line before.
		if (multilineAnchor) {
			Int16 tempX = tContext->cursorX;

			AlignText(tContext, &pContext);
			ContinueAnchor(tContext);

			tContext->cursorX = tempX;
		}

		// Continue any italic text started on the line before.
		if (multilineItalic)
			ContinueItalic(tContext);

		// Continue any underlined text started on the line before.
		if (multilineUnderline)
			ContinueUnderline();

		// Draw the characters and move the cursor... 
		WriteLine(&pContext, tContext, length, true);

		/* End any anchor which was going on the last line; continue it on the
		   next line. */
		if (multilineAnchor)
			StopAnchor(tContext, height);

		/* End any italic text which was going on the last line; continue it on the
		   next line. */
		if (multilineItalic)
			StopItalic(tContext, height);

		/* End any underlined text which was going on the last line; continue it on the
		   next line. */
		if (multilineUnderline)
			StopUnderline();

		tContext->cursorX = TopLeftX() + pContext.left;
		pContext.fontHeight = FntLineHeight();
	}
	if (document->type == DOCTYPE_PHTML_COMPRESSED)
		MemHandleUnlock(GetUncompressDocHandle());
}

// vim: ts=4
