/*
 * $Id: prefsdata.c,v 1.23 2000/10/22 20:17:32 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "const.h"
#include "debug.h"
#include "prefsdata.h"
#include "screen.h"

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/

// ID number and version
#define ViewerOldPrefID 0
#define ViewerPrefID 2
#define ViewerOldPrefSearchStringID 1
#define ViewerPrefSearchStringID 3
#define ViewerVersion 1

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static Preferences pref;
static Char *patterns = NULL;
static UInt16 patternsSize = 0;

/***********************************************************************
 *
 * FUNCTION:    CheckDatabase
 *
 * DESCRIPTION: Check ID for database.
 *
 * PARAMETERS:  db - pointer to the database ID of the database
 *
 * RETURN:      Pointer to the database ID of the database
 *
 * NOTE:        The function will first try to locate the requested
 *              database, if that db is not found (or if db is NULL)
 *              it will locate the first available db. 
 *
 ***********************************************************************/
static LocalID CheckDatabase(LocalID *db)
{
	DmSearchStateType stateInfo;
	UInt16 card, firstCard;
	LocalID id, firstId;
	Err status;

	status = DmGetNextDatabaseByTypeCreator(true, &stateInfo, (UInt32) ViewerDBType, 
						(UInt32) ViewerAppID, false, &card, &id);
	firstId = id;
	firstCard = card;

	if (*db != NULL) {
		while (status == 0) {
			if (*db == id) {
				pref.cardNo = card;
				return *db;
			}
			status = DmGetNextDatabaseByTypeCreator(false, &stateInfo, (UInt32) ViewerDBType, 
								(UInt32) ViewerAppID, false, &card, &id);
		}
	}
	pref.cardNo = firstCard;
	return (*db = firstId);
}

/***********************************************************************
 *
 * FUNCTION:    RemovePref
 *
 * DESCRIPTION: Remove preferences from the Preferences database.
 *
 * PARAMETERS:  prefID - preference to remove
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void RemovePref(UInt16 prefID)
{
	DmOpenRef ref;
	DmSearchStateType state;
	LocalID dbID;
	UInt16 cardNo;
	Int16 index;
    
	DmGetNextDatabaseByTypeCreator(true, &state, 'sprf', 'psys', false, &cardNo, &dbID);
	ref = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);
	if (ref == NULL)
		return;

	index = DmFindResource(ref, ViewerAppID, prefID, NULL); 
	if (index != -1)
		DmRemoveResource(ref, index);

	DmCloseDatabase(ref);
}

/***********************************************************************
 *
 * FUNCTION:    ReadPrefs
 *
 * DESCRIPTION: Retrieve preferences from the Preferences database.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void ReadPrefs(void)
{
	UInt16 prefSize = sizeof(Preferences);
	UInt16 tempSize = 0;

	MemSet(&pref, sizeof(Preferences), 0);

	if (PrefGetAppPreferences((UInt32) ViewerAppID, (UInt16) ViewerPrefID, &pref, 
				  &prefSize, true) == noPreferenceFound) {
		if (PrefGetAppPreferences((UInt32) ViewerAppID, (UInt16) ViewerOldPrefID, NULL, 
				  &tempSize, true) != noPreferenceFound)
			RemovePref(ViewerOldPrefID);

		RemovePref(ViewerPrefSearchStringID);

		pref.dbID = NULL;
		pref.strikethrough = false;
		pref.showDate = true;
		pref.showSize = true;
		pref.multipleSelect = true;
		pref.linkClick = true;
		pref.searchInAllPages = false;
		pref.caseSensitive = false;
		pref.toolbar = TOOLBAR_TOP;
		pref.lastForm = pref.toolbar;
		pref.scrollbar = SCROLLBAR_RIGHT;
		pref.controlMode = MODE1;
		pref.screenDepth = DEFAULT_DEPTH;
		pref.searchEntries = 0;
		pref.categories = 1;
		pref.cardNo = 0;
		pref.toolbarButton = 0;
		pref.filterMode = FILTER_OR;
		pref.underlineMode = false;
		StrCopy(pref.dbName, "");

		MemSet(&pref.select, sizeof(pref.select), SELECT_NONE);
	}
	pref.dbID = DmFindDatabase(pref.cardNo, pref.dbName);
	CheckDatabase(&pref.dbID);
}

/***********************************************************************
 *
 * FUNCTION:    WritePrefs
 *
 * DESCRIPTION: Store preferences in the Preferences database.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void WritePrefs(void)
{
	PrefSetAppPreferences((UInt32) ViewerAppID, (UInt16) ViewerPrefID, 
			      (Int16) ViewerVersion, &pref, sizeof(Preferences), 
			      true);
}

/***********************************************************************
 *
 * FUNCTION:    GetSearchString
 *
 * DESCRIPTION: Retrieve search string 
 *
 * PARAMETERS:  string - string pointer to store result
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void GetSearchString(Char *string)
{
	UInt16 tempSize = 0;
	patternsSize = 0;

	MemSet(string, MAX_PATTERN_LEN + 1, 0);

	if (PrefGetAppPreferences((UInt32) ViewerAppID, (UInt16) ViewerPrefSearchStringID, NULL, 
				  &patternsSize, true) == noPreferenceFound) {
		if (PrefGetAppPreferences((UInt32) ViewerAppID, (UInt16) ViewerOldPrefSearchStringID, NULL, 
				  &tempSize, true) != noPreferenceFound)
			RemovePref(ViewerOldPrefSearchStringID);

		pref.searchEntries = 0;
		StrCopy(string, "");
	} else {
		Char *temp;
		UInt8 i;

		if (patterns)
			MemPtrFree(patterns);
		patterns = (Char *)MemPtrNew(patternsSize);
		PrefGetAppPreferences((UInt32) ViewerAppID, (UInt16) ViewerPrefSearchStringID, 
				      patterns, &patternsSize, true);

		temp = patterns;
		for (i = 0; i < pref.searchEntries - 1; i++)
			temp = (void *) temp + StrLen(temp) + 1;
		StrNCopy(string, temp, MAX_PATTERN_LEN);
	}
}

/***********************************************************************
 *
 * FUNCTION:    GetSearchHistory
 *
 * DESCRIPTION: Retrieve search patterns
 *
 * PARAMETERS:  None
 *
 * RETURN:      Pointer to search pattern
 *
 ***********************************************************************/
Char *GetSearchHistory(void)
{
	return patterns;
}

/***********************************************************************
 *
 * FUNCTION:    SetSearchString
 *
 * DESCRIPTION: Store search patterns
 *
 * PARAMETERS:  string - pattern string
 *
 * RETURN:      Nothing 
 *
 ***********************************************************************/
void SetSearchString(Char *string)
{
	UInt8 i, count = 0;
	UInt16 size;
	Char *temp1 = NULL;
	Char *temp2 = NULL;
	Char *text = NULL;

	if (pref.searchEntries < 10) {
		size = patternsSize + StrLen(string) + 1;
		text = (Char *)MemPtrNew(size);
		temp1 = patterns;
		temp2 = text;

		for (i = 0; i < pref.searchEntries; i++) {
			if (StrCompare(temp1, string)) {
				StrCopy(temp2, temp1);
				temp2 = (void *) temp2 + StrLen(temp2) + 1;
				count++;
			} else
				size -= StrLen(string) + 1;

			temp1 = (void *) temp1 + StrLen(temp1) + 1;
		}
	} else {
		size = patternsSize - StrLen(patterns) + StrLen(string);
		text = (Char *)MemPtrNew(size);
		temp1 = (void *) patterns + StrLen(patterns) + 1;
		temp2 = text;

		for (i = 1; i < pref.searchEntries; i++) {
			if (StrCompare(temp1, string)) {
				StrCopy(temp2, temp1);
				temp2 = (void *) temp2 + StrLen(temp2) + 1;
				count++;
			} else
				size -= StrLen(string) - 1;

			temp1 = (void *) temp1 + StrLen(temp1) + 1;
		}	
	}
	StrCopy(temp2, string);
	pref.searchEntries = count + 1;

	PrefSetAppPreferences((UInt32) ViewerAppID, (UInt16) ViewerPrefSearchStringID, 
			      (Int16) ViewerVersion, text, 
			      size, true);

	MemPtrFree(text);
	if (patterns) {
		MemPtrFree(patterns);
		patterns = NULL;
	}
}

/***********************************************************************
 *
 * FUNCTION:    Prefs
 *
 * DESCRIPTION: Retrieve the value of a preference (through the pointer).
 *
 * PARAMETERS:  None
 *
 * RETURN:      Pointer for the Preferences structure
 *
 ***********************************************************************/
Preferences *Prefs(void)
{
	return &pref;
}
// vim: ts=4
