/*
 * $Id: resultform.c,v 1.15 2000/10/09 17:18:20 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "const.h"
#include "control.h"
#include "control.h"
#include "debug.h"
#include "os.h"
#include "prefsdata.h"
#include "resourceids.h"
#include "resultform.h"
#include "search.h"
#include "util.h"

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define DEFAULT_X 2
#define DEFAULT_Y 30
#define DEFAULT_EXTENT_X 156
#define DEFAULT_EXTENT_Y 11

#define MAX_RESULT 9

#define NOT_FOUND -1

static RectangleType rect = { { DEFAULT_X, DEFAULT_Y } , 
			      { DEFAULT_EXTENT_X, DEFAULT_EXTENT_Y }};

/***********************************************************************
 *
 *      Internal Structures
 *
 ***********************************************************************/
static struct {
	RectangleType coordinates;
	UInt16 record;
	UInt16 pos;
	UInt16 length;
} SearchResult[MAX_RESULT];

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static Int16 count;
static Int16 lineMatch;
static Boolean newLine;

static FieldType *fldPtr;
static Char matches[40];
static Char noMatches[40];
static Char searching[21];
static Char pattern[MAX_PATTERN_LEN + 1];

static FontID initialFontStyle;

/***********************************************************************
 *
 * FUNCTION:    ClearSearchResult
 *
 * DESCRIPTION: Clear result data variables
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ClearSearchResult(void)
{
	MemSet((void *)&SearchResult, sizeof(SearchResult), 0);
	count = 0;
	newLine = false;
	rect.topLeft.y = DEFAULT_Y;
}

/***********************************************************************
 *
 * FUNCTION:    DrawResultHeader
 *
 * DESCRIPTION: Write header info for search result
 *
 * PARAMETERS:  header - header string
 *
 * RETURN:      True if end of document has been reached.
 *
 ***********************************************************************/
Boolean DrawResultHeader(Char *header)
{
	UInt16 len = FntCharsWidth(header, StrLen(header));
	UInt16 deltaX = (DEFAULT_EXTENT_X - len) / 2;
	UInt16 coordY;

	if (newLine) {
		newLine = false;
		count++;
		rect.topLeft.y += DEFAULT_EXTENT_Y;
	}

	coordY = rect.topLeft.y + DEFAULT_EXTENT_Y / 2;

	WinEraseRectangle(&rect, 0);

	WinDrawLine(DEFAULT_X, coordY, deltaX, coordY);
	WinDrawChars(header, StrLen(header), DEFAULT_X + deltaX, coordY - 5);
	WinDrawLine(DEFAULT_X + deltaX + len + 1, coordY, DEFAULT_EXTENT_X - 3, coordY);

	if (count >= MAX_RESULT - 1) {
		count = 0;
		newLine = false;
		rect.topLeft.y = DEFAULT_Y;
		return true;
	}

	return false;
}

/***********************************************************************
 *
 * FUNCTION:    DrawResultString
 *
 * DESCRIPTION: Write search result
 *
 * PARAMETERS:  record - record ID
 *              pos    - poition of search pattern in document
 *              length - length of pattern
 *              result - result string
 *
 * RETURN:      True if end of document has been reached.
 *
 ***********************************************************************/
Boolean DrawResultString(UInt16 record, UInt16 pos, UInt16 length, Char *result)
{
	count++;

	rect.topLeft.y += DEFAULT_EXTENT_Y;
	SearchResult[count].coordinates = rect;
	SearchResult[count].record = record;
	SearchResult[count].pos = pos;
	SearchResult[count].length = length;

	WinDrawChars(result, StrLen(result), rect.topLeft.x, rect.topLeft.y);

	if (count >= MAX_RESULT - 1) {
		count = 0;
		newLine = false;
		rect.topLeft.y = DEFAULT_Y;
		return true;
	}

	newLine = true;

	return false;
}

/***********************************************************************
 *
 * FUNCTION:    ShowCancel
 *
 * DESCRIPTION: Display Cancel button
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ShowCancel(void)
{
	FormType *resultForm = FrmGetFormPtr(frmResult);

	FrmHideObject(resultForm, FrmGetObjectIndex(resultForm, frmResultStop));
	FrmShowObject(resultForm, FrmGetObjectIndex(resultForm, frmResultCancel));
}

/***********************************************************************
 *
 * FUNCTION:    ShowCancelMore
 *
 * DESCRIPTION: Display Cancel and Find More buttons
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ShowCancelMore(void)
{
	FormType *resultForm = FrmGetFormPtr(frmResult);

	ShowCancel();
	FrmShowObject(resultForm, FrmGetObjectIndex(resultForm, frmResultFindMore));
}

/***********************************************************************
 *
 * FUNCTION:    UpdateStatus
 *
 * DESCRIPTION: Display current status: "Matches for <pattern>"
 *                                      "No matches for <pattern>"
 *
 * PARAMETERS:  findMore - set to true if more records remains to 
 *                         be searched
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void UpdateStatus(Boolean findMore)
{
	if (!findMore && count == 0)
		FldSetTextPtr(fldPtr, noMatches);
	else
		FldSetTextPtr(fldPtr, matches);

	FldDrawField(fldPtr);
}

/***********************************************************************
 *
 * FUNCTION:    ResultFormInit
 *
 * DESCRIPTION: Initialize the result form.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void ResultFormInit(void)
{
	FormType *resultForm = FrmGetFormPtr(frmResult);
	Char str[30];

	GetSearchString(pattern);

	SysCopyStringResource(searching, strResultSearching);

	SysCopyStringResource(str, strResultMatches);
	StrPrintF(matches, "%s \"%s\"", str, pattern);

	SysCopyStringResource(str, strResultNoMatches);
	StrPrintF(noMatches, "%s \"%s\"", str, pattern);

	fldPtr = GetObjectPtr(frmResultStatus);
	FldSetTextPtr(fldPtr, searching);

	FrmDrawForm(resultForm);

	ClearSearchResult();

	initialFontStyle = FntSetFont(stdFont);
}

/***********************************************************************
 *
 * FUNCTION:    SearchResultHandleEvent
 *
 * DESCRIPTION: Event handler for the result form.
 *
 * PARAMETERS:  startFromBeginning - start from the first record in DB
 *
 * RETURN:      True if the event was handled
 *
 ***********************************************************************/
static Boolean SearchResultHandleEvent(Boolean startFromBeginning)
{
	EventType event;
	Boolean handled = false;
	Boolean done = true;
	Boolean findMore = false;

	while (1) {
		do {
			EvtGetEvent(&event, evtWaitForever);

			if (SysHandleEvent(&event))
				continue;

			handled = false;

			switch (event.eType) {

			case appStopEvent:
				EvtAddEventToQueue(&event);
				return false;

			case ctlSelectEvent:
				if (event.data.ctlEnter.controlID == frmResultStop) {
					UpdateStatus(false);
					ShowCancelMore();
					handled = true;
				} else
					handled = false;
				break;

			case penUpEvent:
				EvtFlushPenQueue();
				handled = false;
				break;

			case keyDownEvent:
				UpdateStatus(false);
				ShowCancelMore();
				handled = true;
				break;

			default:
				handled = false;
			}

			// Check if the form can handle the event
			if (!handled) {
				FormType *resultForm = FrmGetFormPtr(frmResult);
				FrmHandleEvent (resultForm, &event);
			} else
				return true;

		} while (EvtEventAvail());

		done = SearchInAllPages(pattern, startFromBeginning, &findMore);
		startFromBeginning = false;

		if (done) {
			UpdateStatus(findMore);
			if (findMore)
				ShowCancelMore();
			else
				ShowCancel();

			return true;
		}

	}
}

/***********************************************************************
 *
 * FUNCTION:    ResultFormHandleEvent
 *
 * DESCRIPTION: Event handler for the result form.
 *
 * PARAMETERS:  event - pointer to an EventType structure
 *
 * RETURN:      True if the event was handled
 *
 ***********************************************************************/
Boolean ResultFormHandleEvent(const EventType *event)
{
	Boolean handled = false;

	SET_A4_FROM_A5

	switch (event->eType) {
	case ctlSelectEvent:
		if (event->data.ctlEnter.controlID == frmResultFindMore) {
			FormType *resultForm = FrmGetFormPtr(frmResult);

			RectangleType resultRect = { { DEFAULT_X, DEFAULT_Y }, 
						     { DEFAULT_EXTENT_X, MAX_RESULT * DEFAULT_EXTENT_Y } };

			WinEraseRectangle(&resultRect, 0);

			FldSetTextPtr(fldPtr, searching);
			FldDrawField(fldPtr);

			FrmHideObject(resultForm, FrmGetObjectIndex(resultForm, frmResultCancel));
			FrmHideObject(resultForm, FrmGetObjectIndex(resultForm, frmResultFindMore));
			FrmShowObject(resultForm, FrmGetObjectIndex(resultForm, frmResultStop));

			ClearSearchResult();

			handled = SearchResultHandleEvent(false);
			break;
		} else if (event->data.ctlEnter.controlID != frmResultCancel) 
			break;

		FntSetFont(initialFontStyle);
		FrmReturnToForm(Prefs()->toolbar);
		FrmUpdateForm(Prefs()->toolbar, frmUpdateDoc);
		handled = true;
		break;

	case frmOpenEvent:
		ResultFormInit();
		handled = SearchResultHandleEvent(true);
		break;

	case frmCloseEvent:
		FntSetFont(initialFontStyle);
		handled = false;
		break;

	case penDownEvent:
		MSG(_("pen down at (%d,%d)\n", event->screenX, event->screenY));

		if (event->screenY > TopLeftY() && event->screenY < 130) {
			Int16 i;

			lineMatch = NOT_FOUND;

			for (i = 0; i < MAX_RESULT; i++)
				if (RctPtInRectangle(event->screenX, event->screenY, &(SearchResult[i].coordinates))) {
					WinInvertRectangle(&SearchResult[i].coordinates, 0);
					lineMatch = i;
					break;
				}
			handled = true;
		}
		break;

	case penUpEvent:
		MSG(_("pen up at (%d,%d)\n", event->screenX, event->screenY));

		if (event->screenY > TopLeftY() && event->screenY < 130) {
			if (lineMatch != NOT_FOUND) {
				WinInvertRectangle(&SearchResult[lineMatch].coordinates, 0);

				if (RctPtInRectangle(event->screenX, event->screenY, &(SearchResult[lineMatch].coordinates))) {
					EventType match;

					FrmReturnToForm(Prefs()->toolbar);

					MemSet(&match, sizeof(EventType), 0);

					match.eType = frmGotoEvent;
					match.data.frmGoto.formID = Prefs()->toolbar;
					match.data.frmGoto.recordNum = SearchResult[lineMatch].record;
					match.data.frmGoto.matchPos = SearchResult[lineMatch].pos;
					match.data.frmGoto.matchLen = SearchResult[lineMatch].length;
					match.data.frmGoto.matchFieldNum = 0;
					match.data.frmGoto.matchCustom = 0;
					EvtAddEventToQueue(&match);
				}
				SndPlaySystemSound(sndClick);
			}
			handled = true;
		}
		break;

	default:
		handled = false;
	}

	RESTORE_A4

	return handled;
}

// vim: ts=4
