/*
 * $Id: search.c,v 1.17 2000/10/22 20:17:33 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "debug.h"
#include "document.h"
#include "documentdata.h"
#include "paragraph.h"
#include "prefsdata.h"
#include "resourceids.h"
#include "resultform.h"
#include "search.h"
#include "uncompress.h"
#include "util.h"

static Char GetCaselessValue(Char c) __attribute__ ((section ("plkr")));
static void FindContext(Char *text, Int16 size, Int16 pos, UInt16 function, UInt16 len, Char *result) __attribute__ ((section ("plkr")));
static Boolean DoSearch(Char *text, Int16 size, Char *pattern, Int16 *pos, UInt16 *function, Boolean caseSensitive) __attribute__ ((section ("plkr")));
Boolean SearchInPage(Char *pattern, Boolean startAtTop) __attribute__ ((section ("plkr")));
Boolean SearchInAllPages(Char *pattern, Boolean startFromBeginning, Boolean *findMore) __attribute__ ((section ("plkr")));
void SearchAgain(void) __attribute__ ((section ("plkr")));

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define SEARCH_AGAIN false

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static Boolean foundString;
static UInt16 recordNum;
static Int16 position;

/***********************************************************************
 *
 * FUNCTION:    GetCaselessValue
 *
 * DESCRIPTION: Return caseless but still accented value for character
 *
 * PARAMETERS:  c - character
 *
 * RETURN:      Caseless value
 *
 ***********************************************************************/
static Char GetCaselessValue(Char c)
{
	if (IsUpper(GetCharAttr(), c))
		return (c + ' ');
	else
		return c;
}

/***********************************************************************
 *
 * FUNCTION:    FindContext
 *
 * DESCRIPTION: Find contextual information
 *
 * PARAMETERS:  text     - text string
 *              size     - size of text
 *              pos      - position of the pattern
 *              function - position of function before pattern
 *              len      - length of pattern
 *              result   - will contain the result string after return
 *                         of function
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void FindContext(Char *text, Int16 size, Int16 pos, UInt16 function, UInt16 len, Char *result)
{
	Int16 start, stop;
	Int16 i;

	start = function + 2 + (text[function + 1] & 0x07);
	stop = size;
	for (i = pos; i < pos + 30; i++)
		if (text[i] == '\0') {
			stop = i - 1;
			break;
		}

	if ((stop - start + 1) >= 32) {
		if (start == pos) {
			if ((stop - pos) >= 32)
				stop = pos + 30;
		} else if (stop == (pos + len)) {
			if ((pos - start) >= 32)
				start = pos - (31 - len + 1);
		} else {
			if ((pos - start) >= 16) 
				start = pos - 15;
			if ((stop - pos) >= 16)
				stop = pos + 15;
		}
	}
	StrNCopy(result, &text[start], stop - start + 1);
}

/***********************************************************************
 *
 * FUNCTION:    DoSearch
 *
 * DESCRIPTION: Find pattern in text string
 *
 * PARAMETERS:  text          - text string
 *              size          - size of text
 *              pattern       - pattern to search for
 *              pos           - start position, will contain the position
 *                              of the pattern if found
 *              function      - position of function before pattern,
 *                              set to NULL if you are not interested
 *                              in this information
 *              caseSensitive - indicates if the search should be 
 *                              case-sensitive or not
 *
 * RETURN:      True if pattern is found in text string
 *
 ***********************************************************************/
static Boolean DoSearch(Char *text, Int16 size, Char *pattern, Int16 *pos, 
			UInt16 *function, Boolean caseSensitive)
{
	Int16 m = StrLen(pattern); 
	Int16 i;
	Int16 j;

	if (function)
		*function = 0;

	for (i = *pos, j = 0; j < m && i < size; i++, j++) {
		if (text[i] == '\0') {
			if (function)
				*function = i;
			i += 1 + (text[i + 1] & 0x07);
			j = -1;
			continue;
		}
		if (caseSensitive) {
			if (text[i] != pattern[j]) {
				i -= j;
				j = -1;
			}
		} else {
			if (GetCaselessValue(text[i]) != GetCaselessValue(pattern[j])) {
				i -= j;
				j = -1;
			}
		}
	}
	if (j == m) {
		*pos = i - m;
		return true;
	} else
		return false;
}

/***********************************************************************
 *
 * FUNCTION:    SearchInPage
 *
 * DESCRIPTION: Search in current page
 *
 * PARAMETERS:  pattern       - pattern to search for
 *              startAtTop    - indicates if the search should start
 *                              at the top of the page or from the
 *                              current position
 *
 * RETURN:      False if the search should be stopped.
 *
 ***********************************************************************/
Boolean SearchInPage(Char *pattern, Boolean startAtTop)
{
	Document *document;
	Char *text;
	Int16 size;
	Boolean match = false;
	Boolean status = false;

	document = (Document *) MemHandleLock(GetCurrentDocument());
	FATAL_IF(document == NULL, "MemHandleLock failed\n"); 
	ErrFatalDisplayIf(document == NULL, "SearchInPage: MemHandleLock failed"); 

	if (document->type == DOCTYPE_TBMP || document->type == DOCTYPE_TBMP_COMPRESSED) {
		MemPtrUnlock(document);
		return false;
	}

	if (document->type == DOCTYPE_PHTML_COMPRESSED) {
		text = (Char *) MemHandleLock(GetUncompressDocHandle());
		size = MemPtrSize(text);
	} else {
		text = ((void *) document) + sizeof(Document) + document->paragraphs * sizeof(Paragraph);
		size = document->size;
	}

	if (startAtTop) {
		position = 0;
		foundString = false;
	} else
		position++;

	ClearFindPatternData();

	match = DoSearch(text, size, pattern, &position, NULL, Prefs()->caseSensitive);

	if (match) {
		foundString = true;
		SetFindPatternData(position, StrLen(pattern));
		status = true;

	} else if (foundString) {
		if (FrmAlert(confirmEndOfDoc) == 0) {
			SearchInPage(pattern, true);
			status = true;
		}
	} else {
		FrmAlert(errStringNotFound);
		status = false;
	}

	if (document->type == DOCTYPE_PHTML_COMPRESSED)
		MemHandleUnlock(GetUncompressDocHandle());

	MemPtrUnlock(document);

	return status;
}

/***********************************************************************
 *
 * FUNCTION:    SearchInAllPages
 *
 * DESCRIPTION: Search in all pages
 *
 * PARAMETERS:  pattern            - pattern to search for
 *              startFromBeginning - start from beginning of database
 *              findMore           - indicates if there are remaining
 *                                   records to search through
 *
 * RETURN:      False if the search should be stopped.
 *
 ***********************************************************************/
Boolean SearchInAllPages(Char *pattern, Boolean startFromBeginning, Boolean *findMore)
{
	Char *text;
	Int16 size;
	MemHandle record;
	UInt16 uniqueID;
	Boolean done = false;
	UInt16 numOfRecords = GetNumOfRecords();

	if (startFromBeginning) {
		recordNum = 0;
		position = 0;
	}
	*findMore = true;

	while (1) {
		MemHandle uncompressHandle = NULL;
		Document *document;
		Char header[5];
		Int32 percent;
		Err err;

		if (EvtSysEventAvail(true)) {
			done = false;
			break;
		}

		percent = recordNum * 100L / numOfRecords + 1;
		StrPrintF(header, "%3ld%%", percent);

		done = DrawResultHeader(header);

		if (done)
			break;

		err = ReturnNextDocumentHandle(&recordNum, &record, &uniqueID);

		if (err) {
			recordNum = 0;
			*findMore = false;
			done = true;
			break;
		}
		document = (Document *) MemHandleLock(record);
		FATAL_IF(document == NULL, "MemHandleLock failed\n"); 
		ErrFatalDisplayIf(document == NULL, "SearchInAllPages: MemHandleLock failed"); 

		if (document->type == DOCTYPE_PHTML_COMPRESSED) {
			uncompressHandle = Uncompress(document);
			if (uncompressHandle == NULL)
				goto nextRecord;
			text = (Char *) MemHandleLock(uncompressHandle);
			size = MemPtrSize(text);
		} else {
			text = ((void *) document) + sizeof(Document) + document->paragraphs * sizeof(Paragraph);
			size = document->size;
		}

		while (1) {
			Boolean match;
			UInt16 function;

			match = DoSearch(text, size, pattern, &position, &function, Prefs()->caseSensitive);

			if (match) {
				Char result[32];
				MemSet(result, 32, 0);

				FindContext(text, size, position, function, StrLen(pattern), result);
				done = DrawResultString(uniqueID, position, StrLen(pattern), result);
				position++;
			} else 
				break;

			if (done)
				break;
		}
		if (document->type == DOCTYPE_PHTML_COMPRESSED)
			MemHandleUnlock(uncompressHandle);

	nextRecord:

		MemHandleUnlock(record);

		if (done)
			break;

		recordNum++;
		position = 0;
	}
	return done;
}

/***********************************************************************
 *
 * FUNCTION:    SearchAgain
 *
 * DESCRIPTION: Search again in current page
 *
 * PARAMETERS:  None 
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void SearchAgain(void)
{
	Boolean contSearch;
	Char pattern[16];

	GetSearchString(pattern);

	if (!StrLen(pattern))
		return;

	contSearch = SearchInPage(pattern, SEARCH_AGAIN);

	if (contSearch)
		DoPageMove(GetSearchAdjustment());
}
// vim: ts=4
