/*
 * $Id: uncompress.c,v 1.20 2000/10/03 19:27:34 nordstrom Exp $
 *
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 */

#include "const.h"
#include "db.h"
#include "debug.h"
#include "document.h"
#include "paragraph.h"
#include "resourceids.h"
#include "uncompress.h"
#include "util.h"

#define NOZLIBDEFS
#include "SysZLib.h"

/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/
#define COUNT_BITS 3
#define BUFSIZE 4096
#define ZBUFSIZE 1024

static Char uncompressDBName[] = "Plkr-UnCoMpReSS";

/***********************************************************************
 *
 *      Private variables
 *
 ***********************************************************************/
static MemHandle uncompressDocHandle;
static MemHandle uncompressImageHandle;
static MemHandle uncompressHandle;

static DmOpenRef UncompressDB;

/***********************************************************************
 *
 * FUNCTION:    CreateUncompressDatabase
 *
 * DESCRIPTION: Create database used when uncompressing documents.
 *
 * PARAMETERS:  None
 *
 * RETURN:      The result from the database call
 *
 ***********************************************************************/
Err CreateUncompressDatabase(void)
{
	UInt16 cardNo = 0;
	Err err = 0;

	UncompressDB = DmOpenDatabaseByTypeCreator(UncompressDBType, ViewerAppID, dmModeReadWrite);
	if (!UncompressDB) {
		MemHandle handle;
		UInt16 dbIndex;
		UInt8 i;

		err = DmCreateDatabase (cardNo, uncompressDBName, ViewerAppID, UncompressDBType, false);
		if (err) {
			MSG("Couldn't create Uncompress database\n");
			FrmAlert(warnInsufficientMemory);
			return err;
		}
		UncompressDB = DmOpenDatabaseByTypeCreator(UncompressDBType, ViewerAppID, dmModeReadWrite);

		for (i = 0; i < 3; i++) {
			dbIndex = dmMaxRecordIndex;
			handle = DmNewRecord(UncompressDB, &dbIndex, 1);
			if (handle == NULL)
				err = DmGetLastErr();
			else
				DmReleaseRecord(UncompressDB, dbIndex, true);
		}
		if (err) {
			CloseUncompressDatabase();
			MSG("Couldn't initialize Uncompress database\n");
			FrmAlert(warnInsufficientMemory);
			return err;
		}
	}
	return err;
}

/***********************************************************************
 *
 * FUNCTION:    CloseUncompressDatabase
 *
 * DESCRIPTION: Close and delete Uncompress database.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
void CloseUncompressDatabase(void)
{
	UInt16 cardNo = 0;
	LocalID dbID;

	if (UncompressDB == NULL)
		return;

	CloseDatabase(UncompressDB);
	dbID = DmFindDatabase(cardNo, uncompressDBName);
	if (dbID)
		DmDeleteDatabase(cardNo, dbID);
	UncompressDB = NULL;
}

/***********************************************************************
 *
 * FUNCTION:    AllocateUncompressRecord
 *
 * DESCRIPTION: Allocate space for doc, image or special uncompress record
 *
 * PARAMETERS:  document - pointer to document
 *
 * RETURN:      The handle to the record, or NULL if not defined
 *
 ***********************************************************************/
static MemHandle AllocateUncompressRecord(Document *document)
{
	MemHandle hand = NULL;

	if (document->type == DOCTYPE_PHTML_COMPRESSED) {
		hand = uncompressDocHandle = DmResizeRecord(UncompressDB, UNCOMPRESS_DOC_ID, document->size);
	} else if (document->type == DOCTYPE_TBMP_COMPRESSED) {
		hand = uncompressImageHandle = DmResizeRecord(UncompressDB, UNCOMPRESS_IMAGE_ID, document->size);
	} else if (document->type == DOCTYPE_LINKS_COMPRESSED) {
		hand = uncompressHandle = DmResizeRecord(UncompressDB, UNCOMPRESS_SPECIAL_ID, document->size);
	}
	return hand;
}

/***********************************************************************
 *
 * FUNCTION:    GetUncompressDocHandle
 *
 * DESCRIPTION: Return handle to doc uncompress record
 *
 * PARAMETERS:  None
 *
 * RETURN:      The handle to the record
 *
 ***********************************************************************/
MemHandle GetUncompressDocHandle(void)
{
	return uncompressDocHandle;
}

/***********************************************************************
 *
 * FUNCTION:    GetUncompressImageHandle
 *
 * DESCRIPTION: Return handle to image uncompress record
 *
 * PARAMETERS:  None
 *
 * RETURN:      The handle to the record
 *
 ***********************************************************************/
MemHandle GetUncompressImageHandle(void)
{
	return uncompressImageHandle;
}

/***********************************************************************
 *
 * FUNCTION:    UnDoc
 *
 * DESCRIPTION: Uncompress DOC compressed document/image.
 *
 * PARAMETERS:  document - pointer to compressed document
 *
 * RETURN:      Handle to uncompress record, or NULL if
 *              failed to uncompress data
 *
 ***********************************************************************/
MemHandle UnDoc(Document *document)
{
	UInt16 headerSize = sizeof(Document) + document->paragraphs * sizeof(Paragraph);
	UInt8 *inBuf = ((void *) document) + headerSize;
	UInt16 docSize = MemPtrSize(document) - headerSize;
	UInt16 i, j = 0, k = 0;
	MemHandle uncompressHandle;
	UInt8 *outBuf;
	UInt8 *uncompressPtr;

	if (docSize > document->size)
		return NULL;

	outBuf = (UInt8 *) MemPtrNew(BUFSIZE);
	if (outBuf == NULL)
		return NULL;

        uncompressHandle = AllocateUncompressRecord(document);
	if (uncompressHandle == NULL) {
		MemPtrFree(outBuf);
		return NULL;
	}

        uncompressPtr = (UInt8 *) MemHandleLock(uncompressHandle);
        FATAL_IF(uncompressPtr == NULL, "MemHandleLock failed\n");
        ErrFatalDisplayIf(uncompressPtr == NULL, "Uncompress: MemHandleLock failed");

	while (j < docSize) {
		i = 0;
		while (i < BUFSIZE && j < docSize) {
			UInt16 c;

			c = (UInt16) inBuf[j++];

			if (c > 0 && c < 9)
				while (c--)
					outBuf[i++] = inBuf[j++];

			else if (c < 0x80)
				outBuf[i++] = c;

			else if (c >= 0xc0) {
				outBuf[i++] = ' ';
				outBuf[i++] = c ^ 0x80;
			} else {
				Int16 m, n;

				c <<= 8;
				c += inBuf[j++];
				m = (c & 0x3fff) >> COUNT_BITS;
				n = c & ((1 << COUNT_BITS) - 1);
				n += 2;
				do {
					outBuf[i] = outBuf[i - m];
					i++;
				} while (n--);
			}
		}
		DmWrite(uncompressPtr, k, outBuf, i);
		k += BUFSIZE;
	}
	MemPtrFree(outBuf);
	outBuf = NULL;
	MemHandleUnlock(uncompressHandle);

	return uncompressHandle;
}

/***********************************************************************
 *
 * FUNCTION:    UnZip
 *
 * DESCRIPTION: Uncompress ZLib compressed document/image.
 *
 * PARAMETERS:  document - pointer to compressed document
 *
 * RETURN:      Handle to uncompress record, or NULL if
 *              failed to uncompress data
 *
 ***********************************************************************/
MemHandle UnZip(Document *document)
{
	UInt32 outSize = 0;
	z_stream z;
	Int16 err;
	UInt8 *inBuf;
	UInt16 docSize;
	MemHandle uncompressHandle;
	UInt8 *outBuf;
	UInt8 *uncompressPtr;
	UInt16 headerSize;

	SET_A4_FROM_A5

	headerSize = sizeof(Document) + document->paragraphs * sizeof(Paragraph);
	inBuf = ((void *) document) + headerSize;
	docSize = MemPtrSize(document) - headerSize;

	if (docSize > document->size)
		return NULL;

        uncompressHandle = AllocateUncompressRecord(document);
	if (uncompressHandle == NULL) {
		return NULL;
	}

        uncompressPtr = (UInt8 *) MemHandleLock(uncompressHandle);
        FATAL_IF(uncompressPtr == NULL, "MemHandleLock failed\n");
        ErrFatalDisplayIf(uncompressPtr == NULL, "Uncompress: MemHandleLock failed");

	MemSet(&z, sizeof(z), 0);
	z.next_in = inBuf;
	z.avail_in = docSize;
	err = inflateInit(&z);
	if (err != Z_OK) {
		MemHandleUnlock(uncompressHandle);
		return NULL;
	}

	outBuf = (UInt8 *) MemPtrNew(ZBUFSIZE);
	if (outBuf == NULL) {
		inflateEnd(&z);
		MemHandleUnlock(uncompressHandle);
		return NULL;
	}

	do {
		z.next_out = outBuf;
		z.avail_out = ZBUFSIZE;
	
		err = inflate(&z, Z_FINISH);
		DmWrite(uncompressPtr, outSize, outBuf, ZBUFSIZE - z.avail_out);

		outSize += ZBUFSIZE - z.avail_out;
	} while (z.avail_out != ZBUFSIZE);
	inflateEnd(&z);

	MemPtrFree(outBuf);
	outBuf = NULL;
	MemHandleUnlock(uncompressHandle);

	RESTORE_A4

	return uncompressHandle;
}
// vim: ts=4
