/*
 * $Id: viewer.c,v 1.29 2000/10/03 19:27:35 nordstrom Exp $
 *
 * Viewer - a learning experience for me in programming the Palm Pilot.
 * This program is intended to view HTML documents that have been rendered
 * by an external renderer into a palm-specific, easily displayed format.
 *
 * See the file COPYING for more information about the following license.
 * 
 * Viewer - a part of Plucker, the free off-line HTML viewer for PalmOS
 * Copyright (C) 1998, Mark Ian Lillywhite and Michael Nordstrm
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Mark Lillywhite - plucker@rubberchicken.org
 * http://rubberchicken.org/products/plucker/
 *
 * Originally based on code by Andrew Howlett from his Tex2Hex demo. (Thanks!)
 */

#include "allforms.h"
#include "categorydata.h"
#include "const.h"
#include "debug.h"
#include "document.h"
#include "documentdata.h"
#include "image.h"
#include "os.h"
#include "paragraph.h"
#include "prefsdata.h"
#include "resourceids.h"
#include "screen.h"
#include "session.h"
#include "uncompress.h"
#include "util.h"
#include "viewer.h"

/***********************************************************************
 *
 * FUNCTION:    StartApplication
 *
 * DESCRIPTION: This routine opens the application's database, loads the
 *              saved-state information and initializes global variables.
 *
 * PARAMETERS:  cmdPBP - pointer to command parameter block
 *
 * RETURN:      Error code or zero if successful
 *
 * NOTE:        If the database is broken the user will be alerted of this 
 *              and the database manager form will be displayed instead of
 *              the main form
 *
 ***********************************************************************/
static Err StartApplication(MemPtr cmdPBP)
{
	Err err;
	UInt16 formId;

	err = OS_Start();
	if (err)
		return err;

	ReadPrefs();
	
	SetScreenMode();

	formId = Prefs()->toolbar;

	if (cmdPBP) {
		UInt32 type, creator;
		UInt16 cardNo = ((SysAppLaunchCmdOpenDBType*)cmdPBP)->cardNo;
		LocalID dbID = ((SysAppLaunchCmdOpenDBType*)cmdPBP)->dbID;
	
		DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &type, &creator);
		
		if (creator == ViewerAppID && type == ViewerDBType) {
			Prefs()->dbID = dbID;
			Prefs()->cardNo = cardNo;
		}
	}

	err = CreateUncompressDatabase();
	if (err == dmErrCantOpen)
		return err;
	else if (err)
		formId = frmDBMgr;
	else {
		err = OpenDocumentDatabase(Prefs()->dbID, Prefs()->cardNo);
		if (err == dmErrCantOpen) {
			CloseUncompressDatabase();
			return err;
		} else if (err)
			formId = frmDBMgr;
	}
	InitSessionData();
	ClearFindPatternData();

	FrmGotoForm(formId);

	return 0;
}

/***********************************************************************
 *
 * FUNCTION:    EventLoop
 *
 * DESCRIPTION: This routine is the event loop for the viewer application
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void EventLoop(void)
{
	UInt16 err;
	UInt16 formID;
	FormType *form;
	EventType event;

	do {
		EvtGetEvent(&event, evtWaitForever);

		if (SysHandleEvent(&event))
			continue;

		if (MenuHandleEvent((void *) NULL, &event, &err))
			continue;

		if (event.eType == frmLoadEvent) {
			formID = event.data.frmLoad.formID;
			form = FrmInitForm(formID);
			FrmSetActiveForm(form);
			switch (formID) {
			case frmMainTop:
			case frmMainBottom:
			case frmMainNone:
				FrmSetEventHandler(form, (FormEventHandlerType *) MainFormHandleEvent);
				break;

			case frmPref:
				FrmSetEventHandler(form, (FormEventHandlerType *) PrefsFormHandleEvent);
				break;

			case frmDBMgr:
				FrmSetEventHandler(form, (FormEventHandlerType *) DBmgrFormHandleEvent);
				break;

			case frmDetails:
				FrmSetEventHandler(form, (FormEventHandlerType *) DetailsFormHandleEvent);
				break;

			case frmControl:
				FrmSetEventHandler(form, (FormEventHandlerType *) ControlFormHandleEvent);
				break;

			case frmEmail:
				FrmSetEventHandler(form, (FormEventHandlerType *) EmailFormHandleEvent);
				break;
			case frmSearch:
				FrmSetEventHandler(form, (FormEventHandlerType *) SearchFormHandleEvent);
				break;

			case frmResult:
				FrmSetEventHandler(form, (FormEventHandlerType *) ResultFormHandleEvent);
				break;

			case frmAddBookmark:
				FrmSetEventHandler(form, (FormEventHandlerType *) AddBookmarkFormHandleEvent);
				break;

			case frmEditBookmarks:
				FrmSetEventHandler(form, (FormEventHandlerType *) EditBookmarksFormHandleEvent);
				break;

			case frmRenameDB:
				FrmSetEventHandler(form, (FormEventHandlerType *) RenameDBFormHandleEvent);
				break;

			case frmExternalLinks:
				FrmSetEventHandler(form, (FormEventHandlerType *) ExternalLinksFormHandleEvent);
				break;

			case frmCategory:
				FrmSetEventHandler(form, (FormEventHandlerType *) CategoryFormHandleEvent);
				break;

			case frmNewCategory:
				FrmSetEventHandler(form, (FormEventHandlerType *) CategoryNameFormHandleEvent);
				break;
			}
		}
		FrmDispatchEvent(&event);

	} while (event.eType != appStopEvent);
}

/***********************************************************************
 *
 * FUNCTION:    StopApplication
 *
 * DESCRIPTION: This routine closes the application's database and saves
 *              the current state of the application.
 *
 * PARAMETERS:  None
 *
 * RETURN:      Nothing
 *
 ***********************************************************************/
static void StopApplication(void)
{
	DeleteOffScreenWindow();
	ReleaseTable();

	FrmSaveAllForms();
	FrmCloseAllForms();

	SaveSessionData();
	CloseDocumentDatabase();
	ClosePlkrDatabase();
	CloseUncompressDatabase();
	WritePrefs();
	RestoreScreenMode();

	OS_Stop();
}

/***********************************************************************
 *
 * FUNCTION:    PilotMain
 *
 * DESCRIPTION: This is the main entry point for the viewer application.
 *
 * PARAMETERS:  cmd         - SysAppLaunch Command
 *              cmdPBP      - pointer to command parameter block
 *              launchFlags - launch flag (see SystemMgr.h for details)
 *
 * RETURN:      Error code or zero if successful
 *
 ***********************************************************************/
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err err = 0;

	if (cmd == sysAppLaunchCmdNormalLaunch) {
		err = StartApplication(NULL);
		if (err)
			return err;

		EventLoop();
		StopApplication();
	} else if (cmd == sysAppLaunchCmdOpenDB) {
		err = StartApplication(cmdPBP);
		if (err)
			return err;

		EventLoop();
		StopApplication();
	}
	return err;
}
// vim: ts=4
